"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.failProxy = void 0;
const node_assert_1 = require("node:assert");
const client_utils_1 = require("@fluid-internal/client-utils");
const container_definitions_1 = require("@fluidframework/container-definitions");
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../blobManager/index.js");
const containerHandleContext_js_1 = require("../containerHandleContext.js");
const failProxy = (handler = {}) => {
    const proxy = new Proxy(handler, {
        get: (t, p, r) => {
            if (p === "then") {
                return undefined;
            }
            if (handler !== undefined && p in handler) {
                return Reflect.get(t, p, r);
            }
            throw new Error(`${p.toString()} not implemented`);
        },
    });
    return proxy;
};
exports.failProxy = failProxy;
function createBlobManager(overrides) {
    const runtime = (0, exports.failProxy)({
        baseLogger: (0, internal_2.createChildLogger)(),
        attachState: container_definitions_1.AttachState.Attached,
        resolveHandle: async () => {
            throw new Error("not implemented");
        },
    });
    const routeContext = new containerHandleContext_js_1.ContainerFluidHandleContext("/", runtime, undefined);
    return new index_js_1.BlobManager((0, exports.failProxy)({
        // defaults, these can still be overridden below
        runtime,
        routeContext,
        blobManagerLoadInfo: {},
        stashedBlobs: undefined,
        localBlobIdGenerator: undefined,
        isBlobDeleted: () => false,
        blobRequested: () => { },
        createBlobPayloadPending: false,
        // overrides
        ...overrides,
    }));
}
const blobAttachMessage = {
    clientId: "clientid",
    minimumSequenceNumber: 1,
    referenceSequenceNumber: 1,
    sequenceNumber: 1,
    type: "blobAttach",
    timestamp: Date.now(),
};
describe("BlobHandles", () => {
    it("Create blob", async () => {
        // Deferred promise that will be resolve once we send a blob attach. It is used mainly
        // to simulate correct order or blob operations: create -> onUploadResolve -> process.
        const d = new internal_1.Deferred();
        const blobManager = createBlobManager({
            sendBlobAttachOp(_localId, _storageId) {
                d.resolve();
            },
            stashedBlobs: {},
            localBlobIdGenerator: () => "localId",
            isBlobDeleted: () => false,
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return { id: "blobId" };
                },
                readBlob: async () => {
                    return (0, client_utils_1.stringToBuffer)("contentFromStorage", "utf8");
                },
            }),
        });
        const blob = (0, client_utils_1.stringToBuffer)("content", "utf8");
        const blobHandleP = blobManager.createBlob(blob);
        await d.promise;
        blobManager.processBlobAttachMessage({
            ...blobAttachMessage,
            metadata: {
                localId: "localId",
                blobId: "blobId",
            },
        }, true);
        const blobHandle = await blobHandleP;
        // getting blob handle before attaching from the pending blob list
        node_assert_1.strict.strictEqual((0, client_utils_1.bufferToString)(await blobHandle.get(), "utf8"), "content");
        blobHandle.attachGraph();
        // getting blob handle after attaching from storage
        node_assert_1.strict.strictEqual((0, client_utils_1.bufferToString)(await blobHandle.get(), "utf8"), "contentFromStorage");
    });
    it("Reupload expired blob", async () => {
        const d = new internal_1.Deferred();
        let count = 0;
        const blobManager = createBlobManager({
            sendBlobAttachOp(_localId, _storageId) {
                d.resolve();
            },
            stashedBlobs: {},
            localBlobIdGenerator: () => "localId",
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    count++;
                    return { id: "blobId", minTTLInSeconds: count < 3 ? -1 : undefined };
                },
                readBlob: async () => {
                    return (0, client_utils_1.stringToBuffer)("content", "utf8");
                },
            }),
        });
        const blob = (0, client_utils_1.stringToBuffer)("content", "utf8");
        const blobHandleP = blobManager.createBlob(blob);
        await d.promise;
        node_assert_1.strict.strictEqual(count, 3, "test did not try to reupload");
        blobManager.processBlobAttachMessage({
            ...blobAttachMessage,
            metadata: {
                localId: "localId",
                blobId: "blobId",
            },
        }, true);
        const blobHandle = await blobHandleP;
        blobHandle.attachGraph();
        node_assert_1.strict.strictEqual((0, client_utils_1.bufferToString)(await blobHandle.get(), "utf8"), "content");
    });
});
//# sourceMappingURL=blobHandles.spec.js.map