"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.failProxy = void 0;
const node_assert_1 = require("node:assert");
const client_utils_1 = require("@fluid-internal/client-utils");
const test_pairwise_generator_1 = require("@fluid-private/test-pairwise-generator");
const container_definitions_1 = require("@fluidframework/container-definitions");
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const index_js_1 = require("../blobManager/index.js");
const containerHandleContext_js_1 = require("../containerHandleContext.js");
const failProxy = (handler = {}) => {
    const proxy = new Proxy(handler, {
        get: (t, p, r) => {
            if (p === "then") {
                return undefined;
            }
            if (handler !== undefined && p in handler) {
                return Reflect.get(t, p, r);
            }
            throw new Error(`${p.toString()} not implemented`);
        },
    });
    return proxy;
};
exports.failProxy = failProxy;
function createBlobManager(overrides) {
    const runtime = (0, exports.failProxy)({
        baseLogger: (0, internal_2.createChildLogger)(),
        attachState: container_definitions_1.AttachState.Attached,
        resolveHandle: async () => {
            throw new Error("not implemented");
        },
    });
    const routeContext = new containerHandleContext_js_1.ContainerFluidHandleContext("/", runtime, undefined);
    return new index_js_1.BlobManager((0, exports.failProxy)({
        // defaults, these can still be overridden below
        runtime,
        routeContext,
        blobManagerLoadInfo: {},
        stashedBlobs: undefined,
        localBlobIdGenerator: undefined,
        storage: (0, exports.failProxy)(),
        sendBlobAttachOp: () => { },
        blobRequested: () => { },
        isBlobDeleted: () => false,
        createBlobPayloadPending: false,
        // overrides
        ...overrides,
    }));
}
const olderThanTTL = {
    blob: "olderThanTTL",
    minTTLInSeconds: 100,
    uploadTime: Date.now() - 100 * 1000,
};
const withinTTLHalfLife = {
    blob: "withinTTLHalfLife",
    storageId: "withinTTLHalfLife",
    minTTLInSeconds: 100,
    uploadTime: Date.now() - 25 * 1000,
};
const storageIdWithoutUploadTime = {
    blob: "storageIdWithoutUploadTime",
    storageId: "storageIdWithoutUploadTime",
};
describe("BlobManager.stashed", () => {
    it("No Pending Stashed Uploads", async () => {
        const blobManager = createBlobManager();
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        await blobManager.stashedBlobsUploadP;
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
    });
    it("Stashed blob", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            sendBlobAttachOp(_localId, _storageId) { },
            stashedBlobs: {},
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        const blob = (0, client_utils_1.stringToBuffer)("content", "utf8");
        const sameBlobAsStashedP = blobManager.createBlob(blob);
        const pendingBlobsP = blobManager.attachAndGetPendingBlobs();
        const sameBlobAsStashed = await sameBlobAsStashedP;
        sameBlobAsStashed.attachGraph();
        const pendingBlobs = await pendingBlobsP;
        const blobManager2 = createBlobManager({
            stashedBlobs: pendingBlobs,
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        node_assert_1.strict.strictEqual(blobManager2.hasPendingStashedUploads(), true);
    });
    it("Process blob and complete stashed upload after", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            sendBlobAttachOp(_localId, _storageId) { },
            stashedBlobs: {},
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
            localBlobIdGenerator: () => "stubbed-local-id",
            isBlobDeleted: () => false,
            blobRequested: () => { },
        });
        const blob = (0, client_utils_1.stringToBuffer)("content", "utf8");
        const blobHandleP = blobManager.createBlob(blob);
        const pendingBlobsP = blobManager.attachAndGetPendingBlobs();
        const blobHandle = await blobHandleP;
        blobHandle.attachGraph();
        const pendingBlobs = await pendingBlobsP;
        const createResponse2 = new internal_1.Deferred();
        const blobManager2 = createBlobManager({
            stashedBlobs: pendingBlobs,
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse2.promise;
                },
            }),
            isBlobDeleted: () => false,
        });
        node_assert_1.strict.strictEqual(blobManager2.hasPendingStashedUploads(), true);
        blobManager2.processBlobAttachMessage({
            clientId: "client-id",
            minimumSequenceNumber: 1,
            referenceSequenceNumber: 1,
            sequenceNumber: 1,
            type: "blobAttach",
            metadata: {
                localId: "stubbed-local-id",
                blobId: "stubbed-storage-id",
            },
            timestamp: Date.now(),
        }, true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager2.stashedBlobsUploadP,
        ]);
        createResponse2.resolve({
            id: "new-storage-id",
        });
        await blobManager2.stashedBlobsUploadP;
    });
    it("Already stashed blob", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            stashedBlobs: {
                a: {
                    blob: "a",
                },
            },
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        createResponse.resolve({
            id: "a",
        });
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        node_assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
    it("Stashed blob with upload older than TTL", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            stashedBlobs: {
                olderThanTTL,
            },
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        createResponse.resolve({
            id: "a",
        });
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        node_assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
    it("Stashed blob with upload within TTL half-life", async () => {
        const blobManager = createBlobManager({
            stashedBlobs: {
                withinTTLHalfLife,
            },
        });
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        await blobManager.stashedBlobsUploadP;
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
    });
    it("Stashed blob without upload time", async () => {
        const createResponse = new internal_1.Deferred();
        const blobManager = createBlobManager({
            stashedBlobs: {
                storageIdWithoutUploadTime,
            },
            storage: (0, exports.failProxy)({
                createBlob: async () => {
                    return createResponse.promise;
                },
            }),
        });
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        createResponse.resolve({
            id: "a",
        });
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        node_assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
    it("Mixed stashed blobs", async () => {
        const letUploadsComplete = new internal_1.Deferred();
        const stashedBlobs = (0, test_pairwise_generator_1.generatePairwiseOptions)({
            acked: [undefined, true, false],
            blob: ["content"],
            minTTLInSeconds: [undefined, 100],
            storageId: [undefined, "id"],
            uploadTime: [Date.now() - 100 * 1000, Date.now() - 25 * 1000, undefined],
        }).reduce((pv, cv, i) => {
            if (cv.storageId) {
                cv.storageId += i.toString();
            }
            cv.blob = i.toString();
            pv[i] = cv;
            return pv;
        }, {});
        const blobManager = createBlobManager({
            stashedBlobs,
            storage: (0, exports.failProxy)({
                createBlob: async (b) => {
                    await letUploadsComplete.promise;
                    return { id: `id:${(0, client_utils_1.bufferToString)(b, "utf8")}` };
                },
            }),
        });
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), true);
        letUploadsComplete.resolve();
        await Promise.race([
            new Promise((resolve) => setTimeout(() => resolve(), 10)),
            blobManager.stashedBlobsUploadP,
        ]);
        node_assert_1.strict.strictEqual(blobManager.hasPendingStashedUploads(), false);
        node_assert_1.strict.strictEqual(blobManager.allBlobsAttached, true);
    });
});
//# sourceMappingURL=blobManager.stashed.spec.js.map