"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const node_assert_1 = require("node:assert");
const internal_1 = require("@fluidframework/runtime-definitions/internal");
const channelCollection_js_1 = require("../channelCollection.js");
const index_js_1 = require("../summary/index.js");
describe("Runtime", () => {
    describe("Container Runtime", () => {
        describe("getSummaryForDatastores", () => {
            const enabledMetadata = {
                summaryFormatVersion: 1,
                message: undefined,
            };
            const disabledMetadata = {
                summaryFormatVersion: 1,
                disableIsolatedChannels: true,
                message: undefined,
            };
            const emptyTree = (id) => ({
                id,
                blobs: {},
                trees: {},
            });
            const testSnapshot = {
                id: "root-id",
                blobs: {},
                trees: {
                    [internal_1.channelsTreeName]: {
                        id: "channels-id",
                        blobs: {},
                        trees: {
                            [index_js_1.nonDataStorePaths[0]]: emptyTree("lower-non-datastore-1"),
                            "some-datastore": emptyTree("lower-datastore-1"),
                            [index_js_1.nonDataStorePaths[1]]: emptyTree("lower-non-datastore-2"),
                            "another-datastore": emptyTree("lower-datastore-2"),
                        },
                    },
                    [index_js_1.nonDataStorePaths[0]]: emptyTree("top-non-datastore-1"),
                    "some-datastore": emptyTree("top-datastore-1"),
                    [index_js_1.nonDataStorePaths[1]]: emptyTree("top-non-datastore-2"),
                    "another-datastore": emptyTree("top-datastore-2"),
                },
            };
            it("Should return undefined for undefined snapshots", () => {
                let snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(undefined, undefined);
                (0, node_assert_1.strict)(snapshot === undefined);
                snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(undefined, enabledMetadata);
                (0, node_assert_1.strict)(snapshot === undefined);
                snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(undefined, disabledMetadata);
                (0, node_assert_1.strict)(snapshot === undefined);
                snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(undefined, undefined);
                (0, node_assert_1.strict)(snapshot === undefined);
                snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(undefined, enabledMetadata);
                (0, node_assert_1.strict)(snapshot === undefined);
                snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(undefined, disabledMetadata);
                (0, node_assert_1.strict)(snapshot === undefined);
            });
            it("Should strip out non-datastore paths for versions < 1", () => {
                const snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(testSnapshot, undefined);
                (0, node_assert_1.strict)(snapshot, "Snapshot should be defined");
                node_assert_1.strict.strictEqual(snapshot.id, "root-id", "Should be top-level");
                node_assert_1.strict.strictEqual(Object.keys(snapshot.trees).length, 3, "Should have 3 datastores");
                node_assert_1.strict.strictEqual(snapshot.trees[internal_1.channelsTreeName]?.id, "channels-id", "Should have channels tree as datastore");
                node_assert_1.strict.strictEqual(snapshot.trees["some-datastore"]?.id, "top-datastore-1", "Should have top datastore 1");
                node_assert_1.strict.strictEqual(snapshot.trees["another-datastore"]?.id, "top-datastore-2", "Should have top datastore 2");
            });
            it("Should strip out non-datastore paths for disabled isolated channels", () => {
                const snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(testSnapshot, disabledMetadata);
                (0, node_assert_1.strict)(snapshot, "Snapshot should be defined");
                node_assert_1.strict.strictEqual(snapshot.id, "root-id", "Should be top-level");
                node_assert_1.strict.strictEqual(Object.keys(snapshot.trees).length, 3, "Should have 3 datastores");
                node_assert_1.strict.strictEqual(snapshot.trees[internal_1.channelsTreeName]?.id, "channels-id", "Should have channels tree as datastore");
                node_assert_1.strict.strictEqual(snapshot.trees["some-datastore"]?.id, "top-datastore-1", "Should have top datastore 1");
                node_assert_1.strict.strictEqual(snapshot.trees["another-datastore"]?.id, "top-datastore-2", "Should have top datastore 2");
            });
            it("Should give channels subtree for version 1", () => {
                const snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(testSnapshot, enabledMetadata);
                (0, node_assert_1.strict)(snapshot, "Snapshot should be defined");
                node_assert_1.strict.strictEqual(snapshot.id, "channels-id", "Should be lower-level");
                node_assert_1.strict.strictEqual(Object.keys(snapshot.trees).length, 4, "Should have 4 datastores");
                // Put in variable to avoid type-narrowing bug
                const nonDataStore1 = snapshot.trees[index_js_1.nonDataStorePaths[0]];
                node_assert_1.strict.strictEqual(nonDataStore1?.id, "lower-non-datastore-1", "Should have lower non-datastore 1");
                node_assert_1.strict.strictEqual(snapshot.trees[index_js_1.nonDataStorePaths[1]]?.id, "lower-non-datastore-2", "Should have lower non-datastore 2");
                node_assert_1.strict.strictEqual(snapshot.trees["some-datastore"]?.id, "lower-datastore-1", "Should have lower datastore 1");
                node_assert_1.strict.strictEqual(snapshot.trees["another-datastore"]?.id, "lower-datastore-2", "Should have lower datastore 2");
            });
        });
        describe("detectOutboundReferences", () => {
            it("Can find handles", () => {
                const outboundReferences = [];
                (0, channelCollection_js_1.detectOutboundReferences)("dataStore1", {
                    address: "dds1",
                    someHandle: {
                        type: "__fluid_handle__",
                        url: "routeA",
                    },
                    nested: {
                        anotherHandle: {
                            type: "__fluid_handle__",
                            url: "routeB",
                        },
                        address: "ignored",
                    },
                    array: [
                        {
                            type: "__fluid_handle__",
                            url: "routeC",
                        },
                        {
                            type: "__fluid_handle__",
                            url: "routeD",
                        },
                    ],
                    // eslint-disable-next-line unicorn/no-null
                    deadEnd: null,
                    number: 1,
                    nothing: undefined,
                }, (from, to) => {
                    outboundReferences.push([from, to]);
                });
                node_assert_1.strict.deepEqual(outboundReferences, [
                    ["/dataStore1/dds1", "routeA"],
                    ["/dataStore1/dds1", "routeB"],
                    ["/dataStore1/dds1", "routeC"],
                    ["/dataStore1/dds1", "routeD"],
                ], "Should find both handles");
            });
            it("null contents", () => {
                // eslint-disable-next-line unicorn/no-null
                (0, channelCollection_js_1.detectOutboundReferences)("foo", null, () => {
                    node_assert_1.strict.fail("Should not be called");
                });
            });
        });
    });
});
//# sourceMappingURL=channelCollection.spec.js.map