"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable @typescript-eslint/consistent-type-assertions */
const node_assert_1 = __importDefault(require("node:assert"));
const internal_1 = require("@fluidframework/container-definitions/internal");
const internal_2 = require("@fluidframework/driver-definitions/internal");
const internal_3 = require("@fluidframework/telemetry-utils/internal");
const messageTypes_js_1 = require("../messageTypes.js");
const index_js_1 = require("../opLifecycle/index.js");
const pendingStateManager_js_1 = require("../pendingStateManager.js");
// Make a mock op with distinguishable contents
function op(data = "") {
    return {
        type: messageTypes_js_1.ContainerMessageType.FluidDataStoreOp,
        contents: data,
    };
}
describe("Pending State Manager", () => {
    const mockLogger = new internal_3.MockLogger();
    const logger = (0, internal_3.createChildLogger)({ logger: mockLogger });
    afterEach("ThrowOnErrorLogs", () => {
        // Note: If mockLogger is used within a test,
        // it may inadvertently clear errors such that they're not noticed here
        mockLogger.assertNoErrors();
        mockLogger.clear();
    });
    describe("Rollback", () => {
        let rollbackCalled;
        let rollbackContent;
        let rollbackShouldThrow;
        let batchManager;
        function getMessage(payload) {
            return {
                runtimeOp: op(payload),
            };
        }
        const rollBackCallback = (m) => {
            rollbackCalled = true;
            rollbackContent.push(m);
            if (rollbackShouldThrow) {
                throw new Error();
            }
        };
        beforeEach(async () => {
            rollbackCalled = false;
            rollbackContent = [];
            rollbackShouldThrow = false;
            batchManager = new index_js_1.BatchManager({ canRebase: true });
        });
        it("should do nothing when rolling back empty pending stack", () => {
            const checkpoint = batchManager.checkpoint();
            checkpoint.rollback(rollBackCallback);
            node_assert_1.default.strictEqual(rollbackCalled, false);
            node_assert_1.default.strictEqual(batchManager.empty, true);
        });
        it("should do nothing when rolling back nothing", () => {
            batchManager.push(getMessage("1"), /* reentrant */ false);
            const checkpoint = batchManager.checkpoint();
            checkpoint.rollback(rollBackCallback);
            node_assert_1.default.strictEqual(rollbackCalled, false);
            node_assert_1.default.strictEqual(batchManager.empty, false);
        });
        it("should succeed when rolling back entire pending stack", () => {
            const checkpoint = batchManager.checkpoint();
            batchManager.push(getMessage("11"), /* reentrant */ false);
            batchManager.push(getMessage("22"), /* reentrant */ false);
            batchManager.push(getMessage("33"), /* reentrant */ false);
            checkpoint.rollback(rollBackCallback);
            node_assert_1.default.strictEqual(rollbackCalled, true);
            node_assert_1.default.strictEqual(rollbackContent.length, 3);
            node_assert_1.default.strictEqual(rollbackContent[0].runtimeOp.contents, "33");
            node_assert_1.default.strictEqual(rollbackContent[1].runtimeOp.contents, "22");
            node_assert_1.default.strictEqual(rollbackContent[2].runtimeOp.contents, "11");
            node_assert_1.default.strictEqual(batchManager.empty, true);
        });
        it("should succeed when rolling back part of pending stack", () => {
            batchManager.push(getMessage("11"), /* reentrant */ false);
            const checkpoint = batchManager.checkpoint();
            batchManager.push(getMessage("22"), /* reentrant */ false);
            batchManager.push(getMessage("33"), /* reentrant */ false);
            checkpoint.rollback(rollBackCallback);
            node_assert_1.default.strictEqual(rollbackCalled, true);
            node_assert_1.default.strictEqual(rollbackContent.length, 2);
            node_assert_1.default.strictEqual(rollbackContent[0].runtimeOp.contents, "33");
            node_assert_1.default.strictEqual(rollbackContent[1].runtimeOp.contents, "22");
            node_assert_1.default.strictEqual(batchManager.empty, false);
        });
        it("should throw and close when rollback fails", () => {
            rollbackShouldThrow = true;
            const checkpoint = batchManager.checkpoint();
            batchManager.push(getMessage("11"), /* reentrant */ false);
            node_assert_1.default.throws(() => {
                checkpoint.rollback(rollBackCallback);
            });
            node_assert_1.default.strictEqual(rollbackCalled, true);
        });
    });
    describe("Op processing", () => {
        let pendingStateManager;
        const clientId = "clientId";
        beforeEach(async () => {
            pendingStateManager = new pendingStateManager_js_1.PendingStateManager({
                applyStashedOp: () => {
                    throw new Error();
                },
                clientId: () => "oldClientId",
                connected: () => true,
                reSubmitBatch: () => { },
                isActiveConnection: () => false,
                isAttached: () => true,
            }, undefined /* initialLocalState */, logger);
        });
        const submitBatch = (messages, clientSequenceNumber, localOpMetadata) => {
            pendingStateManager.onFlushBatch(messages.map((message) => ({
                runtimeOp: {
                    type: message.type,
                    contents: message.contents,
                },
                referenceSequenceNumber: message.referenceSequenceNumber, // eslint-disable-line @typescript-eslint/no-non-null-assertion
                metadata: message.metadata,
                localOpMetadata,
            })), clientSequenceNumber ?? messages[0]?.clientSequenceNumber, false /* staged */);
        };
        const processFullBatch = (messages, batchStartCsn, groupedBatch, emptyBatchSequenceNumber, resubmittedBatchId) => pendingStateManager.processInboundMessages({
            type: "fullBatch",
            messages: messages,
            batchStart: {
                batchStartCsn,
                keyMessage: {
                    sequenceNumber: emptyBatchSequenceNumber,
                },
                clientId,
                batchId: resubmittedBatchId,
            },
            length: messages.length,
            groupedBatch,
        }, true /* local */);
        it("Grouped batch is processed correctly", () => {
            const messages = [
                {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 0,
                    referenceSequenceNumber: 0,
                    metadata: { batch: true },
                },
                {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 1,
                    referenceSequenceNumber: 0,
                },
                {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    metadata: { batch: false },
                    clientSequenceNumber: 2,
                    referenceSequenceNumber: 0,
                },
            ];
            submitBatch(messages);
            processFullBatch(messages, 0 /* batchStartCsn */, true /* groupedBatch */);
        });
        it("Ungrouped batch is processed correctly", () => {
            const messages = [
                {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 0,
                    referenceSequenceNumber: 0,
                    metadata: { batch: true },
                },
                {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 1,
                    referenceSequenceNumber: 0,
                },
                {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    metadata: { batch: false },
                    clientSequenceNumber: 2,
                    referenceSequenceNumber: 0,
                },
            ];
            submitBatch(messages);
            pendingStateManager.processInboundMessages({
                type: "batchStartingMessage",
                nextMessage: messages[0],
                batchStart: {
                    batchStartCsn: 0,
                    keyMessage: messages[0],
                    clientId,
                    batchId: undefined,
                },
            }, true /* local */);
            pendingStateManager.processInboundMessages({
                type: "nextBatchMessage",
                nextMessage: messages[1],
            }, true /* local */);
            pendingStateManager.processInboundMessages({
                type: "nextBatchMessage",
                nextMessage: messages[2],
                batchEnd: true,
            }, true /* local */);
        });
        it("empty batch is processed correctly", () => {
            // Empty batch is reflected in the pending state manager as a single message
            // with the following metadata:
            submitBatch([
                {
                    contents: JSON.stringify({ type: "groupedBatch", contents: [] }),
                    referenceSequenceNumber: 0,
                    metadata: { batchId: "batchId" },
                },
            ], 1 /* clientSequenceNumber */, { emptyBatch: true });
            // A groupedBatch is supposed to have nested messages inside its contents,
            // but an empty batch has no nested messages. When processing en empty grouped batch,
            // the psm will expect the next pending message to be an "empty" message as portrayed above.
            processFullBatch([], 1 /* batchStartCsn */, true /* groupedBatch */, 3 /* emptyBatchSequenceNumber */, "batchId" /* resubmittedBatchId */);
        });
        describe("processing out of sync messages will throw and log", () => {
            it("messageTypes do not match", () => {
                const messages = [
                    {
                        clientId,
                        type: internal_2.MessageType.Operation,
                        clientSequenceNumber: 0,
                        referenceSequenceNumber: 0,
                    },
                ];
                submitBatch(messages);
                node_assert_1.default.throws(() => processFullBatch(messages.map((message) => ({
                    ...message,
                    type: "otherType",
                })), 0 /* batchStartCsn */, false /* groupedBatch */), (closeError) => closeError.errorType === internal_1.ContainerErrorTypes.dataProcessingError);
                mockLogger.assertMatch([
                    {
                        eventName: "unexpectedAckReceived",
                        pendingContentScrubbed: JSON.stringify({ type: "op" }),
                        incomingContentScrubbed: JSON.stringify({ type: "otherType" }),
                        contentsMatch: true,
                        pendingLength: 13,
                        incomingLength: 20,
                        mismatchStartIndex: 10,
                        pendingChar: "p",
                        incomingChar: "t",
                    },
                ], "Expected to log scrubbed messages", true /* inlineDetailsProp */);
            });
            it("only one message has undefined content", () => {
                const messages = [
                    {
                        clientId,
                        type: internal_2.MessageType.Operation,
                        clientSequenceNumber: 0,
                        referenceSequenceNumber: 0,
                        contents: {},
                    },
                ];
                submitBatch(messages);
                node_assert_1.default.throws(() => processFullBatch(messages.map((message) => ({
                    ...message,
                    contents: undefined,
                })), 0 /* batchStartCsn */, false /* groupedBatch */), (closeError) => closeError.errorType === internal_1.ContainerErrorTypes.dataProcessingError);
                mockLogger.assertMatch([
                    {
                        eventName: "unexpectedAckReceived",
                        pendingContentScrubbed: JSON.stringify({ type: "op", contents: {} }),
                        incomingContentScrubbed: JSON.stringify({ type: "op" }),
                        contentsMatch: false,
                        pendingLength: 27,
                        incomingLength: 13,
                        mismatchStartIndex: 12,
                        pendingChar: ",",
                        incomingChar: "}",
                    },
                ], "Expected to log scrubbed messages", true /* inlineDetailsProp */);
            });
            it("stringified message content does not match", () => {
                const messages = [
                    {
                        clientId,
                        type: internal_2.MessageType.Operation,
                        clientSequenceNumber: 0,
                        referenceSequenceNumber: 0,
                        contents: {},
                    },
                ];
                submitBatch(messages);
                node_assert_1.default.throws(() => processFullBatch(messages.map((message) => ({
                    ...message,
                    contents: { prop1: true },
                })), 0 /* batchStartCsn */, false /* groupedBatch */), (closeError) => closeError.errorType === internal_1.ContainerErrorTypes.dataProcessingError);
                mockLogger.assertMatch([
                    {
                        eventName: "unexpectedAckReceived",
                        pendingContentScrubbed: JSON.stringify({ type: "op", contents: {} }),
                        incomingContentScrubbed: JSON.stringify({
                            type: "op",
                            contents: { prop1: "boolean" },
                        }),
                        contentsMatch: false,
                        pendingLength: 27,
                        incomingLength: 39,
                        mismatchStartIndex: 25,
                        pendingChar: "}",
                        incomingChar: '"',
                    },
                ], "Expected to log scrubbed messages", true /* inlineDetailsProp */);
            });
            it("stringified message content out of order", () => {
                const message = {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 0,
                    referenceSequenceNumber: 0,
                    contents: {},
                };
                // contents and type are swapped in the stringified message relative to what we typically do/expect
                pendingStateManager.onFlushBatch([
                    {
                        runtimeOp: {
                            contents: message.contents,
                            type: message.type,
                        },
                        referenceSequenceNumber: 0,
                    },
                ], 0 /* clientSequenceNumber */, false /* staged */);
                node_assert_1.default.throws(() => processFullBatch([message], 0 /* batchStartCsn */, false /* groupedBatch */), (closeError) => closeError.errorType === internal_1.ContainerErrorTypes.dataProcessingError);
                mockLogger.assertMatch([
                    {
                        eventName: "unexpectedAckReceived",
                        pendingContentScrubbed: JSON.stringify({ contents: {}, type: "op" }),
                        incomingContentScrubbed: JSON.stringify({ type: "op", contents: {} }),
                        contentsMatch: true,
                        pendingLength: 27,
                        incomingLength: 27,
                        mismatchStartIndex: 2,
                        pendingChar: "c",
                        incomingChar: "t",
                    },
                ], "Expected to log scrubbed messages", true /* inlineDetailsProp */);
            });
            it("stringified message content with unexpected keys", () => {
                const message = {
                    clientId,
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 0,
                    referenceSequenceNumber: 0,
                    contents: {},
                };
                // contents and type are swapped in the stringified message relative to what we typically do/expect
                pendingStateManager.onFlushBatch([
                    {
                        runtimeOp: {
                            type: message.type,
                            contents: message.contents,
                            somethingElse: 123, // Unexpected key
                        },
                        referenceSequenceNumber: 0,
                    },
                ], 0 /* clientSequenceNumber */, false /* staged */);
                node_assert_1.default.throws(() => processFullBatch([message], 0 /* batchStartCsn */, false /* groupedBatch */), (closeError) => closeError.errorType === internal_1.ContainerErrorTypes.dataProcessingError);
                mockLogger.assertMatch([
                    {
                        eventName: "unexpectedAckReceived",
                        pendingContentScrubbed: JSON.stringify({
                            type: "op",
                            contents: {},
                            somethingElse: "number",
                        }),
                        incomingContentScrubbed: JSON.stringify({ type: "op", contents: {} }),
                        contentsMatch: true,
                        pendingLength: 47,
                        incomingLength: 27,
                        mismatchStartIndex: 26,
                        pendingChar: ",",
                        incomingChar: "}",
                    },
                ], "Expected to log scrubbed messages", true /* inlineDetailsProp */);
            });
            it("processing in sync messages will not throw", () => {
                const messages = [
                    {
                        clientId,
                        type: internal_2.MessageType.Operation,
                        clientSequenceNumber: 0,
                        referenceSequenceNumber: 0,
                        contents: { prop1: true },
                    },
                ];
                submitBatch(messages);
                processFullBatch(messages.map((message) => ({
                    ...message,
                    contents: { prop1: true },
                })), 0 /* batchStartCsn */, false /* groupedBatch */);
            });
            it("findFirstCharacterMismatched", () => {
                const nonAsciiChar = String.fromCodePoint(0x80);
                const testCases = [
                    { input: ["", ""], expected: [-1] },
                    { input: ["", "b"], expected: [0, undefined, "b"] },
                    { input: ["a", "b"], expected: [0, "a", "b"] },
                    { input: ["xyz", "xxx"], expected: [1, "y", "x"] },
                    { input: ["xyz", "xy"], expected: [2, "z", undefined] },
                    { input: ["xy", "xxx"], expected: [1, "y", "x"] },
                    { input: ["xyz", "xyz"], expected: [-1] },
                    { input: ["QQ", `Q${nonAsciiChar}`], expected: [1, "Q", "[non-ASCII]"] },
                ];
                for (const { input: [a, b], expected: [i, charA, charB], } of testCases) {
                    node_assert_1.default.deepEqual((0, pendingStateManager_js_1.findFirstCharacterMismatched)(a, b), [i, charA, charB], `Failed input: "${a}", "${b}"`);
                    node_assert_1.default.deepEqual((0, pendingStateManager_js_1.findFirstCharacterMismatched)(b, a), [i, charB, charA], `Failed input: "${b}", "${a}"`);
                }
            });
        });
        describe("getLocalState", () => {
            it("removes ops with seq num lower than snapshot", () => {
                const messages = Array.from({ length: 10 }, (_, i) => ({
                    clientId: "clientId",
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 0,
                    contents: { prop1: true },
                    sequenceNumber: i + 1, // starting with sequence number 1 so first assert does not filter any op
                }));
                submitBatch(messages);
                processFullBatch(messages, 0 /* batchStartCsn */, false /* groupedBatch */);
                let pendingState = pendingStateManager.getLocalState(0).pendingStates;
                node_assert_1.default.strictEqual(pendingState.length, 10);
                pendingState = pendingStateManager.getLocalState(5).pendingStates;
                node_assert_1.default.strictEqual(pendingState.length, 5);
                pendingState = pendingStateManager.getLocalState(10).pendingStates;
                node_assert_1.default.strictEqual(pendingState.length, 0);
            });
            it("throws when trying to get unprocessed ops older than snapshot", () => {
                const messages = Array.from({ length: 10 }, (_, i) => ({
                    clientId: "clientId",
                    type: internal_2.MessageType.Operation,
                    clientSequenceNumber: 0,
                    contents: { prop1: true },
                    referenceSequenceNumber: i,
                }));
                submitBatch(messages);
                node_assert_1.default.throws(() => pendingStateManager.getLocalState(1));
                const pendingState = pendingStateManager.getLocalState(0).pendingStates;
                node_assert_1.default.strictEqual(pendingState.length, 10);
            });
        });
    });
    describe("Local state processing", () => {
        function createPendingStateManager(pendingStates) {
            return new pendingStateManager_js_1.PendingStateManager({
                applyStashedOp: async () => undefined,
                clientId: () => "CLIENT_ID",
                connected: () => true,
                reSubmitBatch: () => { },
                isActiveConnection: () => false,
                isAttached: () => true,
            }, pendingStates ? { pendingStates } : undefined, logger);
        }
        describe("Constructor pendingStates", () => {
            it("Empty local state", () => {
                {
                    const pendingStateManager = createPendingStateManager(undefined);
                    node_assert_1.default.deepStrictEqual(pendingStateManager.initialMessages.toArray(), []);
                }
                {
                    const pendingStateManager = createPendingStateManager([]);
                    node_assert_1.default.deepStrictEqual(pendingStateManager.initialMessages.toArray(), []);
                }
            });
            it("New format", () => {
                const messages = [
                    {
                        type: "message",
                        content: '{"type":"component"}',
                        referenceSequenceNumber: 10,
                    },
                    {
                        type: "message",
                        content: '{"type": "component", "contents": {"prop1": "value"}}',
                        referenceSequenceNumber: 10,
                    },
                ];
                const pendingStateManager = createPendingStateManager(messages);
                node_assert_1.default.deepStrictEqual(pendingStateManager.initialMessages.toArray(), messages);
            });
        });
    });
    describe("replayPendingStates", () => {
        let pendingStateManager;
        const resubmittedBatchIds = [];
        const clientId = "clientId";
        beforeEach(async () => {
            resubmittedBatchIds.length = 0;
            pendingStateManager = new pendingStateManager_js_1.PendingStateManager({
                applyStashedOp: async () => undefined,
                clientId: () => clientId,
                connected: () => true,
                reSubmitBatch: (batch, { batchId }) => {
                    resubmittedBatchIds.push(batchId);
                },
                isActiveConnection: () => false,
                isAttached: () => true,
            }, undefined /* initialLocalState */, logger);
        });
        it("replays pending states", () => {
            const messages = [
                {
                    type: messageTypes_js_1.ContainerMessageType.Rejoin,
                    clientSequenceNumber: 0,
                    referenceSequenceNumber: 0,
                    contents: undefined,
                },
                {
                    type: messageTypes_js_1.ContainerMessageType.Rejoin,
                    clientSequenceNumber: 1,
                    referenceSequenceNumber: 0,
                    contents: undefined,
                },
            ];
            pendingStateManager.onFlushBatch(messages.map((message) => ({
                runtimeOp: {
                    type: message.type,
                    contents: message.contents,
                },
                referenceSequenceNumber: message.referenceSequenceNumber,
            })), 0, false /* staged */);
            pendingStateManager.replayPendingStates();
            node_assert_1.default.strictEqual(resubmittedBatchIds[0], `${clientId}_[0]`);
            node_assert_1.default.strictEqual(resubmittedBatchIds[1], `${clientId}_[0]`);
        });
        it("replays pending states with empty batch", () => {
            pendingStateManager.onFlushEmptyBatch({
                localOpMetadata: { emptyBatch: true },
                referenceSequenceNumber: 0,
                metadata: { batchId: "batchId" },
            }, 0, false /* staged */);
            pendingStateManager.replayPendingStates();
            node_assert_1.default.strictEqual(resubmittedBatchIds[0], "batchId");
        });
    });
    describe("applyStashedOpsAt", () => {
        it("applyStashedOpsAt", async () => {
            const applyStashedOps = [];
            const messages = [
                {
                    type: "message",
                    content: '{"type":"component"}',
                    referenceSequenceNumber: 10,
                    localOpMetadata: undefined,
                    opMetadata: undefined,
                    batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 1, length: 1, staged: false },
                },
                {
                    type: "message",
                    content: '{"type": "component", "contents": {"prop1": "value"}}',
                    referenceSequenceNumber: 11,
                    localOpMetadata: undefined,
                    opMetadata: undefined,
                    batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 2, length: 1, staged: false },
                },
            ];
            const pendingStateManager = new pendingStateManager_js_1.PendingStateManager({
                applyStashedOp: async (content) => applyStashedOps.push(content),
                clientId: () => "clientId",
                connected: () => true,
                reSubmitBatch: () => { },
                isActiveConnection: () => false,
                isAttached: () => true,
            }, { pendingStates: messages }, logger);
            await pendingStateManager.applyStashedOpsAt();
            node_assert_1.default.strictEqual(applyStashedOps.length, 2);
            node_assert_1.default.strictEqual(pendingStateManager.pendingMessagesCount, 2);
        });
        it("applyStashedOpsAt for empty batch", async () => {
            const applyStashedOps = [];
            const messages = [
                {
                    type: "message",
                    content: '{"type":"groupedBatch", "contents": []}',
                    referenceSequenceNumber: 10,
                    opMetadata: undefined,
                    localOpMetadata: { emptyBatch: true },
                    batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 1, length: 1, staged: false },
                },
            ];
            const pendingStateManager = new pendingStateManager_js_1.PendingStateManager({
                applyStashedOp: async (content) => applyStashedOps.push(content),
                clientId: () => "clientId",
                connected: () => true,
                reSubmitBatch: () => { },
                isActiveConnection: () => false,
                isAttached: () => true,
            }, { pendingStates: messages }, logger);
            await pendingStateManager.applyStashedOpsAt();
            node_assert_1.default.strictEqual(applyStashedOps.length, 0);
            node_assert_1.default.strictEqual(pendingStateManager.pendingMessagesCount, 1);
        });
    });
    describe("Pending messages state", () => {
        const forInitialMessages = [
            { type: "message", content: '{"type":"component"}', referenceSequenceNumber: 10 },
            {
                type: "message",
                content: '{"type": "component", "contents": {"prop1": "value"}}',
                referenceSequenceNumber: 10,
            },
        ];
        const forFlushedMessages = [
            {
                runtimeOp: { type: "component" },
                referenceSequenceNumber: 10,
            },
            {
                runtimeOp: { type: "component" },
                referenceSequenceNumber: 10,
            },
        ];
        function createPendingStateManager(pendingStates) {
            return new pendingStateManager_js_1.PendingStateManager({
                applyStashedOp: async () => undefined,
                clientId: () => "CLIENT_ID",
                connected: () => true,
                reSubmitBatch: () => { },
                isActiveConnection: () => false,
                isAttached: () => true,
            }, pendingStates ? { pendingStates } : undefined, logger);
        }
        it("no pending or initial messages", () => {
            const pendingStateManager = createPendingStateManager(undefined);
            node_assert_1.default.strictEqual(pendingStateManager.hasPendingMessages(), false, "There shouldn't be pending messages");
            node_assert_1.default.strictEqual(pendingStateManager.pendingMessagesCount, 0, "Pending messages count should be 0");
        });
        it("has pending messages but no initial messages", () => {
            const pendingStateManager = createPendingStateManager(undefined);
            // let each message be its own batch
            for (const message of forFlushedMessages) {
                pendingStateManager.onFlushBatch([
                    {
                        runtimeOp: message.runtimeOp,
                        referenceSequenceNumber: message.referenceSequenceNumber,
                    },
                ], 0, false /* staged */);
            }
            node_assert_1.default.strictEqual(pendingStateManager.hasPendingMessages(), true, "There should be pending messages");
            node_assert_1.default.strictEqual(pendingStateManager.pendingMessagesCount, forFlushedMessages.length, "Pending messages count should be same as pending messages");
        });
        it("has initial messages but no pending messages", () => {
            const pendingStateManager = createPendingStateManager(forInitialMessages);
            node_assert_1.default.strictEqual(pendingStateManager.hasPendingMessages(), true, "There should be initial messages");
            node_assert_1.default.strictEqual(pendingStateManager.pendingMessagesCount, forFlushedMessages.length, "Pending messages count should be same as initial messages");
        });
        it("has both pending messages and initial messages", () => {
            const pendingStateManager = createPendingStateManager(forInitialMessages);
            // let each message be its own batch
            for (const message of forFlushedMessages) {
                pendingStateManager.onFlushBatch([
                    {
                        runtimeOp: message.runtimeOp,
                        referenceSequenceNumber: message.referenceSequenceNumber,
                    },
                ], 0, false /* staged */);
            }
            node_assert_1.default.strictEqual(pendingStateManager.hasPendingMessages(), true, "There should be pending messages");
            node_assert_1.default.strictEqual(pendingStateManager.pendingMessagesCount, forInitialMessages.length + forFlushedMessages.length, "Pending messages count should be same as pending + initial messages");
        });
    });
    describe("Minimum sequence number", () => {
        const forInitialMessages = [
            {
                type: "message",
                content: '{"type":"component"}',
                referenceSequenceNumber: 10,
                localOpMetadata: undefined,
                opMetadata: undefined,
                batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 1, length: 1, staged: false },
            },
            {
                type: "message",
                content: '{"type": "component", "contents": {"prop1": "value"}}',
                referenceSequenceNumber: 11,
                localOpMetadata: undefined,
                opMetadata: undefined,
                batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 2, length: 1, staged: false },
            },
            {
                type: "message",
                content: '{"type": "component", "contents": {"prop2": "value"}}',
                referenceSequenceNumber: 12,
                localOpMetadata: undefined,
                opMetadata: undefined,
                batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 3, length: 1, staged: false },
            },
            {
                type: "message",
                content: '{"type": "component", "contents": {"prop3": "value"}}',
                referenceSequenceNumber: 12,
                localOpMetadata: undefined,
                opMetadata: undefined,
                batchInfo: { clientId: "CLIENT_ID", batchStartCsn: 4, length: 1, staged: false },
            },
        ];
        const forFlushedMessages = forInitialMessages.map((message) => ({
            runtimeOp: JSON.parse(message.content),
            referenceSequenceNumber: message.referenceSequenceNumber,
        }));
        function createPendingStateManager(pendingStates) {
            return new pendingStateManager_js_1.PendingStateManager({
                applyStashedOp: async () => undefined,
                clientId: () => "123",
                connected: () => true,
                reSubmitBatch: () => { },
                isActiveConnection: () => false,
                isAttached: () => true,
            }, pendingStates ? { pendingStates } : undefined /* initialLocalState */, logger);
        }
        it("minimum sequence number can be retrieved from initial messages", async () => {
            const pendingStateManager = createPendingStateManager(forInitialMessages);
            await pendingStateManager.applyStashedOpsAt();
            node_assert_1.default.strictEqual(pendingStateManager.minimumPendingMessageSequenceNumber, 10, "minimum sequence number should be the first message");
            pendingStateManager.replayPendingStates();
            node_assert_1.default.strictEqual(pendingStateManager.minimumPendingMessageSequenceNumber, undefined, "Should have processed messages and thus have no min seq number");
        });
        it("minimum sequence number can be retrieved from pending messages", async () => {
            const pendingStateManager = createPendingStateManager();
            node_assert_1.default.strictEqual(pendingStateManager.minimumPendingMessageSequenceNumber, undefined, "No pending messages should mean no minimum seq number");
            // Each message has a different reference sequence number so let them each be their own batch
            for (const message of forFlushedMessages) {
                pendingStateManager.onFlushBatch([
                    {
                        runtimeOp: message.runtimeOp,
                        referenceSequenceNumber: message.referenceSequenceNumber,
                    },
                ], 0, false /* staged */);
            }
            node_assert_1.default.strictEqual(pendingStateManager.minimumPendingMessageSequenceNumber, 10, "has pending messages and thus a minimum seq number");
            pendingStateManager.replayPendingStates();
            node_assert_1.default.strictEqual(pendingStateManager.minimumPendingMessageSequenceNumber, undefined, "Should no minimum sequence number as there are no messages");
        });
    });
});
//# sourceMappingURL=pendingStateManager.spec.js.map