"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const node_assert_1 = require("node:assert");
const internal_1 = require("@fluidframework/container-definitions/internal");
const internal_2 = require("@fluidframework/core-interfaces/internal");
const internal_3 = require("@fluidframework/telemetry-utils/internal");
const internal_4 = require("@fluidframework/test-runtime-utils/internal");
const sinon_1 = __importDefault(require("sinon"));
const containerRuntime_js_1 = require("../containerRuntime.js");
const packageVersion_js_1 = require("../packageVersion.js");
const runtimeLayerCompatState_js_1 = require("../runtimeLayerCompatState.js");
const dataStoreCreationHelper_js_1 = require("./dataStoreCreationHelper.js");
function validateFailureProperties(error, isGenerationCompatible, layerGeneration, layerType, unsupportedFeatures) {
    (0, node_assert_1.strict)(error instanceof internal_3.UsageError, "The error should be a UsageError");
    node_assert_1.strict.strictEqual(error.errorType, internal_2.FluidErrorTypes.usageError, "Error type should be usageError");
    const telemetryProps = error.getTelemetryProperties();
    (0, node_assert_1.strict)(typeof telemetryProps.errorDetails === "string", "Error details should be present");
    const properties = JSON.parse(telemetryProps.errorDetails);
    node_assert_1.strict.strictEqual(properties.isGenerationCompatible, isGenerationCompatible, "Generation compatibility not as expected");
    node_assert_1.strict.strictEqual(properties.runtimeVersion, packageVersion_js_1.pkgVersion, "Runtime version not as expected");
    node_assert_1.strict.strictEqual(properties.runtimeGeneration, runtimeLayerCompatState_js_1.runtimeCompatDetailsForLoader.generation, "Runtime generation not as expected");
    node_assert_1.strict.deepStrictEqual(properties.unsupportedFeatures, unsupportedFeatures, "Unsupported features not as expected");
    if (layerType === "Loader") {
        node_assert_1.strict.strictEqual(properties.loaderVersion, packageVersion_js_1.pkgVersion, "Loader version not as expected");
        node_assert_1.strict.strictEqual(properties.loaderGeneration, layerGeneration, "Loader generation not as expected");
    }
    else {
        node_assert_1.strict.strictEqual(properties.dataStoreVersion, packageVersion_js_1.pkgVersion, "DataStore version not as expected");
        node_assert_1.strict.strictEqual(properties.dataStoreGeneration, layerGeneration, "DataStore generation not as expected");
    }
    return true;
}
describe("Runtime Layer compatibility", () => {
    describe("Runtime <-> Loader compatibility", () => {
        let originalRequiredFeatures;
        beforeEach(() => {
            originalRequiredFeatures = runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures;
        });
        afterEach(() => {
            runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures =
                [...originalRequiredFeatures];
        });
        /**
         * These tests validates that the layer compat state and validateLoaderCompatibility function correctly
         * validates the compatibility between Runtime and Loader layers.
         */
        describe("validateLoaderCompatibility", () => {
            it("Runtime is compatible with old Loader (pre-enforcement)", () => {
                // Older Loader will not have ILayerCompatDetails defined.
                node_assert_1.strict.doesNotThrow(() => (0, runtimeLayerCompatState_js_1.validateLoaderCompatibility)(undefined /* maybeloaderCompatDetailsForRuntime */, () => {
                    throw new Error("should not dispose");
                }), "Runtime should be compatible with older Loader");
            });
            it("Runtime generation and features are compatible with Loader", () => {
                runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures = ["feature1", "feature2"];
                const loaderCompatDetailsForRuntime = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: runtimeLayerCompatState_js_1.loaderSupportRequirements.minSupportedGeneration,
                    supportedFeatures: new Set(runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures),
                };
                node_assert_1.strict.doesNotThrow(() => (0, runtimeLayerCompatState_js_1.validateLoaderCompatibility)(loaderCompatDetailsForRuntime, () => {
                    throw new Error("should not dispose");
                }), "Runtime should be compatible with Loader layer");
            });
            it("Runtime generation is incompatible with Loader", () => {
                const disposeFn = sinon_1.default.fake();
                runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures = ["feature1", "feature2"];
                const loaderGeneration = runtimeLayerCompatState_js_1.loaderSupportRequirements.minSupportedGeneration - 1;
                const loaderCompatDetailsForRuntime = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: loaderGeneration,
                    supportedFeatures: new Set(runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures),
                };
                node_assert_1.strict.throws(() => (0, runtimeLayerCompatState_js_1.validateLoaderCompatibility)(loaderCompatDetailsForRuntime, disposeFn), (e) => validateFailureProperties(e, false /* isGenerationCompatible */, loaderGeneration, "Loader"), "Runtime should be incompatible with Loader layer");
                (0, node_assert_1.strict)(disposeFn.calledOnce, "Dispose should be called");
            });
            it("Runtime features are incompatible with Loader", () => {
                const disposeFn = sinon_1.default.fake();
                const requiredFeatures = ["feature2", "feature3"];
                runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures = requiredFeatures;
                const loaderCompatDetailsForRuntime = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: runtimeLayerCompatState_js_1.loaderSupportRequirements.minSupportedGeneration,
                    supportedFeatures: new Set(),
                };
                node_assert_1.strict.throws(() => (0, runtimeLayerCompatState_js_1.validateLoaderCompatibility)(loaderCompatDetailsForRuntime, disposeFn), (e) => validateFailureProperties(e, true /* isGenerationCompatible */, runtimeLayerCompatState_js_1.loaderSupportRequirements.minSupportedGeneration, "Loader", requiredFeatures), "Runtime should be incompatible with Loader layer");
                (0, node_assert_1.strict)(disposeFn.calledOnce, "Dispose should be called");
            });
            it("Runtime generation and features are both incompatible with Loader", () => {
                const disposeFn = sinon_1.default.fake();
                const loaderGeneration = runtimeLayerCompatState_js_1.loaderSupportRequirements.minSupportedGeneration - 1;
                const requiredFeatures = ["feature2"];
                runtimeLayerCompatState_js_1.loaderSupportRequirements.requiredFeatures = requiredFeatures;
                const loaderCompatDetailsForRuntime = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: loaderGeneration,
                    supportedFeatures: new Set(),
                };
                node_assert_1.strict.throws(() => (0, runtimeLayerCompatState_js_1.validateLoaderCompatibility)(loaderCompatDetailsForRuntime, disposeFn), (e) => validateFailureProperties(e, false /* isGenerationCompatible */, loaderGeneration, "Loader", requiredFeatures), "Runtime should be incompatible with Loader layer");
                (0, node_assert_1.strict)(disposeFn.calledOnce, "Dispose should be called");
            });
        });
        /**
         * These tests validates that the Runtime layer compatibility is correctly enforced during container runtime creation.
         */
        describe("Container Runtime creation", () => {
            const mockProvideEntryPoint = async () => ({
                myProp: "myValue",
            });
            const localGetMockContext = (compatibilityDetails) => {
                return {
                    attachState: internal_1.AttachState.Attached,
                    deltaManager: new internal_4.MockDeltaManager(),
                    audience: new internal_4.MockAudience(),
                    quorum: new internal_4.MockQuorumClients(),
                    taggedLogger: (0, internal_3.createChildLogger)({}),
                    clientDetails: { capabilities: { interactive: true } },
                    closeFn: (error) => { },
                    updateDirtyContainerState: (_dirty) => { },
                    getLoadedFromVersion: () => undefined,
                    ILayerCompatDetails: compatibilityDetails,
                };
            };
            it("Older Loader is compatible", async () => {
                await node_assert_1.strict.doesNotReject(async () => containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: localGetMockContext(),
                    registryEntries: [],
                    existing: false,
                    provideEntryPoint: mockProvideEntryPoint,
                }), "Older Loader should be compatible");
            });
            it("Loader with generation >= minSupportedGeneration is compatible", async () => {
                const loaderCompatDetailsForRuntime = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: runtimeLayerCompatState_js_1.loaderSupportRequirements.minSupportedGeneration,
                    supportedFeatures: new Set(),
                };
                await node_assert_1.strict.doesNotReject(async () => containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: localGetMockContext(loaderCompatDetailsForRuntime),
                    registryEntries: [],
                    existing: false,
                    provideEntryPoint: mockProvideEntryPoint,
                }), "Loader with generation >= minSupportedGeneration should be compatible");
            });
            it("Loader with generation < minSupportedGeneration is not compatible", async () => {
                const loaderGeneration = runtimeLayerCompatState_js_1.loaderSupportRequirements.minSupportedGeneration - 1;
                const loaderCompatDetailsForRuntime = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: loaderGeneration,
                    supportedFeatures: new Set(),
                };
                await node_assert_1.strict.rejects(async () => containerRuntime_js_1.ContainerRuntime.loadRuntime({
                    context: localGetMockContext(loaderCompatDetailsForRuntime),
                    registryEntries: [],
                    existing: false,
                    provideEntryPoint: mockProvideEntryPoint,
                }), (error) => validateFailureProperties(error, false /* isGenerationCompatible */, loaderGeneration, "Loader"), "Loader with generation < minSupportedGeneration should be incompatible");
            });
        });
    });
    describe("Runtime <-> DataStore compatibility", () => {
        let originalRequiredFeatures;
        beforeEach(() => {
            originalRequiredFeatures = runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures;
        });
        afterEach(() => {
            runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures = [...originalRequiredFeatures];
        });
        /**
         * These tests validates that the layer compat state and validateDataStoreCompatibility function correctly
         * validate the compatibility between Runtime and DataStore layers.
         */
        describe("validateDataStoreCompatibility", () => {
            it("Runtime is compatible with old DataStore (pre-enforcement)", () => {
                // Older DataStore will not have ILayerCompatDetails defined.
                node_assert_1.strict.doesNotThrow(() => (0, runtimeLayerCompatState_js_1.validateDatastoreCompatibility)(undefined /* maybeDataStoreCompatDetails */, () => {
                    throw new Error("should not dispose");
                }), "Runtime should be compatible with older DataStore");
            });
            it("Runtime generation and features are compatible with DataStore", () => {
                runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures = ["feature1", "feature2"];
                const dataStoreCompatDetails = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: runtimeLayerCompatState_js_1.dataStoreSupportRequirements.minSupportedGeneration,
                    supportedFeatures: new Set(runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures),
                };
                node_assert_1.strict.doesNotThrow(() => (0, runtimeLayerCompatState_js_1.validateDatastoreCompatibility)(dataStoreCompatDetails, () => {
                    throw new Error("should not dispose");
                }), "Runtime should be compatible with DataStore layer");
            });
            it("Runtime generation is incompatible with DataStore", () => {
                const disposeFn = sinon_1.default.fake();
                runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures = ["feature1", "feature2"];
                const dataStoreGeneration = runtimeLayerCompatState_js_1.dataStoreSupportRequirements.minSupportedGeneration - 1;
                const dataStoreCompatDetails = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: dataStoreGeneration,
                    supportedFeatures: new Set(runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures),
                };
                node_assert_1.strict.throws(() => (0, runtimeLayerCompatState_js_1.validateDatastoreCompatibility)(dataStoreCompatDetails, disposeFn), (e) => validateFailureProperties(e, false /* isGenerationCompatible */, dataStoreGeneration, "DataStore"), "Runtime should be incompatible with DataStore layer");
                (0, node_assert_1.strict)(disposeFn.calledOnce, "Dispose should be called");
            });
            it("Runtime features are incompatible with DataStore", () => {
                const disposeFn = sinon_1.default.fake();
                const requiredFeatures = ["feature2", "feature3"];
                runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures = requiredFeatures;
                const dataStoreCompatDetails = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: runtimeLayerCompatState_js_1.dataStoreSupportRequirements.minSupportedGeneration,
                    supportedFeatures: new Set(),
                };
                node_assert_1.strict.throws(() => (0, runtimeLayerCompatState_js_1.validateDatastoreCompatibility)(dataStoreCompatDetails, disposeFn), (e) => validateFailureProperties(e, true /* isGenerationCompatible */, runtimeLayerCompatState_js_1.dataStoreSupportRequirements.minSupportedGeneration, "DataStore", requiredFeatures), "Runtime should be incompatible with DataStore layer");
                (0, node_assert_1.strict)(disposeFn.calledOnce, "Dispose should be called");
            });
            it("Runtime generation and features are both incompatible with DataStore", () => {
                const disposeFn = sinon_1.default.fake();
                const dataStoreGeneration = runtimeLayerCompatState_js_1.dataStoreSupportRequirements.minSupportedGeneration - 1;
                const requiredFeatures = ["feature2"];
                runtimeLayerCompatState_js_1.dataStoreSupportRequirements.requiredFeatures = requiredFeatures;
                const dataStoreCompatDetails = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: dataStoreGeneration,
                    supportedFeatures: new Set(),
                };
                node_assert_1.strict.throws(() => (0, runtimeLayerCompatState_js_1.validateDatastoreCompatibility)(dataStoreCompatDetails, disposeFn), (e) => validateFailureProperties(e, false /* isGenerationCompatible */, dataStoreGeneration, "DataStore", requiredFeatures), "Runtime should be incompatible with DataStore layer");
                (0, node_assert_1.strict)(disposeFn.calledOnce, "Dispose should be called");
            });
        });
        /**
         * These tests validates that the Runtime layer compatibility is correctly enforced during data store context realization.
         */
        describe("FluidDataStoreContext realization", () => {
            it("Older DataStore is compatible", async () => {
                const localDataStoreContext = (0, dataStoreCreationHelper_js_1.createLocalDataStoreContext)({});
                await node_assert_1.strict.doesNotReject(async () => localDataStoreContext.realize(), "Older DataStore should be compatible");
            });
            it("DataStore with generation >= minSupportedGeneration is compatible", async () => {
                const dataStoreCompatDetails = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: runtimeLayerCompatState_js_1.dataStoreSupportRequirements.minSupportedGeneration,
                    supportedFeatures: new Set(),
                };
                const localDataStoreContext = (0, dataStoreCreationHelper_js_1.createLocalDataStoreContext)({}, dataStoreCompatDetails);
                await node_assert_1.strict.doesNotReject(async () => localDataStoreContext.realize(), "DataStore with generation >= minSupportedGeneration should be compatible");
            });
            it("DataStore with generation < minSupportedGeneration is not compatible", async () => {
                const dataStoreGeneration = runtimeLayerCompatState_js_1.dataStoreSupportRequirements.minSupportedGeneration - 1;
                const dataStoreCompatDetails = {
                    pkgVersion: packageVersion_js_1.pkgVersion,
                    generation: dataStoreGeneration,
                    supportedFeatures: new Set(),
                };
                const localDataStoreContext = (0, dataStoreCreationHelper_js_1.createLocalDataStoreContext)({}, dataStoreCompatDetails);
                await node_assert_1.strict.rejects(async () => localDataStoreContext.realize(), (error) => validateFailureProperties(error, false /* isGenerationCompatible */, dataStoreGeneration, "DataStore"), "DataStore with generation < minSupportedGeneration should be incompatible");
            });
        });
    });
});
//# sourceMappingURL=runtimeLayerCompatValidation.spec.js.map