"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.createNewSignalEnvelope = exports.ContainerRuntime = exports.loadContainerRuntime = exports.getSingleUseLegacyLogCallback = exports.makeLegacySendBatchFn = exports.getDeviceSpec = exports.agentSchedulerId = exports.isUnpackedRuntimeMessage = exports.defaultPendingOpsRetryDelayMs = exports.defaultPendingOpsWaitTimeoutMs = exports.defaultRuntimeHeaderData = exports.InactiveResponseHeaderKey = exports.TombstoneResponseHeaderKey = exports.DeletedResponseHeaderKey = void 0;
const client_utils_1 = require("@fluid-internal/client-utils");
const container_definitions_1 = require("@fluidframework/container-definitions");
const internal_1 = require("@fluidframework/container-definitions/internal");
const internal_2 = require("@fluidframework/core-utils/internal");
const driver_definitions_1 = require("@fluidframework/driver-definitions");
const internal_3 = require("@fluidframework/driver-definitions/internal");
const internal_4 = require("@fluidframework/driver-utils/internal");
const internal_5 = require("@fluidframework/id-compressor/internal");
const internal_6 = require("@fluidframework/runtime-definitions/internal");
const internal_7 = require("@fluidframework/runtime-utils/internal");
const internal_8 = require("@fluidframework/telemetry-utils/internal");
const uuid_1 = require("uuid");
const batchTracker_js_1 = require("./batchTracker.js");
const index_js_1 = require("./blobManager/index.js");
const channelCollection_js_1 = require("./channelCollection.js");
const compatUtils_js_1 = require("./compatUtils.js");
const compressionDefinitions_js_1 = require("./compressionDefinitions.js");
const connectionTelemetry_js_1 = require("./connectionTelemetry.js");
const containerHandleContext_js_1 = require("./containerHandleContext.js");
const dataStore_js_1 = require("./dataStore.js");
const dataStoreRegistry_js_1 = require("./dataStoreRegistry.js");
const deltaManagerProxies_js_1 = require("./deltaManagerProxies.js");
const deltaScheduler_js_1 = require("./deltaScheduler.js");
const index_js_2 = require("./gc/index.js");
const inboundBatchAggregator_js_1 = require("./inboundBatchAggregator.js");
const messageTypes_js_1 = require("./messageTypes.js");
const index_js_3 = require("./opLifecycle/index.js");
const packageVersion_js_1 = require("./packageVersion.js");
const pendingStateManager_js_1 = require("./pendingStateManager.js");
const runCounter_js_1 = require("./runCounter.js");
const runtimeLayerCompatState_js_1 = require("./runtimeLayerCompatState.js");
const signalTelemetryProcessing_js_1 = require("./signalTelemetryProcessing.js");
const index_js_4 = require("./summary/index.js");
const throttler_js_1 = require("./throttler.js");
/**
 * Creates an error object to be thrown / passed to Container's close fn in case of an unknown message type.
 * The parameters are typed to support compile-time enforcement of handling all known types/behaviors
 *
 * @param unknownContainerRuntimeMessageType - Typed as something unexpected, to ensure all known types have been
 * handled before calling this function (e.g. in a switch statement).
 *
 * @param codePath - The code path where the unexpected message type was encountered.
 *
 * @param sequencedMessage - The sequenced message that contained the unexpected message type.
 *
 */
function getUnknownMessageTypeError(unknownContainerRuntimeMessageType, codePath, sequencedMessage) {
    return internal_8.DataProcessingError.create("Runtime message of unknown type", codePath, sequencedMessage, {
        messageDetails: {
            type: unknownContainerRuntimeMessageType,
        },
    });
}
/**
 * Error responses when requesting a deleted object will have this header set to true
 * @internal
 */
exports.DeletedResponseHeaderKey = "wasDeleted";
/**
 * Tombstone error responses will have this header set to true
 * @legacy
 * @alpha
 */
exports.TombstoneResponseHeaderKey = "isTombstoned";
/**
 * Inactive error responses will have this header set to true
 * @legacy
 * @alpha
 *
 * @deprecated this header is deprecated and will be removed in the future. The functionality corresponding
 * to this was experimental and is no longer supported.
 */
exports.InactiveResponseHeaderKey = "isInactive";
/**
 * Default values for Runtime Headers
 */
exports.defaultRuntimeHeaderData = {
    wait: true,
    viaHandle: false,
    allowTombstone: false,
};
const defaultStagingCommitOptions = { squash: false };
const maxConsecutiveReconnectsKey = "Fluid.ContainerRuntime.MaxConsecutiveReconnects";
// The actual limit is 1Mb (socket.io and Kafka limits)
// We can't estimate it fully, as we
// - do not know what properties relay service will add
// - we do not stringify final op, thus we do not know how much escaping will be added.
const defaultMaxBatchSizeInBytes = 700 * 1024;
const defaultChunkSizeInBytes = 204800;
/**
 * The default time to wait for pending ops to be processed during summarization
 */
exports.defaultPendingOpsWaitTimeoutMs = 1000;
/**
 * The default time to delay a summarization retry attempt when there are pending ops
 */
exports.defaultPendingOpsRetryDelayMs = 1000;
/**
 * Instead of refreshing from latest because we do not have 100% confidence in the state
 * of the current system, we should close the summarizer and let it recover.
 * This delay's goal is to prevent tight restart loops
 */
const defaultCloseSummarizerDelayMs = 5000; // 5 seconds
/**
 * Checks whether a message.type is one of the values in ContainerMessageType
 */
function isUnpackedRuntimeMessage(message) {
    return Object.values(messageTypes_js_1.ContainerMessageType).includes(message.type);
}
exports.isUnpackedRuntimeMessage = isUnpackedRuntimeMessage;
/**
 * Legacy ID for the built-in AgentScheduler.  To minimize disruption while removing it, retaining this as a
 * special-case for document dirty state.  Ultimately we should have no special-cases from the
 * ContainerRuntime's perspective.
 * @internal
 */
exports.agentSchedulerId = "_scheduler";
// safely check navigator and get the hardware spec value
function getDeviceSpec() {
    try {
        if (typeof navigator === "object" && navigator !== null) {
            return {
                // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-assignment
                deviceMemory: navigator.deviceMemory,
                hardwareConcurrency: navigator.hardwareConcurrency,
            };
        }
    }
    catch {
        // Eat the error
    }
    return {};
}
exports.getDeviceSpec = getDeviceSpec;
/**
 * Older loader doesn't have a submitBatchFn member, this is the older way of submitting a batch.
 * Rather than exposing the submitFn (now deprecated) and IDeltaManager (dangerous to hand out) to the Outbox,
 * we can provide a partially-applied function to keep those items private to the ContainerRuntime.
 */
const makeLegacySendBatchFn = (submitFn, deltaManager) => (batch) => {
    // Default to negative one to match Container.submitBatch behavior
    let clientSequenceNumber = -1;
    for (const message of batch.messages) {
        clientSequenceNumber = submitFn(internal_3.MessageType.Operation, 
        // For back-compat (submitFn only works on deserialized content)
        message.contents === undefined ? undefined : JSON.parse(message.contents), true, // batch
        message.metadata);
    }
    deltaManager.flush();
    return clientSequenceNumber;
};
exports.makeLegacySendBatchFn = makeLegacySendBatchFn;
/**
 * Extract last message from the snapshot metadata.
 * Uses legacy property if not using explicit schema control, otherwise uses the new property.
 * This allows new runtime to make documents not openable for old runtimes, one explicit document schema control is enabled.
 * Please see addMetadataToSummary() as well
 */
function lastMessageFromMetadata(metadata) {
    return metadata?.documentSchema?.runtime?.explicitSchemaControl
        ? metadata?.lastMessage
        : metadata?.message;
}
/**
 * There is some ancient back-compat code that we'd like to instrument
 * to understand if/when it is hit.
 * We only want to log this once, to avoid spamming telemetry if we are wrong and these cases are hit commonly.
 */
let getSingleUseLegacyLogCallback = (logger, type) => {
    return (codePath) => {
        logger.sendTelemetryEvent({
            eventName: "LegacyMessageFormat",
            details: { codePath, type },
        });
        // Now that we've logged, prevent future logging (globally).
        // eslint-disable-next-line unicorn/consistent-function-scoping
        exports.getSingleUseLegacyLogCallback = () => () => { };
    };
};
exports.getSingleUseLegacyLogCallback = getSingleUseLegacyLogCallback;
/**
 * This is meant to be used by a {@link @fluidframework/container-definitions#IRuntimeFactory} to instantiate a container runtime.
 * @param params - An object which specifies all required and optional params necessary to instantiate a runtime.
 * @returns A runtime which provides all the functionality necessary to bind with the loader layer via the {@link @fluidframework/container-definitions#IRuntime} interface and provide a runtime environment via the {@link @fluidframework/container-runtime-definitions#IContainerRuntime} interface.
 * @legacy
 * @alpha
 */
async function loadContainerRuntime(params) {
    return ContainerRuntime.loadRuntime(params);
}
exports.loadContainerRuntime = loadContainerRuntime;
const defaultMaxConsecutiveReconnects = 7;
/**
 * Represents the runtime of the container. Contains helper functions/state of the container.
 * It will define the store level mappings.
 *
 * @internal
 */
class ContainerRuntime extends client_utils_1.TypedEventEmitter {
    /**
     * Load the stores from a snapshot and returns the runtime.
     * @param params - An object housing the runtime properties:
     * - context - Context of the container.
     * - registryEntries - Mapping from data store types to their corresponding factories.
     * - existing - Pass 'true' if loading from an existing snapshot.
     * - requestHandler - (optional) Request handler for the request() method of the container runtime.
     * Only relevant for back-compat while we remove the request() method and move fully to entryPoint as the main pattern.
     * - runtimeOptions - Additional options to be passed to the runtime
     * - containerScope - runtime services provided with context
     * - containerRuntimeCtor - Constructor to use to create the ContainerRuntime instance.
     * This allows mixin classes to leverage this method to define their own async initializer.
     * - provideEntryPoint - Promise that resolves to an object which will act as entryPoint for the Container.
     * - minVersionForCollab - Minimum version of the FF runtime that this runtime supports collaboration with.
     * This object should provide all the functionality that the Container is expected to provide to the loader layer.
     */
    static async loadRuntime(params) {
        const { context, registryEntries, existing, requestHandler, provideEntryPoint, runtimeOptions = {}, containerScope = {}, containerRuntimeCtor = ContainerRuntime, minVersionForCollab = compatUtils_js_1.defaultMinVersionForCollab, } = params;
        // If taggedLogger exists, use it. Otherwise, wrap the vanilla logger:
        // back-compat: Remove the TaggedLoggerAdapter fallback once all the host are using loader > 0.45
        const backCompatContext = context;
        const passLogger = backCompatContext.taggedLogger ??
            // eslint-disable-next-line import/no-deprecated
            new internal_8.TaggedLoggerAdapter(backCompatContext.logger);
        const logger = (0, internal_8.createChildLogger)({
            logger: passLogger,
            properties: {
                all: {
                    runtimeVersion: packageVersion_js_1.pkgVersion,
                },
            },
        });
        const mc = (0, internal_8.loggerToMonitoringContext)(logger);
        // Some options require a minimum version of the FF runtime to operate, so the default configs will be generated
        // based on the minVersionForCollab.
        // For example, if minVersionForCollab is set to "1.0.0", the default configs will ensure compatibility with FF runtime
        // 1.0.0 or later. If the minVersionForCollab is set to "2.10.0", the default values will be generated to ensure compatibility
        // with FF runtime 2.10.0 or later.
        if (!(0, compatUtils_js_1.isValidMinVersionForCollab)(minVersionForCollab)) {
            throw new internal_8.UsageError(`Invalid minVersionForCollab: ${minVersionForCollab}. It must be an existing FF version (i.e. 2.22.1).`);
        }
        const defaultsAffectingDocSchema = (0, compatUtils_js_1.getMinVersionForCollabDefaults)(minVersionForCollab);
        // The following are the default values for the options that do not affect the DocumentSchema.
        const defaultsNotAffectingDocSchema = {
            summaryOptions: {},
            loadSequenceNumberVerification: "close",
            maxBatchSizeInBytes: defaultMaxBatchSizeInBytes,
            chunkSizeInBytes: defaultChunkSizeInBytes,
        };
        const defaultConfigs = {
            ...defaultsAffectingDocSchema,
            ...defaultsNotAffectingDocSchema,
        };
        // Here we set each option to its corresponding default config value if it's not provided in runtimeOptions.
        // Note: We cannot do a simple object merge of defaultConfigs/runtimeOptions because in most cases we don't want
        // a option that is undefined in runtimeOptions to override the default value (except for idCompressor, see below).
        const { summaryOptions = defaultConfigs.summaryOptions, gcOptions = defaultConfigs.gcOptions, loadSequenceNumberVerification = defaultConfigs.loadSequenceNumberVerification, maxBatchSizeInBytes = defaultConfigs.maxBatchSizeInBytes, chunkSizeInBytes = defaultConfigs.chunkSizeInBytes, explicitSchemaControl = defaultConfigs.explicitSchemaControl, enableGroupedBatching = defaultConfigs.enableGroupedBatching, flushMode = defaultConfigs.flushMode, 
        // If batching is disabled then we should disable compression as well. If batching is disabled and compression
        // is enabled via runtimeOptions, we will throw an error later.
        compressionOptions = enableGroupedBatching === false
            ? compressionDefinitions_js_1.disabledCompressionConfig
            : defaultConfigs.compressionOptions, createBlobPayloadPending = defaultConfigs.createBlobPayloadPending, } = runtimeOptions;
        // The logic for enableRuntimeIdCompressor is a bit different. Since `undefined` represents a logical state (off)
        // we need to check it's explicitly set in runtimeOptions. If so, we should use that value even if it's undefined.
        const enableRuntimeIdCompressor = "enableRuntimeIdCompressor" in runtimeOptions
            ? runtimeOptions.enableRuntimeIdCompressor
            : defaultConfigs.enableRuntimeIdCompressor;
        const registry = new dataStoreRegistry_js_1.FluidDataStoreRegistry(registryEntries);
        const tryFetchBlob = async (blobName) => {
            const blobId = context.baseSnapshot?.blobs[blobName];
            if (context.baseSnapshot && blobId) {
                // IContainerContext storage api return type still has undefined in 0.39 package version.
                // So once we release 0.40 container-defn package we can remove this check.
                (0, internal_2.assert)(context.storage !== undefined, 0x1f5 /* "Attached state should have storage" */);
                return (0, internal_4.readAndParse)(context.storage, blobId);
            }
        };
        const [chunks, recentBatchInfo, metadata, electedSummarizerData, aliases, serializedIdCompressor,] = await Promise.all([
            tryFetchBlob(index_js_4.chunksBlobName),
            tryFetchBlob(index_js_4.recentBatchInfoBlobName),
            tryFetchBlob(index_js_4.metadataBlobName),
            tryFetchBlob(index_js_4.electedSummarizerBlobName),
            tryFetchBlob(index_js_4.aliasBlobName),
            tryFetchBlob(index_js_4.idCompressorBlobName),
        ]);
        // read snapshot blobs needed for BlobManager to load
        const blobManagerLoadInfo = await (0, index_js_1.loadBlobManagerLoadInfo)(context);
        const messageAtLastSummary = lastMessageFromMetadata(metadata);
        // Verify summary runtime sequence number matches protocol sequence number.
        const runtimeSequenceNumber = messageAtLastSummary?.sequenceNumber;
        const protocolSequenceNumber = context.deltaManager.initialSequenceNumber;
        // When we load with pending state, we reuse an old snapshot so we don't expect these numbers to match
        if (!context.pendingLocalState && runtimeSequenceNumber !== undefined) {
            // Unless bypass is explicitly set, then take action when sequence numbers mismatch.
            // eslint-disable-next-line unicorn/no-lonely-if -- Separate if statements make flow easier to parse
            if (loadSequenceNumberVerification !== "bypass" &&
                runtimeSequenceNumber !== protocolSequenceNumber) {
                // Message to OCEs:
                // You can hit this error with runtimeSequenceNumber === -1 in < 2.0 RC3 builds.
                // This would indicate that explicit schema control is enabled in current (2.0 RC3+) builds and it
                // results in addMetadataToSummary() creating a poison pill for older runtimes in the form of a -1 sequence number.
                // Older runtimes do not understand new schema, and thus could corrupt document if they proceed, thus we are using
                // this poison pill to prevent them from proceeding.
                // "Load from summary, runtime metadata sequenceNumber !== initialSequenceNumber"
                const error = new internal_8.DataCorruptionError(
                // pre-0.58 error message: SummaryMetadataMismatch
                "Summary metadata mismatch", { runtimeVersion: packageVersion_js_1.pkgVersion, runtimeSequenceNumber, protocolSequenceNumber });
                if (loadSequenceNumberVerification === "log") {
                    logger.sendErrorEvent({ eventName: "SequenceNumberMismatch" }, error);
                }
                else {
                    context.closeFn(error);
                }
            }
        }
        let desiredIdCompressorMode;
        switch (mc.config.getBoolean("Fluid.ContainerRuntime.IdCompressorEnabled")) {
            case true: {
                desiredIdCompressorMode = "on";
                break;
            }
            case false: {
                desiredIdCompressorMode = undefined;
                break;
            }
            default: {
                desiredIdCompressorMode = enableRuntimeIdCompressor;
                break;
            }
        }
        // Enabling the IdCompressor is a one-way operation and we only want to
        // allow new containers to turn it on.
        let idCompressorMode;
        if (existing) {
            // This setting has to be sticky for correctness:
            // 1) if compressior is OFF, it can't be enabled, as already running clients (in given document session) do not know
            //    how to process compressor ops
            // 2) if it's ON, then all sessions should load compressor right away
            // 3) Same logic applies for "delayed" mode
            // Maybe in the future we will need to enabled (and figure how to do it safely) "delayed" -> "on" change.
            // We could do "off" -> "on" transition too, if all clients start loading compressor (but not using it initially) and
            // do so for a while - this will allow clients to eventually disregard "off" setting (when it's safe so) and start
            // using compressor in future sessions.
            // Everyting is possible, but it needs to be designed and executed carefully, when such need arises.
            idCompressorMode = metadata?.documentSchema?.runtime
                ?.idCompressorMode;
            // This is the only exception to the rule above - we have proper plumbing to load ID compressor on schema change
            // event. It is loaded async (relative to op processing), so this conversion is only safe for off -> delayed conversion!
            // Clients do not expect ID compressor ops unless ID compressor is On for them, and that could be achieved only through
            // explicit schema change, i.e. only if explicitSchemaControl is on.
            // Note: it would be better if we throw on combination of options (explicitSchemaControl = off, desiredIdCompressorMode === "delayed")
            // that is not supported. But our service tests are oblivious to these problems and throwing here will cause a ton of failures
            // We ignored incompatible ID compressor changes from the start (they were sticky), so that's not a new problem being introduced...
            if (idCompressorMode === undefined &&
                desiredIdCompressorMode === "delayed" &&
                explicitSchemaControl) {
                idCompressorMode = desiredIdCompressorMode;
            }
        }
        else {
            idCompressorMode = desiredIdCompressorMode;
        }
        const createIdCompressorFn = () => {
            /**
             * Because the IdCompressor emits so much telemetry, this function is used to sample
             * approximately 5% of all clients. Only the given percentage of sessions will emit telemetry.
             */
            const idCompressorEventSampler = (() => {
                const isIdCompressorTelemetryEnabled = Math.random() < 0.05;
                return {
                    sample: () => {
                        return isIdCompressorTelemetryEnabled;
                    },
                };
            })();
            const compressorLogger = (0, internal_8.createSampledLogger)(logger, idCompressorEventSampler);
            const pendingLocalState = context.pendingLocalState;
            if (pendingLocalState?.pendingIdCompressorState !== undefined) {
                return (0, internal_5.deserializeIdCompressor)(pendingLocalState.pendingIdCompressorState, compressorLogger);
            }
            else if (serializedIdCompressor === undefined) {
                return (0, internal_5.createIdCompressor)(compressorLogger);
            }
            else {
                return (0, internal_5.deserializeIdCompressor)(serializedIdCompressor, (0, internal_5.createSessionId)(), compressorLogger);
            }
        };
        const compressionLz4 = compressionOptions.minimumBatchSizeInBytes !== Number.POSITIVE_INFINITY &&
            compressionOptions.compressionAlgorithm === "lz4";
        const documentSchemaController = new index_js_4.DocumentsSchemaController(existing, protocolSequenceNumber, metadata?.documentSchema, {
            explicitSchemaControl,
            compressionLz4,
            idCompressorMode,
            opGroupingEnabled: enableGroupedBatching,
            createBlobPayloadPending,
            disallowedVersions: [],
        }, (schema) => {
            runtime.onSchemaChange(schema);
        });
        if (compressionLz4 && !enableGroupedBatching) {
            throw new internal_8.UsageError("If compression is enabled, op grouping must be enabled too");
        }
        const featureGatesForTelemetry = {};
        // Make sure we've got all the options including internal ones
        const internalRuntimeOptions = {
            summaryOptions,
            gcOptions,
            loadSequenceNumberVerification,
            flushMode,
            compressionOptions,
            maxBatchSizeInBytes,
            chunkSizeInBytes,
            enableRuntimeIdCompressor,
            enableGroupedBatching,
            explicitSchemaControl,
            createBlobPayloadPending,
        };
        const runtime = new containerRuntimeCtor(context, registry, metadata, electedSummarizerData, chunks ?? [], aliases ?? [], internalRuntimeOptions, containerScope, logger, existing, blobManagerLoadInfo, context.storage, createIdCompressorFn, documentSchemaController, featureGatesForTelemetry, provideEntryPoint, minVersionForCollab, requestHandler, undefined, // summaryConfiguration
        recentBatchInfo);
        runtime.blobManager.stashedBlobsUploadP.then(() => {
            // make sure we didn't reconnect before the promise resolved
            if (runtime.delayConnectClientId !== undefined && !runtime.disposed) {
                runtime.delayConnectClientId = undefined;
                runtime.setConnectionStateCore(true, runtime.delayConnectClientId);
            }
        }, (error) => runtime.closeFn(error));
        // Apply stashed ops with a reference sequence number equal to the sequence number of the snapshot,
        // or zero. This must be done before Container replays saved ops.
        await runtime.pendingStateManager.applyStashedOpsAt(runtimeSequenceNumber ?? 0);
        // Initialize the base state of the runtime before it's returned.
        await runtime.initializeBaseState(context.loader);
        return runtime;
    }
    get clientId() {
        return this._getClientId();
    }
    get storage() {
        return this._storage;
    }
    get containerRuntime() {
        return this;
    }
    get flushMode() {
        return this._flushMode;
    }
    get scope() {
        return this.containerScope;
    }
    get IFluidDataStoreRegistry() {
        return this.registry;
    }
    get attachState() {
        return this._getAttachState();
    }
    /**
     * Current session schema - defines what options are on & off.
     * It's overlap of document schema (controlled by summary & ops) and options controlling this session.
     * For example, document schema might have compression ON, but feature gates / runtime options turn it Off.
     * In such case it will be off in session schema (i.e. this session should not use compression), but this client
     * has to deal with compressed ops as other clients might send them.
     * And in reverse, session schema can have compression Off, but feature gates / runtime options want it On.
     * In such case it will be off in session schema, however this client will propose change to schema, and once / if
     * this op roundtrips, compression will be On. Client can't send compressed ops until it's change in schema.
     */
    get sessionSchema() {
        return this.documentsSchemaController.sessionSchema.runtime;
    }
    /**
     * {@inheritDoc @fluidframework/runtime-definitions#IContainerRuntimeBase.idCompressor}
     */
    get idCompressor() {
        // Expose ID Compressor only if it's On from the start.
        // If container uses delayed mode, then we can only expose generateDocumentUniqueId() and nothing else.
        // That's because any other usage will require immidiate loading of ID Compressor in next sessions in order
        // to reason over such things as session ID space.
        if (this.sessionSchema.idCompressorMode === "on") {
            (0, internal_2.assert)(this._idCompressor !== undefined, 0x8ea /* compressor should have been loaded */);
            return this._idCompressor;
        }
    }
    /**
     * {@inheritDoc @fluidframework/runtime-definitions#IContainerRuntimeBase.generateDocumentUniqueId}
     */
    generateDocumentUniqueId() {
        return this._idCompressor?.generateDocumentUniqueId() ?? (0, uuid_1.v4)();
    }
    get IFluidHandleContext() {
        return this.handleContext;
    }
    /**
     * This is a proxy to the delta manager provided by the container context (innerDeltaManager). It restricts certain
     * accesses such as sets "read-only" mode for the summarizer client. This is the default delta manager that should
     * be used unless the innerDeltaManager is required.
     */
    get deltaManager() {
        return this._deltaManager;
    }
    /**
     * Invokes the given callback and expects that no ops are submitted
     * until execution finishes. If an op is submitted, an error will be raised.
     *
     * @param callback - the callback to be invoked
     */
    ensureNoDataModelChanges(callback) {
        return this.dataModelChangeRunner.run(callback);
    }
    get connected() {
        return this._connected;
    }
    /**
     * clientId of parent (non-summarizing) container that owns summarizer container
     */
    get summarizerClientId() {
        return this.summarizerClientElection?.electedClientId;
    }
    get disposed() {
        return this._disposed;
    }
    /**
     * If false, loading or using a Tombstoned object should merely log, not fail.
     * @deprecated NOT SUPPORTED - hardcoded to return false since it's deprecated.
     */
    // eslint-disable-next-line @typescript-eslint/class-literal-property-style
    get gcTombstoneEnforcementAllowed() {
        return false;
    }
    /**
     * If true, throw an error when a tombstone data store is used.
     * @deprecated NOT SUPPORTED - hardcoded to return false since it's deprecated.
     */
    // eslint-disable-next-line @typescript-eslint/class-literal-property-style
    get gcThrowOnTombstoneUsage() {
        return false;
    }
    /**
     * The compatibility details of the Runtime layer that is exposed to the Loader layer
     * for validating Loader-Runtime compatibility.
     */
    get ILayerCompatDetails() {
        return runtimeLayerCompatState_js_1.runtimeCompatDetailsForLoader;
    }
    /***/
    constructor(context, registry, metadata, electedSummarizerData, chunks, dataStoreAliasMap, runtimeOptions, containerScope, 
    // Create a custom ITelemetryBaseLogger to output telemetry events.
    baseLogger, existing, blobManagerLoadInfo, _storage, createIdCompressorFn, documentsSchemaController, featureGatesForTelemetry, provideEntryPoint, minVersionForCollab, requestHandler, 
    // // eslint-disable-next-line unicorn/no-object-as-default-parameter
    summaryConfiguration = {
        // the defaults
        ...index_js_4.DefaultSummaryConfiguration,
        // the runtime configuration overrides
        ...runtimeOptions.summaryOptions?.summaryConfigOverrides,
    }, recentBatchInfo) {
        super();
        this.registry = registry;
        this.metadata = metadata;
        this.electedSummarizerData = electedSummarizerData;
        this.runtimeOptions = runtimeOptions;
        this.containerScope = containerScope;
        this.baseLogger = baseLogger;
        this._storage = _storage;
        this.createIdCompressorFn = createIdCompressorFn;
        this.documentsSchemaController = documentsSchemaController;
        this.minVersionForCollab = minVersionForCollab;
        this.requestHandler = requestHandler;
        this.summaryConfiguration = summaryConfiguration;
        this.imminentClosure = false;
        this.isReadOnly = () => this.deltaManager.readOnlyInfo.readonly === true;
        // We accumulate Id compressor Ops while Id compressor is not loaded yet (only for "delayed" mode)
        // Once it loads, it will process all such ops and we will stop accumulating further ops - ops will be processes as they come in.
        this.pendingIdCompressorOps = [];
        this.batchRunner = new runCounter_js_1.RunCounter();
        this.flushTaskExists = false;
        this.consecutiveReconnects = 0;
        this.dataModelChangeRunner = new runCounter_js_1.RunCounter();
        this._disposed = false;
        this.emitDirtyDocumentEvent = true;
        this.signalTelemetryManager = new signalTelemetryProcessing_js_1.SignalTelemetryManager();
        /**
         * It a cache for holding mapping for loading groupIds with its snapshot from the service. Add expiry policy of 1 minute.
         * Starting with 1 min and based on recorded usage we can tweak it later on.
         */
        this.snapshotCacheForLoadingGroupIds = new internal_2.PromiseCache({
            expiry: { policy: "absolute", durationMs: 60000 },
        });
        this.notifyReadOnlyState = (readonly) => this.channelCollection.notifyReadOnlyState(readonly);
        /**
         * Enter Staging Mode, such that ops submitted to the ContainerRuntime will not be sent to the ordering service.
         * To exit Staging Mode, call either discardChanges or commitChanges on the Stage Controls returned from this method.
         *
         * @returns StageControlsExperimental - Controls for exiting Staging Mode.
         */
        // eslint-disable-next-line import/no-deprecated
        this.enterStagingMode = () => {
            if (this.stageControls !== undefined) {
                throw new Error("already in staging mode");
            }
            // Make sure all BatchManagers are empty before entering staging mode,
            // since we mark whole batches as "staged" or not to indicate whether to submit them.
            this.outbox.flush();
            const exitStagingMode = (discardOrCommit) => () => {
                // Final flush of any last staged changes
                this.outbox.flush();
                this.stageControls = undefined;
                discardOrCommit();
                this.channelCollection.notifyStagingMode(false);
            };
            // eslint-disable-next-line import/no-deprecated
            const stageControls = {
                discardChanges: exitStagingMode(() => {
                    // Pop all staged batches from the PSM and roll them back in LIFO order
                    this.pendingStateManager.popStagedBatches(({ runtimeOp, localOpMetadata }) => {
                        (0, internal_2.assert)(runtimeOp !== undefined, 0xb82 /* Staged batches expected to have runtimeOp defined */);
                        this.rollback(runtimeOp, localOpMetadata);
                    });
                    if (this.attachState === container_definitions_1.AttachState.Attached) {
                        this.updateDocumentDirtyState(this.pendingMessagesCount !== 0);
                    }
                }),
                commitChanges: (optionsParam) => {
                    const options = { ...defaultStagingCommitOptions, ...optionsParam };
                    return exitStagingMode(() => {
                        this.pendingStateManager.replayPendingStates({
                            onlyStagedBatches: true,
                            squash: options.squash ?? false,
                        });
                    })();
                },
            };
            this.stageControls = stageControls;
            this.channelCollection.notifyStagingMode(true);
            return this.stageControls;
        };
        this.readAndParseBlob = async (id) => (0, internal_4.readAndParse)(this.storage, id);
        const { options, clientDetails, connected, baseSnapshot, submitFn, submitBatchFn, submitSummaryFn, submitSignalFn, disposeFn, closeFn, deltaManager, quorum, audience, pendingLocalState, supportedFeatures, snapshotWithContents, } = context;
        // In old loaders without dispose functionality, closeFn is equivalent but will also switch container to readonly mode
        this.disposeFn = disposeFn ?? closeFn;
        // Validate that the Loader is compatible with this Runtime.
        const maybeloaderCompatDetailsForRuntime = context;
        (0, runtimeLayerCompatState_js_1.validateLoaderCompatibility)(maybeloaderCompatDetailsForRuntime.ILayerCompatDetails, this.disposeFn);
        this.mc = (0, internal_8.createChildMonitoringContext)({
            logger: this.baseLogger,
            namespace: "ContainerRuntime",
        });
        // If we support multiple algorithms in the future, then we would need to manage it here carefully.
        // We can use runtimeOptions.compressionOptions.compressionAlgorithm, but only if it's in the schema list!
        // If it's not in the list, then we will need to either use no compression, or fallback to some other (supported by format)
        // compression.
        const compressionOptions = {
            minimumBatchSizeInBytes: this.sessionSchema.compressionLz4
                ? runtimeOptions.compressionOptions.minimumBatchSizeInBytes
                : Number.POSITIVE_INFINITY,
            compressionAlgorithm: compressionDefinitions_js_1.CompressionAlgorithms.lz4,
        };
        (0, internal_2.assert)((0, internal_1.isIDeltaManagerFull)(deltaManager), 0xa80 /* Invalid delta manager */);
        this.innerDeltaManager = deltaManager;
        // Here we could wrap/intercept on these functions to block/modify outgoing messages if needed.
        // This makes ContainerRuntime the final gatekeeper for outgoing messages.
        // back-compat: ADO #1385: Make this call unconditional in the future
        this.submitSummaryFn =
            submitSummaryFn ??
                ((summaryOp, refseq) => submitFn(internal_3.MessageType.Summarize, summaryOp, false));
        this.submitSignalFn = submitSignalFn;
        // TODO: After IContainerContext.options is removed, we'll just create a new blank object {} here.
        // Values are generally expected to be set from the runtime side.
        this.options = options ?? {};
        this.clientDetails = clientDetails;
        const isSummarizerClient = this.clientDetails.type === index_js_4.summarizerClientType;
        this.loadedFromVersionId = context.getLoadedFromVersion()?.id;
        // eslint-disable-next-line unicorn/consistent-destructuring
        this._getClientId = () => context.clientId;
        // eslint-disable-next-line unicorn/consistent-destructuring
        this._getAttachState = () => context.attachState;
        this.getAbsoluteUrl = async (relativeUrl) => {
            // eslint-disable-next-line unicorn/consistent-destructuring
            if (context.getAbsoluteUrl === undefined) {
                throw new Error("Driver does not implement getAbsoluteUrl");
            }
            if (this.attachState !== container_definitions_1.AttachState.Attached) {
                return undefined;
            }
            return context.getAbsoluteUrl(relativeUrl);
        };
        // TODO: Consider that the Container could just listen to these events itself, or even more appropriately maybe the
        // customer should observe dirty state on the runtime (the owner of dirty state) directly, rather than on the IContainer.
        this.on("dirty", () => context.updateDirtyContainerState(true));
        this.on("saved", () => context.updateDirtyContainerState(false));
        // Telemetry for when the container is attached and subsequently saved for the first time.
        // These events are useful for investigating the validity of container "saved" eventing upon attach.
        // See this.setAttachState() and this.updateDocumentDirtyState() for more details on "attached" and "saved" events.
        this.once("attached", () => {
            this.mc.logger.sendTelemetryEvent({
                eventName: "Attached",
                details: {
                    dirtyContainer: this.dirtyContainer,
                    hasPendingMessages: this.hasPendingMessages(),
                },
            });
        });
        this.once("saved", () => this.mc.logger.sendTelemetryEvent({
            eventName: "Saved",
            details: { attachState: this.attachState },
        }));
        // In cases of summarizer, we want to dispose instead since consumer doesn't interact with this container
        this.closeFn = isSummarizerClient ? this.disposeFn : closeFn;
        let loadSummaryNumber;
        // Get the container creation metadata. For new container, we initialize these. For existing containers,
        // get the values from the metadata blob.
        if (existing) {
            this.createContainerMetadata = {
                createContainerRuntimeVersion: metadata?.createContainerRuntimeVersion,
                createContainerTimestamp: metadata?.createContainerTimestamp,
            };
            // summaryNumber was renamed from summaryCount. For older docs that haven't been opened for a long time,
            // the count is reset to 0.
            loadSummaryNumber = metadata?.summaryNumber ?? 0;
        }
        else {
            this.createContainerMetadata = {
                createContainerRuntimeVersion: packageVersion_js_1.pkgVersion,
                createContainerTimestamp: Date.now(),
            };
            loadSummaryNumber = 0;
        }
        this.nextSummaryNumber = loadSummaryNumber + 1;
        this.messageAtLastSummary = lastMessageFromMetadata(metadata);
        // Note that we only need to pull the *initial* connected state from the context.
        // Later updates come through calls to setConnectionState.
        this._connected = connected;
        this.mc.logger.sendTelemetryEvent({
            eventName: "GCFeatureMatrix",
            metadataValue: JSON.stringify(metadata?.gcFeatureMatrix),
            inputs: JSON.stringify({
                // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
                gcOptions_gcGeneration: runtimeOptions.gcOptions[index_js_2.gcGenerationOptionName],
            }),
        });
        this.telemetryDocumentId = metadata?.telemetryDocumentId ?? (0, uuid_1.v4)();
        const opGroupingManager = new index_js_3.OpGroupingManager({
            groupedBatchingEnabled: this.groupedBatchingEnabled,
        }, this.mc.logger);
        const opSplitter = new index_js_3.OpSplitter(chunks, submitBatchFn, runtimeOptions.chunkSizeInBytes, runtimeOptions.maxBatchSizeInBytes, this.mc.logger);
        this.remoteMessageProcessor = new index_js_3.RemoteMessageProcessor(opSplitter, new index_js_3.OpDecompressor(this.mc.logger), opGroupingManager);
        const pendingRuntimeState = pendingLocalState;
        this.pendingStateManager = new pendingStateManager_js_1.PendingStateManager({
            applyStashedOp: this.applyStashedOp.bind(this),
            clientId: () => this.clientId,
            connected: () => this.connected,
            reSubmitBatch: this.reSubmitBatch.bind(this),
            isActiveConnection: () => this.innerDeltaManager.active,
            isAttached: () => this.attachState !== container_definitions_1.AttachState.Detached,
        }, pendingRuntimeState?.pending, this.baseLogger);
        let outerDeltaManager = this.innerDeltaManager;
        this.useDeltaManagerOpsProxy =
            this.mc.config.getBoolean("Fluid.ContainerRuntime.DeltaManagerOpsProxy") === true;
        // The summarizerDeltaManager Proxy is used to lie to the summarizer to convince it is in the right state as a summarizer client.
        outerDeltaManager = deltaManagerProxies_js_1.DeltaManagerSummarizerProxy.wrapIfSummarizer(outerDeltaManager);
        // The DeltaManagerPendingOpsProxy is used to control the minimum sequence number
        // It allows us to lie to the layers below so that they can maintain enough local state for rebasing ops.
        if (this.useDeltaManagerOpsProxy) {
            const pendingOpsDeltaManagerProxy = new deltaManagerProxies_js_1.DeltaManagerPendingOpsProxy(outerDeltaManager, this.pendingStateManager);
            outerDeltaManager = pendingOpsDeltaManagerProxy;
        }
        // always wrap the exposed delta manager in at least on layer of proxying
        this._deltaManager =
            outerDeltaManager instanceof deltaManagerProxies_js_1.BaseDeltaManagerProxy
                ? outerDeltaManager
                : new deltaManagerProxies_js_1.BaseDeltaManagerProxy(outerDeltaManager);
        this.handleContext = new containerHandleContext_js_1.ContainerFluidHandleContext("", this);
        if (this.summaryConfiguration.state === "enabled") {
            (0, index_js_4.validateSummaryHeuristicConfiguration)(this.summaryConfiguration);
        }
        this.summariesDisabled = (0, index_js_4.isSummariesDisabled)(this.summaryConfiguration);
        this.maxConsecutiveReconnects =
            this.mc.config.getNumber(maxConsecutiveReconnectsKey) ?? defaultMaxConsecutiveReconnects;
        // If the context has ILayerCompatDetails, it supports referenceSequenceNumbers since that features
        // predates ILayerCompatDetails.
        const referenceSequenceNumbersSupported = maybeloaderCompatDetailsForRuntime.ILayerCompatDetails === undefined
            ? supportedFeatures?.get("referenceSequenceNumbers") === true
            : true;
        if (runtimeOptions.flushMode === internal_6.FlushModeExperimental.Async &&
            !referenceSequenceNumbersSupported) {
            // The loader does not support reference sequence numbers, falling back on FlushMode.TurnBased
            this.mc.logger.sendErrorEvent({ eventName: "FlushModeFallback" });
            this._flushMode = internal_6.FlushMode.TurnBased;
        }
        else {
            this._flushMode = runtimeOptions.flushMode;
        }
        this.offlineEnabled =
            this.mc.config.getBoolean("Fluid.Container.enableOfflineLoad") ?? false;
        if (this.offlineEnabled && this._flushMode !== internal_6.FlushMode.TurnBased) {
            const error = new internal_8.UsageError("Offline mode is only supported in turn-based mode");
            this.closeFn(error);
            throw error;
        }
        // DuplicateBatchDetection is only enabled if Offline Load is enabled
        // It maintains a cache of all batchIds/sequenceNumbers within the collab window.
        // Don't waste resources doing so if not needed.
        if (this.offlineEnabled) {
            this.duplicateBatchDetector = new index_js_3.DuplicateBatchDetector(recentBatchInfo);
        }
        // eslint-disable-next-line unicorn/consistent-destructuring
        if (context.attachState === container_definitions_1.AttachState.Attached) {
            const maxSnapshotCacheDurationMs = this._storage?.policies?.maximumCacheDurationMs;
            if (maxSnapshotCacheDurationMs !== undefined &&
                maxSnapshotCacheDurationMs > 5 * 24 * 60 * 60 * 1000) {
                // This is a runtime enforcement of what's already explicit in the policy's type itself,
                // which dictates the value is either undefined or exactly 5 days in ms.
                // As long as the actual value is less than 5 days, the assumptions GC makes here are valid.
                throw new internal_8.UsageError("Driver's maximumCacheDurationMs policy cannot exceed 5 days");
            }
        }
        this.garbageCollector = index_js_2.GarbageCollector.create({
            runtime: this,
            gcOptions: runtimeOptions.gcOptions,
            baseSnapshot,
            baseLogger: this.mc.logger,
            existing,
            metadata,
            createContainerMetadata: this.createContainerMetadata,
            isSummarizerClient,
            getNodePackagePath: async (nodePath) => this.getGCNodePackagePath(nodePath),
            getLastSummaryTimestampMs: () => this.messageAtLastSummary?.timestamp,
            readAndParseBlob: async (id) => (0, internal_4.readAndParse)(this.storage, id),
            submitMessage: (message) => this.submit(message),
            sessionExpiryTimerStarted: pendingRuntimeState?.sessionExpiryTimerStarted,
        });
        const loadedFromSequenceNumber = this.deltaManager.initialSequenceNumber;
        // If the base snapshot was generated when isolated channels were disabled, set the summary reference
        // sequence to undefined so that this snapshot will not be used for incremental summaries. This is for
        // back-compat and will rarely happen so its okay to re-summarize everything in the first summary.
        const summaryReferenceSequenceNumber = baseSnapshot === undefined || metadata?.disableIsolatedChannels === true
            ? undefined
            : loadedFromSequenceNumber;
        this.summarizerNode = (0, index_js_4.createRootSummarizerNodeWithGC)((0, internal_8.createChildLogger)({ logger: this.baseLogger, namespace: "SummarizerNode" }), 
        // Summarize function to call when summarize is called. Summarizer node always tracks summary state.
        async (fullTree, trackState, telemetryContext) => this.summarizeInternal(fullTree, trackState, telemetryContext), 
        // Latest change sequence number, no changes since summary applied yet
        loadedFromSequenceNumber, summaryReferenceSequenceNumber, {
            // Must set to false to prevent sending summary handle which would be pointing to
            // a summary with an older protocol state.
            canReuseHandle: false,
            // If GC should not run, let the summarizer node know so that it does not track GC state.
            gcDisabled: !this.garbageCollector.shouldRunGC,
        }, 
        // Function to get GC data if needed. This will always be called by the root summarizer node to get GC data.
        async (fullGC) => this.getGCDataInternal(fullGC), 
        // Function to get the GC details from the base snapshot we loaded from.
        async () => this.garbageCollector.getBaseGCDetails());
        const parentContext = (0, channelCollection_js_1.wrapContext)(this);
        if (snapshotWithContents !== undefined) {
            this.isSnapshotInstanceOfISnapshot = true;
        }
        // Due to a mismatch between different layers in terms of
        // what is the interface of passing signals, we need the
        // downstream stores to wrap the signal.
        parentContext.submitSignal = (type, content, targetClientId) => {
            // Future: Can the `content` argument type be IEnvelope?
            // verifyNotClosed is called in FluidDataStoreContext, which is *the* expected caller.
            const envelope1 = content;
            const envelope2 = createNewSignalEnvelope(envelope1.address, type, envelope1.contents);
            if (targetClientId === undefined) {
                this.signalTelemetryManager.applyTrackingToBroadcastSignalEnvelope(envelope2);
            }
            this.submitSignalFn(envelope2, targetClientId);
        };
        let snapshot = (0, channelCollection_js_1.getSummaryForDatastores)(baseSnapshot, metadata);
        if (snapshot !== undefined && snapshotWithContents !== undefined) {
            snapshot = {
                ...snapshotWithContents,
                snapshotTree: snapshot,
            };
        }
        this.channelCollection = new channelCollection_js_1.ChannelCollection(snapshot, parentContext, this.mc.logger, (props) => this.garbageCollector.nodeUpdated({
            ...props,
            timestampMs: props.timestampMs ?? this.getCurrentReferenceTimestampMs(),
        }), (path) => this.garbageCollector.isNodeDeleted(path), new Map(dataStoreAliasMap), async (runtime) => provideEntryPoint);
        this._deltaManager.on("readonly", this.notifyReadOnlyState);
        this.blobManager = new index_js_1.BlobManager({
            routeContext: this.handleContext,
            blobManagerLoadInfo,
            storage: this.storage,
            sendBlobAttachOp: (localId, blobId) => {
                if (!this.disposed) {
                    this.submit({ type: messageTypes_js_1.ContainerMessageType.BlobAttach, contents: undefined }, undefined, {
                        localId,
                        blobId,
                    });
                }
            },
            blobRequested: (blobPath) => this.garbageCollector.nodeUpdated({
                node: { type: "Blob", path: blobPath },
                reason: "Loaded",
                timestampMs: this.getCurrentReferenceTimestampMs(),
            }),
            isBlobDeleted: (blobPath) => this.garbageCollector.isNodeDeleted(blobPath),
            runtime: this,
            stashedBlobs: pendingRuntimeState?.pendingAttachmentBlobs,
            createBlobPayloadPending: this.sessionSchema.createBlobPayloadPending === true,
        });
        this.deltaScheduler = new deltaScheduler_js_1.DeltaScheduler(this.innerDeltaManager, this, (0, internal_8.createChildLogger)({ logger: this.baseLogger, namespace: "DeltaScheduler" }));
        this.inboundBatchAggregator = new inboundBatchAggregator_js_1.InboundBatchAggregator(this.innerDeltaManager, () => this.clientId, (0, internal_8.createChildLogger)({ logger: this.baseLogger, namespace: "InboundBatchAggregator" }));
        const legacySendBatchFn = (0, exports.makeLegacySendBatchFn)(submitFn, this.innerDeltaManager);
        this.skipSafetyFlushDuringProcessStack =
            // Keep the old flag name even though we renamed the class member (it shipped in 2.31.0)
            this.mc.config.getBoolean("Fluid.ContainerRuntime.DisableFlushBeforeProcess") === true;
        this.outbox = new index_js_3.Outbox({
            shouldSend: () => this.canSendOps(),
            pendingStateManager: this.pendingStateManager,
            submitBatchFn,
            legacySendBatchFn,
            compressor: new index_js_3.OpCompressor(this.mc.logger),
            splitter: opSplitter,
            config: {
                compressionOptions,
                maxBatchSizeInBytes: runtimeOptions.maxBatchSizeInBytes,
                // If we disable flush before process, we must be ready to flush partial batches
                flushPartialBatches: this.skipSafetyFlushDuringProcessStack,
            },
            logger: this.mc.logger,
            groupingManager: opGroupingManager,
            getCurrentSequenceNumbers: () => ({
                // Note: These sequence numbers only change when DeltaManager processes an incoming op
                referenceSequenceNumber: this.deltaManager.lastSequenceNumber,
                clientSequenceNumber: this._processedClientSequenceNumber,
            }),
            reSubmit: this.reSubmit.bind(this),
            opReentrancy: () => this.dataModelChangeRunner.running,
        });
        this._quorum = quorum;
        this._quorum.on("removeMember", (clientId) => {
            this.remoteMessageProcessor.clearPartialMessagesFor(clientId);
        });
        this._audience = audience;
        if (audience.getSelf === undefined) {
            // back-compat, added in 2.0 RC3.
            // Purpose: deal with cases when we run against old loader that does not have newly added capabilities
            audience.getSelf = () => {
                const clientId = this._getClientId();
                return clientId === undefined
                    ? undefined
                    : ({
                        clientId,
                        client: audience.getMember(clientId),
                    });
            };
            let oldClientId = this.clientId;
            this.on("connected", () => {
                const clientId = this.clientId;
                (0, internal_2.assert)(clientId !== undefined, 0x975 /* can't be undefined */);
                audience.emit("selfChanged", { clientId: oldClientId }, { clientId, client: audience.getMember(clientId) });
                oldClientId = clientId;
            });
        }
        const closeSummarizerDelayOverride = this.mc.config.getNumber("Fluid.ContainerRuntime.Test.CloseSummarizerDelayOverrideMs");
        this.closeSummarizerDelayMs =
            closeSummarizerDelayOverride ?? defaultCloseSummarizerDelayMs;
        this.dirtyContainer =
            this.attachState !== container_definitions_1.AttachState.Attached || this.hasPendingMessages();
        context.updateDirtyContainerState(this.dirtyContainer);
        if (!this.skipSafetyFlushDuringProcessStack) {
            // Reference Sequence Number may have just changed, and it must be consistent across a batch,
            // so we should flush now to clear the way for the next ops.
            // NOTE: This will be redundant whenever CR.process was called for the op (since we flush there too) -
            // But we need this coverage for old loaders that don't call ContainerRuntime.process for non-runtime messages.
            // (We have to call flush _before_ processing a runtime op, but after is ok for non-runtime op)
            this.deltaManager.on("op", () => this.flush());
        }
        // logging hardware telemetry
        this.baseLogger.send({
            category: "generic",
            eventName: "DeviceSpec",
            ...getDeviceSpec(),
        });
        this.mc.logger.sendTelemetryEvent({
            eventName: "ContainerLoadStats",
            ...this.createContainerMetadata,
            ...this.channelCollection.containerLoadStats,
            summaryNumber: loadSummaryNumber,
            summaryFormatVersion: metadata?.summaryFormatVersion,
            disableIsolatedChannels: metadata?.disableIsolatedChannels,
            gcVersion: metadata?.gcFeature,
            options: JSON.stringify(runtimeOptions),
            idCompressorModeMetadata: metadata?.documentSchema?.runtime?.idCompressorMode,
            idCompressorMode: this.sessionSchema.idCompressorMode,
            sessionRuntimeSchema: JSON.stringify(this.sessionSchema),
            featureGates: JSON.stringify({
                ...featureGatesForTelemetry,
                closeSummarizerDelayOverride,
                disableFlushBeforeProcess: this.skipSafetyFlushDuringProcessStack,
            }),
            telemetryDocumentId: this.telemetryDocumentId,
            groupedBatchingEnabled: this.groupedBatchingEnabled,
            initialSequenceNumber: this.deltaManager.initialSequenceNumber,
            minVersionForCollab: this.minVersionForCollab,
        });
        (0, connectionTelemetry_js_1.ReportOpPerfTelemetry)(this.clientId, this._deltaManager, this, this.baseLogger);
        (0, batchTracker_js_1.BindBatchTracker)(this, this.baseLogger);
        this.entryPoint = new internal_2.LazyPromise(async () => {
            if (this._summarizer !== undefined) {
                return this._summarizer;
            }
            return provideEntryPoint(this);
        });
        // If we loaded from pending state, then we need to skip any ops that are already accounted in such
        // saved state, i.e. all the ops marked by Loader layer sa savedOp === true.
        this.skipSavedCompressorOps = pendingRuntimeState?.pendingIdCompressorState !== undefined;
    }
    onSchemaChange(schema) {
        this.mc.logger.sendTelemetryEvent({
            eventName: "SchemaChangeAccept",
            sessionRuntimeSchema: JSON.stringify(schema),
        });
        // Most of the settings will be picked up only by new sessions (i.e. after reload).
        // We can make it better in the future (i.e. start to use op compression right away), but for simplicity
        // this is not done.
        // But ID compressor is special. It's possible, that in future, we will remove "stickiness" of ID compressor setting
        // and will allow to start using it. If that were to happen, we want to ensure that we do not break eventual consistency
        // promises. To do so, we need to initialize id compressor right away.
        // As it's implemented right now (with async initialization), this will only work for "off" -> "delayed" transitions.
        // Anything else is too risky, and requires ability to initialize ID compressor synchronously!
        if (schema.runtime.idCompressorMode !== undefined) {
            this.loadIdCompressor();
        }
    }
    getCreateChildSummarizerNodeFn(id, createParam) {
        return (summarizeInternal, getGCDataFn) => this.summarizerNode.createChild(summarizeInternal, id, createParam, undefined, getGCDataFn);
    }
    deleteChildSummarizerNode(id) {
        return this.summarizerNode.deleteChild(id);
    }
    // #region `IFluidParentContext` APIs that should not be called on Root
    makeLocallyVisible() {
        (0, internal_2.assert)(false, 0x8eb /* should not be called */);
    }
    setChannelDirty(address) {
        (0, internal_2.assert)(false, 0x909 /* should not be called */);
    }
    // #endregion
    /**
     * Initializes the state from the base snapshot this container runtime loaded from.
     */
    async initializeBaseState(loader) {
        await this.initializeSummarizer(loader);
        if (this.sessionSchema.idCompressorMode === "on" ||
            (this.sessionSchema.idCompressorMode === "delayed" && this.connected)) {
            this._idCompressor = this.createIdCompressorFn();
            // This is called from loadRuntime(), long before we process any ops, so there should be no ops accumulated yet.
            (0, internal_2.assert)(this.pendingIdCompressorOps.length === 0, 0x8ec /* no pending ops */);
        }
        await this.garbageCollector.initializeBaseState();
    }
    async initializeSummarizer(loader) {
        if (this.summariesDisabled) {
            this.mc.logger.sendTelemetryEvent({ eventName: "SummariesDisabled" });
            return;
        }
        const { maxOpsSinceLastSummary = 0, initialSummarizerDelayMs = 0 } = (0, index_js_4.isSummariesDisabled)(this.summaryConfiguration)
            ? {}
            : {
                ...this.summaryConfiguration,
                initialSummarizerDelayMs: 
                // back-compat: initialSummarizerDelayMs was moved from ISummaryRuntimeOptions
                //   to ISummaryConfiguration in 0.60.
                this.runtimeOptions.summaryOptions.initialSummarizerDelayMs ??
                    this.summaryConfiguration.initialSummarizerDelayMs,
            };
        const summaryCollection = new index_js_4.SummaryCollection(this.deltaManager, this.baseLogger);
        const orderedClientLogger = (0, internal_8.createChildLogger)({
            logger: this.baseLogger,
            namespace: "OrderedClientElection",
        });
        const orderedClientCollection = new index_js_4.OrderedClientCollection(orderedClientLogger, this.innerDeltaManager, this._quorum);
        const orderedClientElectionForSummarizer = new index_js_4.OrderedClientElection(orderedClientLogger, orderedClientCollection, this.electedSummarizerData ?? this.innerDeltaManager.lastSequenceNumber, index_js_4.SummarizerClientElection.isClientEligible, this.mc.config.getBoolean("Fluid.ContainerRuntime.OrderedClientElection.EnablePerformanceEvents"));
        this.summarizerClientElection = new index_js_4.SummarizerClientElection(orderedClientLogger, summaryCollection, orderedClientElectionForSummarizer, maxOpsSinceLastSummary);
        const isSummarizerClient = this.clientDetails.type === index_js_4.summarizerClientType;
        if (isSummarizerClient) {
            // We want to dynamically import any thing inside summaryDelayLoadedModule module only when we are the summarizer client,
            // so that all non summarizer clients don't have to load the code inside this module.
            const module = await import(
            /* webpackChunkName: "summarizerDelayLoadedModule" */ "./summary/index.js");
            this._summarizer = new module.Summarizer(this /* ISummarizerRuntime */, () => this.summaryConfiguration, this /* ISummarizerInternalsProvider */, this.handleContext, summaryCollection, async (runtime) => module.RunWhileConnectedCoordinator.create(runtime, 
            // Summarization runs in summarizer client and needs access to the real (non-proxy) active
            // information. The proxy delta manager would always return false for summarizer client.
            () => this.innerDeltaManager.active));
        }
        else if (index_js_4.SummarizerClientElection.clientDetailsPermitElection(this.clientDetails)) {
            // Only create a SummaryManager and SummarizerClientElection
            // if summaries are enabled and we are not the summarizer client.
            const defaultAction = () => {
                if (summaryCollection.opsSinceLastAck > maxOpsSinceLastSummary) {
                    this.mc.logger.sendTelemetryEvent({ eventName: "SummaryStatus:Behind" });
                    // unregister default to no log on every op after falling behind
                    // and register summary ack handler to re-register this handler
                    // after successful summary
                    summaryCollection.once(internal_3.MessageType.SummaryAck, () => {
                        this.mc.logger.sendTelemetryEvent({
                            eventName: "SummaryStatus:CaughtUp",
                        });
                        // we've caught up, so re-register the default action to monitor for
                        // falling behind, and unregister ourself
                        summaryCollection.on("default", defaultAction);
                    });
                    summaryCollection.off("default", defaultAction);
                }
            };
            summaryCollection.on("default", defaultAction);
            // Create the SummaryManager and mark the initial state
            this.summaryManager = new index_js_4.SummaryManager(this.summarizerClientElection, this, // IConnectedState
            summaryCollection, this.baseLogger, (0, index_js_4.formCreateSummarizerFn)(loader), new throttler_js_1.Throttler(60 * 1000, // 60 sec delay window
            30 * 1000, // 30 sec max delay
            // throttling function increases exponentially (0ms, 40ms, 80ms, 160ms, etc)
            (0, throttler_js_1.formExponentialFn)({ coefficient: 20, initialDelay: 0 })), {
                initialDelayMs: initialSummarizerDelayMs,
            });
            // Forward events from SummaryManager
            for (const eventName of [
                "summarize",
                "summarizeAllAttemptsFailed",
                "summarizerStop",
                "summarizerStart",
                "summarizerStartupFailed",
            ]) {
                this.summaryManager?.on(eventName, (...args) => {
                    this.emit(eventName, ...args);
                });
            }
            this.summaryManager.start();
        }
    }
    dispose(error) {
        if (this._disposed) {
            return;
        }
        this._disposed = true;
        this.mc.logger.sendTelemetryEvent({
            eventName: "ContainerRuntimeDisposed",
            isDirty: this.isDirty,
            lastSequenceNumber: this.deltaManager.lastSequenceNumber,
            attachState: this.attachState,
        }, error);
        if (this.summaryManager !== undefined) {
            this.summaryManager.dispose();
        }
        this.garbageCollector.dispose();
        this._summarizer?.dispose();
        this.channelCollection.dispose();
        this.pendingStateManager.dispose();
        this.inboundBatchAggregator.dispose();
        this.deltaScheduler.dispose();
        this._deltaManager.dispose();
        this.emit("dispose");
        this.removeAllListeners();
    }
    /**
     * Api to fetch the snapshot from the service for a loadingGroupIds.
     * @param loadingGroupIds - LoadingGroupId for which the snapshot is asked for.
     * @param pathParts - Parts of the path, which we want to extract from the snapshot tree.
     * @returns - snapshotTree and the sequence number of the snapshot.
     */
    async getSnapshotForLoadingGroupId(loadingGroupIds, pathParts) {
        const sortedLoadingGroupIds = loadingGroupIds.sort();
        (0, internal_2.assert)(this.storage.getSnapshot !== undefined, 0x8ed /* getSnapshot api should be defined if used */);
        let loadedFromCache = true;
        // Lookup up in the cache, if not present then make the network call as multiple datastores could
        // be in same loading group. So, once we have fetched the snapshot for that loading group on
        // any request, then cache that as same group could be requested in future too.
        const snapshot = await this.snapshotCacheForLoadingGroupIds.addOrGet(sortedLoadingGroupIds.join(","), async () => {
            (0, internal_2.assert)(this.storage.getSnapshot !== undefined, 0x8ee /* getSnapshot api should be defined if used */);
            loadedFromCache = false;
            return this.storage.getSnapshot({
                cacheSnapshot: false,
                scenarioName: "snapshotForLoadingGroupId",
                loadingGroupIds: sortedLoadingGroupIds,
            });
        });
        this.mc.logger.sendTelemetryEvent({
            eventName: "GroupIdSnapshotFetched",
            details: JSON.stringify({
                fromCache: loadedFromCache,
                loadingGroupIds: loadingGroupIds.join(","),
            }),
        });
        // Find the snapshotTree inside the returned snapshot based on the path as given in the request.
        const hasIsolatedChannels = (0, index_js_4.rootHasIsolatedChannels)(this.metadata);
        const snapshotTreeForPath = this.getSnapshotTreeForPath(snapshot.snapshotTree, pathParts, hasIsolatedChannels);
        (0, internal_2.assert)(snapshotTreeForPath !== undefined, 0x8ef /* no snapshotTree for the path */);
        const snapshotSeqNumber = snapshot.sequenceNumber;
        (0, internal_2.assert)(snapshotSeqNumber !== undefined, 0x8f0 /* snapshotSeqNumber should be present */);
        // This assert fires if we get a snapshot older than the snapshot we loaded from. This is a service issue.
        // Snapshots should only move forward. If we observe an older snapshot than the one we loaded from, then likely
        // the file has been overwritten or service lost data.
        if (snapshotSeqNumber < this.deltaManager.initialSequenceNumber) {
            throw internal_8.DataProcessingError.create("Downloaded snapshot older than snapshot we loaded from", "getSnapshotForLoadingGroupId", undefined, {
                loadingGroupIds: sortedLoadingGroupIds.join(","),
                snapshotSeqNumber,
                initialSequenceNumber: this.deltaManager.initialSequenceNumber,
            });
        }
        // If the snapshot is ahead of the last seq number of the delta manager, then catch up before
        // returning the snapshot.
        if (snapshotSeqNumber > this.deltaManager.lastSequenceNumber) {
            // If this is a summarizer client, which is trying to load a group and it finds that there is
            // another snapshot from which the summarizer loaded and it is behind, then just give up as
            // the summarizer state is not up to date.
            // This should be a recoverable scenario and shouldn't happen as we should process the ack first.
            if (this._summarizer !== undefined) {
                throw new Error("Summarizer client behind, loaded newer snapshot with loadingGroupId");
            }
            // We want to catchup from sequenceNumber to targetSequenceNumber
            const props = {
                eventName: "GroupIdSnapshotCatchup",
                loadingGroupIds: sortedLoadingGroupIds.join(","),
                targetSequenceNumber: snapshotSeqNumber, // This is so we reuse some columns in telemetry
                sequenceNumber: this.deltaManager.lastSequenceNumber, // This is so we reuse some columns in telemetry
            };
            const event = internal_8.PerformanceEvent.start(this.mc.logger, {
                ...props,
            });
            // If the inbound deltas queue is paused or disconnected, we expect a reconnect and unpause
            // as long as it's not a summarizer client.
            if (this._deltaManager.inbound.paused) {
                props.inboundPaused = this._deltaManager.inbound.paused; // reusing telemetry
            }
            const defP = new internal_2.Deferred();
            this.deltaManager.on("op", (message) => {
                if (message.sequenceNumber >= snapshotSeqNumber) {
                    defP.resolve(true);
                }
            });
            await defP.promise;
            event.end(props);
        }
        return { snapshotTree: snapshotTreeForPath, sequenceNumber: snapshotSeqNumber };
    }
    /**
     * Api to find a snapshot tree inside a bigger snapshot tree based on the path in the pathParts array.
     * @param snapshotTree - snapshot tree to look into.
     * @param pathParts - Part of the path, which we want to extract from the snapshot tree.
     * @param hasIsolatedChannels - whether the channels are present inside ".channels" subtree. Older
     * snapshots will not have trees inside ".channels", so check that.
     * @returns - requested snapshot tree based on the path parts.
     */
    getSnapshotTreeForPath(snapshotTree, pathParts, hasIsolatedChannels) {
        let childTree = snapshotTree;
        for (const part of pathParts) {
            if (hasIsolatedChannels) {
                childTree = childTree?.trees[internal_6.channelsTreeName];
            }
            childTree = childTree?.trees[part];
        }
        return childTree;
    }
    /**
     * Notifies this object about the request made to the container.
     * @param request - Request made to the handler.
     * @deprecated Will be removed in future major release. This method needs to stay private until LTS version of Loader moves to "2.0.0-internal.7.0.0".
     */
    // @ts-expect-error expected to be used by LTS Loaders and Containers
    async request(request) {
        try {
            const parser = internal_7.RequestParser.create(request);
            const id = parser.pathParts[0];
            if (id === index_js_4.summarizerRequestUrl && parser.pathParts.length === 1) {
                if (this._summarizer !== undefined) {
                    return {
                        status: 200,
                        mimeType: "fluid/object",
                        value: this._summarizer,
                    };
                }
                return (0, internal_7.create404Response)(request);
            }
            if (this.requestHandler !== undefined) {
                // eslint-disable-next-line @typescript-eslint/return-await -- Adding an await here causes test failures
                return this.requestHandler(parser, this);
            }
            return (0, internal_7.create404Response)(request);
        }
        catch (error) {
            return (0, internal_7.exceptionToResponse)(error);
        }
    }
    /**
     * Resolves URI representing handle
     * @param request - Request made to the handler.
     */
    async resolveHandle(request) {
        try {
            const requestParser = internal_7.RequestParser.create(request);
            const id = requestParser.pathParts[0];
            if (id === "_channels") {
                // eslint-disable-next-line @typescript-eslint/return-await -- Adding an await here causes test failures
                return this.resolveHandle(requestParser.createSubRequest(1));
            }
            if (id === index_js_1.blobManagerBasePath && requestParser.isLeaf(2)) {
                const localId = requestParser.pathParts[1];
                const payloadPending = requestParser.headers?.[internal_7.RuntimeHeaders.payloadPending] === true;
                if (!this.blobManager.hasBlob(localId) &&
                    requestParser.headers?.[internal_7.RuntimeHeaders.wait] === false) {
                    return (0, internal_7.create404Response)(request);
                }
                const blob = await this.blobManager.getBlob(localId, payloadPending);
                return {
                    status: 200,
                    mimeType: "fluid/object",
                    value: blob,
                };
            }
            else if (requestParser.pathParts.length > 0) {
                return await this.channelCollection.request(request);
            }
            return (0, internal_7.create404Response)(request);
        }
        catch (error) {
            return (0, internal_7.exceptionToResponse)(error);
        }
    }
    /**
     * {@inheritDoc @fluidframework/container-definitions#IRuntime.getEntryPoint}
     */
    async getEntryPoint() {
        return this.entryPoint;
    }
    internalId(maybeAlias) {
        return this.channelCollection.internalId(maybeAlias);
    }
    /**
     * Adds the container's metadata to the given summary tree.
     */
    addMetadataToSummary(summaryTree) {
        // The last message processed at the time of summary. If there are no new messages, use the message from the
        // last summary.
        const message = (0, index_js_4.extractSummaryMetadataMessage)(this.deltaManager.lastMessage) ??
            this.messageAtLastSummary;
        const documentSchema = this.documentsSchemaController.summarizeDocumentSchema(this.deltaManager.lastSequenceNumber);
        // Is document schema explicit control on?
        const explicitSchemaControl = documentSchema?.runtime.explicitSchemaControl;
        const metadata = {
            ...this.createContainerMetadata,
            // Increment the summary number for the next summary that will be generated.
            summaryNumber: this.nextSummaryNumber++,
            summaryFormatVersion: 1,
            ...this.garbageCollector.getMetadata(),
            telemetryDocumentId: this.telemetryDocumentId,
            // If explicit document schema control is not on, use legacy way to supply last message (using 'message' property).
            // Otherwise use new 'lastMessage' property, but also put content into the 'message' property that cases old
            // runtimes (that preceed document schema control capabilities) to close container on load due to mismatch in
            // last message's sequence number.
            // See also lastMessageFromMetadata()
            message: explicitSchemaControl
                ? { sequenceNumber: -1 }
                : message,
            lastMessage: explicitSchemaControl ? message : undefined,
            documentSchema,
        };
        (0, internal_7.addBlobToSummary)(summaryTree, index_js_4.metadataBlobName, JSON.stringify(metadata));
    }
    addContainerStateToSummary(summaryTree, fullTree, trackState, telemetryContext) {
        this.addMetadataToSummary(summaryTree);
        if (this._idCompressor) {
            const idCompressorState = JSON.stringify(this._idCompressor.serialize(false));
            (0, internal_7.addBlobToSummary)(summaryTree, index_js_4.idCompressorBlobName, idCompressorState);
        }
        if (this.remoteMessageProcessor.partialMessages.size > 0) {
            const content = JSON.stringify([...this.remoteMessageProcessor.partialMessages]);
            (0, internal_7.addBlobToSummary)(summaryTree, index_js_4.chunksBlobName, content);
        }
        const recentBatchInfo = this.duplicateBatchDetector?.getRecentBatchInfoForSummary(telemetryContext);
        if (recentBatchInfo !== undefined) {
            (0, internal_7.addBlobToSummary)(summaryTree, index_js_4.recentBatchInfoBlobName, JSON.stringify(recentBatchInfo));
        }
        const dataStoreAliases = this.channelCollection.aliases;
        if (dataStoreAliases.size > 0) {
            (0, internal_7.addBlobToSummary)(summaryTree, index_js_4.aliasBlobName, JSON.stringify([...dataStoreAliases]));
        }
        if (this.summarizerClientElection) {
            const electedSummarizerContent = JSON.stringify(this.summarizerClientElection?.serialize());
            (0, internal_7.addBlobToSummary)(summaryTree, index_js_4.electedSummarizerBlobName, electedSummarizerContent);
        }
        const blobManagerSummary = this.blobManager.summarize();
        // Some storage (like git) doesn't allow empty tree, so we can omit it.
        // and the blob manager can handle the tree not existing when loading
        if (Object.keys(blobManagerSummary.summary.tree).length > 0) {
            (0, internal_7.addSummarizeResultToSummary)(summaryTree, index_js_1.blobsTreeName, blobManagerSummary);
        }
        const gcSummary = this.garbageCollector.summarize(fullTree, trackState, telemetryContext);
        if (gcSummary !== undefined) {
            (0, internal_7.addSummarizeResultToSummary)(summaryTree, internal_6.gcTreeKey, gcSummary);
        }
    }
    // Track how many times the container tries to reconnect with pending messages.
    // This happens when the connection state is changed and we reset the counter
    // when we are able to process a local op or when there are no pending messages.
    // If this counter reaches a max, it's a good indicator that the container
    // is not making progress and it is stuck in a retry loop.
    shouldContinueReconnecting() {
        if (this.maxConsecutiveReconnects <= 0) {
            // Feature disabled, we never stop reconnecting
            return true;
        }
        if (!this.hasPendingMessages()) {
            // If there are no pending messages, we can always reconnect
            this.resetReconnectCount();
            return true;
        }
        if (this.consecutiveReconnects === Math.floor(this.maxConsecutiveReconnects / 2)) {
            // If we're halfway through the max reconnects, send an event in order
            // to better identify false positives, if any. If the rate of this event
            // matches Container Close count below, we can safely cut down
            // maxConsecutiveReconnects to half.
            this.mc.logger.sendTelemetryEvent({
                eventName: "ReconnectsWithNoProgress",
                attempts: this.consecutiveReconnects,
                pendingMessages: this.pendingMessagesCount,
            });
        }
        return this.consecutiveReconnects < this.maxConsecutiveReconnects;
    }
    resetReconnectCount() {
        this.consecutiveReconnects = 0;
    }
    replayPendingStates() {
        // We need to be able to send ops to replay states
        if (!this.canSendOps()) {
            return;
        }
        // We need to temporary clear the dirty flags and disable
        // dirty state change events to detect whether replaying ops
        // has any effect.
        // Save the old state, reset to false, disable event emit
        const oldState = this.dirtyContainer;
        this.dirtyContainer = false;
        (0, internal_2.assert)(this.emitDirtyDocumentEvent, 0x127 /* "dirty document event not set on replay" */);
        this.emitDirtyDocumentEvent = false;
        let newState;
        try {
            this.submitIdAllocationOpIfNeeded(true);
            // replay the ops
            this.pendingStateManager.replayPendingStates();
        }
        finally {
            // Save the new start and restore the old state, re-enable event emit
            newState = this.dirtyContainer;
            this.dirtyContainer = oldState;
            this.emitDirtyDocumentEvent = true;
        }
        // Officially transition from the old state to the new state.
        this.updateDocumentDirtyState(newState);
    }
    /**
     * Parse an op's type and actual content from given serialized content
     * ! Note: this format needs to be in-line with what is set in the "ContainerRuntime.submit(...)" method
     */
    parseLocalOpContent(serializedContents) {
        (0, internal_2.assert)(serializedContents !== undefined, 0x6d5 /* content must be defined */);
        const message = JSON.parse(serializedContents);
        (0, internal_2.assert)(message.type !== undefined, 0x6d6 /* incorrect op content format */);
        return message;
    }
    async applyStashedOp(serializedOpContent) {
        // Pending State contains serialized contents, so parse it here.
        const opContents = this.parseLocalOpContent(serializedOpContent);
        switch (opContents.type) {
            case messageTypes_js_1.ContainerMessageType.FluidDataStoreOp:
            case messageTypes_js_1.ContainerMessageType.Attach:
            case messageTypes_js_1.ContainerMessageType.Alias: {
                return this.channelCollection.applyStashedOp(opContents);
            }
            case messageTypes_js_1.ContainerMessageType.IdAllocation: {
                // IDs allocation ops in stashed state are ignored because the tip state of the compressor
                // is serialized into the pending state. This is done because generation of new IDs during
                // stashed op application (or, later, resubmit) must generate new IDs and if the compressor
                // was loaded from a state serialized at the same time as the summary tree in the stashed state
                // then it would generate IDs that collide with any in later stashed ops.
                // In the future, IdCompressor could be extended to have an "applyStashedOp" or similar method
                // and the runtime could filter out all ID allocation ops from the stashed state and apply them
                // before applying the rest of the stashed ops. This would accomplish the same thing but with
                // better performance in future incremental stashed state creation.
                (0, internal_2.assert)(this.sessionSchema.idCompressorMode !== undefined, 0x8f1 /* ID compressor should be in use */);
                return;
            }
            case messageTypes_js_1.ContainerMessageType.DocumentSchemaChange: {
                return;
            }
            case messageTypes_js_1.ContainerMessageType.BlobAttach: {
                return;
            }
            case messageTypes_js_1.ContainerMessageType.Rejoin: {
                throw new Error("rejoin not expected here");
            }
            case messageTypes_js_1.ContainerMessageType.GC: {
                // GC op is only sent in summarizer which should never have stashed ops.
                throw new internal_8.LoggingError("GC op not expected to be stashed in summarizer");
            }
            default: {
                const error = getUnknownMessageTypeError(opContents.type, "applyStashedOp" /* codePath */);
                this.closeFn(error);
                throw error;
            }
        }
    }
    loadIdCompressor() {
        if (this._idCompressor === undefined &&
            this.sessionSchema.idCompressorMode !== undefined) {
            this._idCompressor = this.createIdCompressorFn();
            // Finalize any ranges we received while the compressor was turned off.
            const ops = this.pendingIdCompressorOps;
            this.pendingIdCompressorOps = [];
            for (const range of ops) {
                this._idCompressor.finalizeCreationRange(range);
            }
            (0, internal_2.assert)(this.pendingIdCompressorOps.length === 0, 0x976 /* No new ops added */);
        }
    }
    setConnectionState(connected, clientId) {
        // Validate we have consistent state
        const currentClientId = this._audience.getSelf()?.clientId;
        (0, internal_2.assert)(clientId === currentClientId, 0x977 /* input clientId does not match Audience */);
        (0, internal_2.assert)(this.clientId === currentClientId, 0x978 /* this.clientId does not match Audience */);
        if (connected && this.sessionSchema.idCompressorMode === "delayed") {
            this.loadIdCompressor();
        }
        if (connected === false && this.delayConnectClientId !== undefined) {
            this.delayConnectClientId = undefined;
            this.mc.logger.sendTelemetryEvent({
                eventName: "UnsuccessfulConnectedTransition",
            });
            // Don't propagate "disconnected" event because we didn't propagate the previous "connected" event
            return;
        }
        if (!connected) {
            this.documentsSchemaController.onDisconnect();
        }
        // If there are stashed blobs in the pending state, we need to delay
        // propagation of the "connected" event until we have uploaded them to
        // ensure we don't submit ops referencing a blob that has not been uploaded
        const connecting = connected && !this._connected;
        if (connecting && this.blobManager.hasPendingStashedUploads()) {
            (0, internal_2.assert)(!this.delayConnectClientId, 0x791 /* Connect event delay must be canceled before subsequent connect event */);
            (0, internal_2.assert)(!!clientId, 0x792 /* Must have clientId when connecting */);
            this.delayConnectClientId = clientId;
            return;
        }
        this.setConnectionStateCore(connected, clientId);
    }
    setConnectionStateCore(connected, clientId) {
        (0, internal_2.assert)(!this.delayConnectClientId, 0x394 /* connect event delay must be cleared before propagating connect event */);
        this.verifyNotClosed();
        // There might be no change of state due to Container calling this API after loading runtime.
        const changeOfState = this._connected !== connected;
        const reconnection = changeOfState && !connected;
        // We need to flush the ops currently collected by Outbox to preserve original order.
        // This flush NEEDS to happen before we set the ContainerRuntime to "connected".
        // We want these ops to get to the PendingStateManager without sending to service and have them return to the Outbox upon calling "replayPendingStates".
        if (changeOfState && connected) {
            this.flush();
        }
        this._connected = connected;
        if (connected) {
            (0, internal_2.assert)(this.attachState === container_definitions_1.AttachState.Attached, 0x3cd /* Connection is possible only if container exists in storage */);
            if (changeOfState) {
                this.signalTelemetryManager.resetTracking();
            }
        }
        // Fail while disconnected
        if (reconnection) {
            this.consecutiveReconnects++;
            if (!this.shouldContinueReconnecting()) {
                this.closeFn(internal_8.DataProcessingError.create("Runtime detected too many reconnects with no progress syncing local ops.", "setConnectionState", undefined, {
                    dataLoss: 1,
                    attempts: this.consecutiveReconnects,
                    pendingMessages: this.pendingMessagesCount,
                }));
                return;
            }
        }
        if (changeOfState) {
            this.replayPendingStates();
        }
        this.channelCollection.setConnectionState(connected, clientId);
        this.garbageCollector.setConnectionState(connected, clientId);
        (0, internal_8.raiseConnectedEvent)(this.mc.logger, this, connected, clientId);
    }
    async notifyOpReplay(message) {
        await this.pendingStateManager.applyStashedOpsAt(message.sequenceNumber);
    }
    /**
     * Processes the op.
     * @param messageCopy - Sequenced message for a distributed document.
     * @param local - true if the message was originally generated by the client receiving it.
     */
    process({ ...messageCopy }, local) {
        // spread operator above ensure we make a shallow copy of message, as the processing flow will modify it.
        // There might be multiple container instances receiving the same message.
        this.verifyNotClosed();
        if (!this.skipSafetyFlushDuringProcessStack) {
            // Reference Sequence Number may be about to change, and it must be consistent across a batch, so flush now
            this.flush();
        }
        this.ensureNoDataModelChanges(() => {
            this.processInboundMessageOrBatch(messageCopy, local);
        });
    }
    /**
     * Implementation of core logic for {@link ContainerRuntime.process}, once preconditions are established
     *
     * @param messageCopy - Shallow copy of the sequenced message. If it's a virtualized batch, we'll process
     * all messages in the batch here.
     */
    processInboundMessageOrBatch(messageCopy, local) {
        // Whether or not the message appears to be a runtime message from an up-to-date client.
        // It may be a legacy runtime message (ie already unpacked and ContainerMessageType)
        // or something different, like a system message.
        const hasModernRuntimeMessageEnvelope = messageCopy.type === internal_3.MessageType.Operation;
        const savedOp = messageCopy.metadata?.savedOp;
        const logLegacyCase = (0, exports.getSingleUseLegacyLogCallback)(this.mc.logger, messageCopy.type);
        let runtimeBatch = hasModernRuntimeMessageEnvelope || isUnpackedRuntimeMessage(messageCopy);
        if (runtimeBatch) {
            // We expect runtime messages to have JSON contents - deserialize it in place.
            (0, index_js_3.ensureContentsDeserialized)(messageCopy);
        }
        if (hasModernRuntimeMessageEnvelope) {
            // If the message has the modern message envelope, then process it here.
            // Here we unpack the message (decompress, unchunk, and/or ungroup) into a batch of messages with ContainerMessageType
            const inboundResult = this.remoteMessageProcessor.process(messageCopy, logLegacyCase);
            if (inboundResult === undefined) {
                // This means the incoming message is an incomplete part of a message or batch
                // and we need to process more messages before the rest of the system can understand it.
                return;
            }
            if ("batchStart" in inboundResult) {
                const batchStart = inboundResult.batchStart;
                const result = this.duplicateBatchDetector?.processInboundBatch(batchStart);
                if (result?.duplicate) {
                    const error = new internal_8.DataCorruptionError("Duplicate batch - The same batch was sequenced twice", { batchId: batchStart.batchId });
                    this.mc.logger.sendTelemetryEvent({
                        eventName: "DuplicateBatch",
                        details: {
                            batchId: batchStart.batchId,
                            clientId: batchStart.clientId,
                            batchStartCsn: batchStart.batchStartCsn,
                            size: inboundResult.length,
                            duplicateBatchSequenceNumber: result.otherSequenceNumber,
                            ...(0, internal_8.extractSafePropertiesFromMessage)(batchStart.keyMessage),
                        },
                    }, error);
                    throw error;
                }
            }
            // Reach out to PendingStateManager, either to zip localOpMetadata into the *local* message list,
            // or to check to ensure the *remote* messages don't match the batchId of a pending local batch.
            // This latter case would indicate that the container has forked - two copies are trying to persist the same local changes.
            let messagesWithPendingState = this.pendingStateManager.processInboundMessages(inboundResult, local);
            if (inboundResult.type !== "fullBatch") {
                (0, internal_2.assert)(messagesWithPendingState.length === 1, 0xa3d /* Partial batch should have exactly one message */);
            }
            if (messagesWithPendingState.length === 0) {
                (0, internal_2.assert)(inboundResult.type === "fullBatch", 0xa3e /* Empty batch is always considered a full batch */);
                /**
                 * We need to process an empty batch, which will execute expected actions while processing even if there
                 * are no inner runtime messages.
                 *
                 * Empty batches are produced by the outbox on resubmit when the resubmit flow resulted in no runtime
                 * messages.
                 * This can happen if changes from a remote client "cancel out" the pending changes being resubmitted by
                 * this client.  We submit an empty batch if "offline load" (aka rehydrating from stashed state) is
                 * enabled, to ensure we account for this batch when comparing batchIds, checking for a forked container.
                 * Otherwise, we would not realize this container has forked in the case where it did fork, and a batch
                 * became empty but wasn't submitted as such.
                 */
                messagesWithPendingState = [
                    {
                        message: inboundResult.batchStart.keyMessage,
                        localOpMetadata: undefined,
                    },
                ];
                // Empty batch message is a non-runtime message as it was generated by the op grouping manager.
                runtimeBatch = false;
            }
            const locationInBatch = inboundResult.type === "fullBatch"
                ? { batchStart: true, batchEnd: true }
                : inboundResult.type === "batchStartingMessage"
                    ? { batchStart: true, batchEnd: false }
                    : { batchStart: false, batchEnd: inboundResult.batchEnd === true };
            this.processInboundMessages(messagesWithPendingState, locationInBatch, local, savedOp, runtimeBatch, inboundResult.type === "fullBatch"
                ? inboundResult.groupedBatch
                : false /* groupedBatch */);
        }
        else {
            this.processInboundMessages([{ message: messageCopy, localOpMetadata: undefined }], { batchStart: true, batchEnd: true }, // Single message
            local, savedOp, runtimeBatch, false /* groupedBatch */);
        }
        if (local) {
            // If we have processed a local op, this means that the container is
            // making progress and we can reset the counter for how many times
            // we have consecutively replayed the pending states
            this.resetReconnectCount();
        }
    }
    /**
     * Processes inbound message(s). It calls delta scheduler according to the messages' location in the batch.
     * @param messagesWithMetadata - messages to process along with their metadata.
     * @param locationInBatch - Are we processing the start and/or end of a batch?
     * @param local - true if the messages were originally generated by the client receiving it.
     * @param savedOp - true if the message is a replayed saved op.
     * @param runtimeBatch - true if these are runtime messages.
     * @param groupedBatch - true if these messages are part of a grouped op batch.
     */
    processInboundMessages(messagesWithMetadata, locationInBatch, local, savedOp, runtimeBatch, groupedBatch) {
        if (locationInBatch.batchStart) {
            const firstMessage = messagesWithMetadata[0]?.message;
            (0, internal_2.assert)(firstMessage !== undefined, 0xa31 /* Batch must have at least one message */);
            this.emit("batchBegin", firstMessage);
        }
        let error;
        try {
            if (!runtimeBatch) {
                for (const { message } of messagesWithMetadata) {
                    this.observeNonRuntimeMessage(message);
                }
                return;
            }
            // Updates a message's minimum sequence number to the minimum sequence number that container
            // runtime is tracking and sets _processedClientSequenceNumber. It returns the updated message.
            const updateSequenceNumbers = (message) => {
                // Set the minimum sequence number to the containerRuntime's understanding of minimum sequence number.
                message.minimumSequenceNumber =
                    this.useDeltaManagerOpsProxy &&
                        this.deltaManager.minimumSequenceNumber < message.minimumSequenceNumber
                        ? this.deltaManager.minimumSequenceNumber
                        : message.minimumSequenceNumber;
                this._processedClientSequenceNumber = message.clientSequenceNumber;
                return message;
            };
            // Non-grouped batch messages are processed one at a time.
            if (!groupedBatch) {
                for (const { message, localOpMetadata } of messagesWithMetadata) {
                    updateSequenceNumbers(message);
                    this.validateAndProcessRuntimeMessages(message, [
                        {
                            contents: message.contents,
                            localOpMetadata,
                            clientSequenceNumber: message.clientSequenceNumber,
                        },
                    ], local, savedOp);
                    this.emit("op", message, true /* runtimeMessage */);
                }
                return;
            }
            let bunchedMessagesContent = [];
            let previousMessage;
            // Process the previous bunch of messages.
            const processBunchedMessages = () => {
                (0, internal_2.assert)(previousMessage !== undefined, 0xa67 /* previous message must exist */);
                this.validateAndProcessRuntimeMessages(previousMessage, bunchedMessagesContent, local, savedOp);
                bunchedMessagesContent = [];
            };
            /**
             * For grouped batch messages, bunch contiguous messages of the same type and process them together.
             * This is an optimization mainly for DDSes, where it can process a bunch of ops together. DDSes
             * like merge tree or shared tree can process ops more efficiently when they are bunched together.
             */
            for (const { message, localOpMetadata } of messagesWithMetadata) {
                const currentMessage = updateSequenceNumbers(message);
                if (previousMessage && previousMessage.type !== currentMessage.type) {
                    processBunchedMessages();
                }
                previousMessage = currentMessage;
                bunchedMessagesContent.push({
                    contents: message.contents,
                    localOpMetadata,
                    clientSequenceNumber: message.clientSequenceNumber,
                });
            }
            // Process the last bunch of messages.
            processBunchedMessages();
            // Send the "op" events for the messages now that the ops have been processed.
            for (const { message } of messagesWithMetadata) {
                this.emit("op", message, true /* runtimeMessage */);
            }
        }
        catch (error_) {
            error = error_;
            throw error;
        }
        finally {
            if (locationInBatch.batchEnd) {
                const lastMessage = messagesWithMetadata[messagesWithMetadata.length - 1]?.message;
                (0, internal_2.assert)(lastMessage !== undefined, 0xa32 /* Batch must have at least one message */);
                this.emit("batchEnd", error, lastMessage);
            }
        }
    }
    /**
     * Observes messages that are not intended for the runtime layer, updating/notifying Runtime systems as needed.
     * @param message - non-runtime message to process.
     */
    observeNonRuntimeMessage(message) {
        // Set the minimum sequence number to the containerRuntime's understanding of minimum sequence number.
        if (this.deltaManager.minimumSequenceNumber < message.minimumSequenceNumber) {
            message.minimumSequenceNumber = this.deltaManager.minimumSequenceNumber;
        }
        this._processedClientSequenceNumber = message.clientSequenceNumber;
        // If there are no more pending messages after processing a local message,
        // the document is no longer dirty.
        if (!this.hasPendingMessages()) {
            this.updateDocumentDirtyState(false);
        }
        // The DeltaManager used to do this, but doesn't anymore as of Loader v2.4
        // Anyone listening to our "op" event would expect the contents to be parsed per this same logic
        if (typeof message.contents === "string" &&
            message.contents !== "" &&
            message.type !== internal_3.MessageType.ClientLeave) {
            message.contents = JSON.parse(message.contents);
        }
        this.emit("op", message, false /* runtimeMessage */);
    }
    /**
     * Process runtime messages. The messages here are contiguous messages in a batch.
     * Assuming the messages in the given bunch are also a TypedContainerRuntimeMessage, checks its type and dispatch
     * the messages to the appropriate handler in the runtime.
     * Throws a DataProcessingError if the message looks like but doesn't conform to a known TypedContainerRuntimeMessage type.
     * @param message - The core message with common properties for all the messages.
     * @param messageContents - The contents, local metadata and clientSequenceNumbers of the messages.
     * @param local - true if the messages were originally generated by the client receiving it.
     * @param savedOp - true if the message is a replayed saved op.
     *
     */
    validateAndProcessRuntimeMessages(message, messagesContent, local, savedOp) {
        // If there are no more pending messages after processing a local message,
        // the document is no longer dirty.
        if (!this.hasPendingMessages()) {
            this.updateDocumentDirtyState(false);
        }
        // Get the contents without the localOpMetadata because not all message types know about localOpMetadata.
        const contents = messagesContent.map((c) => c.contents);
        switch (message.type) {
            case messageTypes_js_1.ContainerMessageType.FluidDataStoreOp:
            case messageTypes_js_1.ContainerMessageType.Attach:
            case messageTypes_js_1.ContainerMessageType.Alias: {
                // Remove the metadata from the message before sending it to the channel collection. The metadata
                // is added by the container runtime and is not part of the message that the channel collection and
                // layers below it expect.
                this.channelCollection.processMessages({ envelope: message, messagesContent, local });
                break;
            }
            case messageTypes_js_1.ContainerMessageType.BlobAttach: {
                this.blobManager.processBlobAttachMessage(message, local);
                break;
            }
            case messageTypes_js_1.ContainerMessageType.IdAllocation: {
                this.processIdCompressorMessages(contents, savedOp);
                break;
            }
            case messageTypes_js_1.ContainerMessageType.GC: {
                this.garbageCollector.processMessages(contents, message.timestamp, local);
                break;
            }
            case messageTypes_js_1.ContainerMessageType.ChunkedOp: {
                // From observability POV, we should not expose the rest of the system (including "op" events on object) to these messages.
                // Also resetReconnectCount() would be wrong - see comment that was there before this change was made.
                (0, internal_2.assert)(false, 0x93d /* should not even get here */);
            }
            case messageTypes_js_1.ContainerMessageType.Rejoin: {
                break;
            }
            case messageTypes_js_1.ContainerMessageType.DocumentSchemaChange: {
                this.documentsSchemaController.processDocumentSchemaMessages(contents, local, message.sequenceNumber);
                break;
            }
            default: {
                const error = getUnknownMessageTypeError(message.type, "validateAndProcessRuntimeMessage" /* codePath */, message);
                this.closeFn(error);
                throw error;
            }
        }
    }
    processIdCompressorMessages(messageContents, savedOp) {
        for (const range of messageContents) {
            // Don't re-finalize the range if we're processing a "savedOp" in
            // stashed ops flow. The compressor is stashed with these ops already processed.
            // That said, in idCompressorMode === "delayed", we might not serialize ID compressor, and
            // thus we need to process all the ops.
            if (!(this.skipSavedCompressorOps && savedOp === true)) {
                // Some other client turned on the id compressor. If we have not turned it on,
                // put it in a pending queue and delay finalization.
                if (this._idCompressor === undefined) {
                    (0, internal_2.assert)(this.sessionSchema.idCompressorMode !== undefined, 0x93c /* id compressor should be enabled */);
                    this.pendingIdCompressorOps.push(range);
                }
                else {
                    (0, internal_2.assert)(this.pendingIdCompressorOps.length === 0, 0x979 /* there should be no pending ops! */);
                    this._idCompressor.finalizeCreationRange(range);
                }
            }
        }
    }
    processSignal(message, local) {
        const envelope = message.content;
        const transformed = {
            clientId: message.clientId,
            content: envelope.contents.content,
            type: envelope.contents.type,
            targetClientId: message.targetClientId,
        };
        // Only collect signal telemetry for broadcast messages sent by the current client.
        if (message.clientId === this.clientId) {
            this.signalTelemetryManager.trackReceivedSignal(envelope, this.mc.logger, this.consecutiveReconnects);
        }
        if (envelope.address === undefined) {
            // No address indicates a container signal message.
            this.emit("signal", transformed, local);
            return;
        }
        // Due to a mismatch between different layers in terms of
        // what is the interface of passing signals, we need to adjust
        // the signal envelope before sending it to the datastores to be processed
        const envelope2 = {
            address: envelope.address,
            contents: transformed.content,
        };
        transformed.content = envelope2;
        this.channelCollection.processSignal(transformed, local);
    }
    /**
     * Flush the current batch of ops to the ordering service for sequencing
     * This method is not expected to be called in the middle of a batch.
     * @remarks - If it throws (e.g. if the batch is too large to send), the container will be closed.
     *
     * @param resubmitInfo - If defined, indicates this is a resubmission of a batch with the given Batch info needed for resubmit.
     */
    flush(resubmitInfo) {
        try {
            (0, internal_2.assert)(!this.batchRunner.running, 0x24c /* "Cannot call `flush()` while manually accumulating a batch (e.g. under orderSequentially) */);
            this.outbox.flush(resubmitInfo);
            (0, internal_2.assert)(this.outbox.isEmpty, 0x3cf /* reentrancy */);
        }
        catch (error) {
            const error2 = (0, internal_8.normalizeError)(error, {
                props: {
                    orderSequentiallyCalls: this.batchRunner.runs,
                },
            });
            this.closeFn(error2);
            throw error2;
        }
    }
    /**
     * {@inheritDoc @fluidframework/runtime-definitions#IContainerRuntimeBase.orderSequentially}
     */
    orderSequentially(callback) {
        let checkpoint;
        const checkpointDirtyState = this.dirtyContainer;
        // eslint-disable-next-line import/no-deprecated
        let stageControls;
        if (this.mc.config.getBoolean("Fluid.ContainerRuntime.EnableRollback")) {
            if (!this.batchRunner.running && !this.inStagingMode) {
                stageControls = this.enterStagingMode();
            }
            // Note: we are not touching any batches other than mainBatch here, for two reasons:
            // 1. It would not help, as other batches are flushed independently from main batch.
            // 2. There is no way to undo process of data store creation, blob creation, ID compressor ops, or other things tracked by other batches.
            checkpoint = this.outbox.getBatchCheckpoints().mainBatch;
        }
        const result = this.batchRunner.run(() => {
            try {
                return callback();
            }
            catch (error) {
                if (checkpoint) {
                    // This will throw and close the container if rollback fails
                    try {
                        checkpoint.rollback((message) => this.rollback(message.runtimeOp, message.localOpMetadata));
                        // reset the dirty state after rollback to what it was before to keep it consistent
                        if (this.dirtyContainer !== checkpointDirtyState) {
                            this.updateDocumentDirtyState(checkpointDirtyState);
                        }
                        stageControls?.discardChanges();
                        stageControls = undefined;
                    }
                    catch (error_) {
                        const error2 = (0, internal_8.wrapError)(error_, (message) => {
                            return internal_8.DataProcessingError.create(`RollbackError: ${message}`, "checkpointRollback", undefined);
                        });
                        this.closeFn(error2);
                        throw error2;
                    }
                }
                else {
                    this.closeFn((0, internal_8.wrapError)(error, (errorMessage) => new internal_8.GenericError(`orderSequentially callback exception: ${errorMessage}`, error, {
                        orderSequentiallyCalls: this.batchRunner.runs,
                    })));
                }
                throw error; // throw the original error for the consumer of the runtime
            }
        });
        stageControls?.commitChanges({ squash: false });
        // We don't flush on TurnBased since we expect all messages in the same JS turn to be part of the same batch
        if (this.flushMode !== internal_6.FlushMode.TurnBased && !this.batchRunner.running) {
            this.flush();
        }
        return result;
    }
    /**
     * If true, the ContainerRuntime is not submitting any new ops to the ordering service.
     * Ops submitted to the ContainerRuntime while in Staging Mode will be queued in the PendingStateManager,
     * either to be discarded or committed later (via the Stage Controls returned from enterStagingMode).
     */
    get inStagingMode() {
        return this.stageControls !== undefined;
    }
    /**
     * Returns the aliased data store's entryPoint, given the alias.
     * @param alias - The alias for the data store.
     * @returns The data store's entry point ({@link @fluidframework/core-interfaces#IFluidHandle}) if it exists and is aliased.
     * Returns undefined if no data store has been assigned the given alias.
     */
    async getAliasedDataStoreEntryPoint(alias) {
        // Back-comapatibility:
        // There are old files that were created without using data store aliasing feature, but
        // used createRoot*DataStore*() (already removed) API. Such data stores will have isRoot = true,
        // and internalID provided by user. The expectation is that such files behave as new files, where
        // same data store instances created using aliasing feature.
        // Please also see note on name collisions in DataStores.createDataStoreId()
        await this.channelCollection.waitIfPendingAlias(alias);
        const internalId = this.internalId(alias);
        const context = await this.channelCollection.getDataStoreIfAvailable(internalId, {
            wait: false,
        });
        // If the data store is not available or not an alias, return undefined.
        if (context === undefined || !(await context.isRoot())) {
            return undefined;
        }
        const channel = await context.realize();
        if (channel.entryPoint === undefined) {
            throw new internal_8.UsageError("entryPoint must be defined on data store runtime for using getAliasedDataStoreEntryPoint");
        }
        this.garbageCollector.nodeUpdated({
            node: { type: "DataStore", path: `/${internalId}` },
            reason: "Loaded",
            packagePath: context.packagePath,
            timestampMs: this.getCurrentReferenceTimestampMs(),
        });
        return channel.entryPoint;
    }
    createDetachedDataStore(pkg, loadingGroupId) {
        return this.channelCollection.createDetachedDataStore(pkg, loadingGroupId);
    }
    async createDataStore(pkg, loadingGroupId) {
        const context = this.channelCollection.createDataStoreContext(Array.isArray(pkg) ? pkg : [pkg], loadingGroupId);
        return (0, dataStore_js_1.channelToDataStore)(await context.realize(), context.id, this.channelCollection, this.mc.logger);
    }
    canSendOps() {
        // Note that the real (non-proxy) delta manager is needed here to get the readonly info. This is because
        // container runtime's ability to send ops depend on the actual readonly state of the delta manager.
        return (this.connected && !this.innerDeltaManager.readOnlyInfo.readonly && !this.imminentClosure);
    }
    /**
     * Typically ops are batched and later flushed together, but in some cases we want to flush immediately.
     */
    currentlyBatching() {
        return this.flushMode !== internal_6.FlushMode.Immediate || this.batchRunner.running;
    }
    getQuorum() {
        return this._quorum;
    }
    getAudience() {
        return this._audience;
    }
    /**
     * Returns true of container is dirty, i.e. there are some pending local changes that
     * either were not sent out to delta stream or were not yet acknowledged.
     */
    get isDirty() {
        return this.dirtyContainer;
    }
    isContainerMessageDirtyable({ type, contents, }) {
        // Certain container runtime messages should not mark the container dirty such as the old built-in
        // AgentScheduler and Garbage collector messages.
        switch (type) {
            case messageTypes_js_1.ContainerMessageType.Attach: {
                const attachMessage = contents;
                if (attachMessage.id === exports.agentSchedulerId) {
                    return false;
                }
                break;
            }
            case messageTypes_js_1.ContainerMessageType.FluidDataStoreOp: {
                const envelope = contents;
                if (envelope.address === exports.agentSchedulerId) {
                    return false;
                }
                break;
            }
            case messageTypes_js_1.ContainerMessageType.IdAllocation:
            case messageTypes_js_1.ContainerMessageType.DocumentSchemaChange:
            case messageTypes_js_1.ContainerMessageType.GC: {
                return false;
            }
            default: {
                break;
            }
        }
        return true;
    }
    /**
     * Submits the signal to be sent to other clients.
     * @param type - Type of the signal.
     * @param content - Content of the signal. Should be a JSON serializable object or primitive.
     * @param targetClientId - When specified, the signal is only sent to the provided client id.
     *
     * @remarks
     *
     * The `targetClientId` parameter here is currently intended for internal testing purposes only.
     * Support for this option at container runtime is planned to be deprecated in the future.
     *
     */
    submitSignal(type, content, targetClientId) {
        this.verifyNotClosed();
        const envelope = createNewSignalEnvelope(undefined /* address */, type, content);
        if (targetClientId === undefined) {
            this.signalTelemetryManager.applyTrackingToBroadcastSignalEnvelope(envelope);
        }
        this.submitSignalFn(envelope, targetClientId);
    }
    setAttachState(attachState) {
        if (attachState === container_definitions_1.AttachState.Attaching) {
            (0, internal_2.assert)(this.attachState === container_definitions_1.AttachState.Attaching, 0x12d /* "Container Context should already be in attaching state" */);
        }
        else {
            (0, internal_2.assert)(this.attachState === container_definitions_1.AttachState.Attached, 0x12e /* "Container Context should already be in attached state" */);
            this.emit("attached");
        }
        if (attachState === container_definitions_1.AttachState.Attached && !this.hasPendingMessages()) {
            this.updateDocumentDirtyState(false);
        }
        this.channelCollection.setAttachState(attachState);
    }
    /**
     * Create a summary. Used when attaching or serializing a detached container.
     *
     * @param blobRedirectTable - A table passed during the attach process. While detached, blob upload is supported
     * using IDs generated locally. After attach, these IDs cannot be used, so this table maps the old local IDs to the
     * new storage IDs so requests can be redirected.
     * @param telemetryContext - summary data passed through the layers for telemetry purposes
     */
    createSummary(blobRedirectTable, telemetryContext) {
        if (blobRedirectTable) {
            this.blobManager.setRedirectTable(blobRedirectTable);
        }
        // We can finalize any allocated IDs since we're the only client
        const idRange = this._idCompressor?.takeNextCreationRange();
        if (idRange !== undefined) {
            (0, internal_2.assert)(idRange.ids === undefined || idRange.ids.firstGenCount === 1, 0x93e /* No other ranges should be taken while container is detached. */);
            this._idCompressor?.finalizeCreationRange(idRange);
        }
        const summarizeResult = this.channelCollection.getAttachSummary(telemetryContext);
        // Wrap data store summaries in .channels subtree.
        (0, index_js_4.wrapSummaryInChannelsTree)(summarizeResult);
        this.addContainerStateToSummary(summarizeResult, true /* fullTree */, false /* trackState */, telemetryContext);
        return summarizeResult.summary;
    }
    /**
     * Builds the Summary tree including all the channels and the container state.
     *
     * @remarks - Unfortunately, this function is accessed in a non-typesafe way by a legacy first-party partner,
     * so until we can provide a proper API for their scenario, we need to ensure this function doesn't change.
     */
    async summarizeInternal(fullTree, trackState, telemetryContext) {
        const summarizeResult = await this.channelCollection.summarize(fullTree, trackState, telemetryContext);
        // Wrap data store summaries in .channels subtree.
        (0, index_js_4.wrapSummaryInChannelsTree)(summarizeResult);
        const pathPartsForChildren = [internal_6.channelsTreeName];
        this.loadIdCompressor();
        this.addContainerStateToSummary(summarizeResult, fullTree, trackState, telemetryContext);
        return {
            ...summarizeResult,
            id: "",
            pathPartsForChildren,
        };
    }
    /**
     * Returns a summary of the runtime at the current sequence number.
     */
    async summarize(options) {
        this.verifyNotClosed();
        const { fullTree = false, trackState = true, summaryLogger = this.mc.logger, runGC = this.garbageCollector.shouldRunGC, runSweep, fullGC, } = options;
        const telemetryContext = new internal_7.TelemetryContext();
        // Add the options that are used to generate this summary to the telemetry context.
        telemetryContext.setMultiple("fluid_Summarize", "Options", {
            fullTree,
            trackState,
            runGC,
            fullGC,
            runSweep,
        });
        try {
            if (runGC) {
                await this.collectGarbage({ logger: summaryLogger, runSweep, fullGC }, telemetryContext);
            }
            const { stats, summary } = await this.summarizerNode.summarize(fullTree, trackState, telemetryContext);
            (0, internal_2.assert)(summary.type === driver_definitions_1.SummaryType.Tree, 0x12f /* "Container Runtime's summarize should always return a tree" */);
            return { stats, summary };
        }
        finally {
            summaryLogger.sendTelemetryEvent({
                eventName: "SummarizeTelemetry",
                details: telemetryContext.serialize(),
            });
        }
    }
    async getGCDataInternal(fullGC) {
        return this.channelCollection.getGCData(fullGC);
    }
    /**
     * Generates and returns the GC data for this container.
     * @param fullGC - true to bypass optimizations and force full generation of GC data.
     * @see IGarbageCollectionRuntime.getGCData
     */
    async getGCData(fullGC) {
        const builder = new internal_7.GCDataBuilder();
        const dsGCData = await this.summarizerNode.getGCData(fullGC);
        builder.addNodes(dsGCData.gcNodes);
        const blobsGCData = this.blobManager.getGCData(fullGC);
        builder.addNodes(blobsGCData.gcNodes);
        return builder.getGCData();
    }
    /**
     * After GC has run, called to notify this container's nodes of routes that are used in it.
     * @param usedRoutes - The routes that are used in all nodes in this Container.
     * @see IGarbageCollectionRuntime.updateUsedRoutes
     */
    updateUsedRoutes(usedRoutes) {
        // Update our summarizer node's used routes. Updating used routes in summarizer node before
        // summarizing is required and asserted by the the summarizer node. We are the root and are
        // always referenced, so the used routes is only self-route (empty string).
        this.summarizerNode.updateUsedRoutes([""]);
        const { dataStoreRoutes } = this.getDataStoreAndBlobManagerRoutes(usedRoutes);
        this.channelCollection.updateUsedRoutes(dataStoreRoutes);
    }
    /**
     * After GC has run and identified nodes that are sweep ready, this is called to delete the sweep ready nodes.
     * @param sweepReadyRoutes - The routes of nodes that are sweep ready and should be deleted.
     * @returns The routes of nodes that were deleted.
     */
    deleteSweepReadyNodes(sweepReadyRoutes) {
        const { dataStoreRoutes, blobManagerRoutes } = this.getDataStoreAndBlobManagerRoutes(sweepReadyRoutes);
        return [
            ...this.channelCollection.deleteSweepReadyNodes(dataStoreRoutes),
            ...this.blobManager.deleteSweepReadyNodes(blobManagerRoutes),
        ];
    }
    /**
     * This is called to update objects that are tombstones.
     *
     * A Tombstoned object has been unreferenced long enough that GC knows it won't be referenced again.
     * Tombstoned objects are eventually deleted by GC.
     *
     * @param tombstonedRoutes - Data store and attachment blob routes that are tombstones in this Container.
     */
    updateTombstonedRoutes(tombstonedRoutes) {
        const { dataStoreRoutes } = this.getDataStoreAndBlobManagerRoutes(tombstonedRoutes);
        this.channelCollection.updateTombstonedRoutes(dataStoreRoutes);
    }
    /**
     * Returns a server generated referenced timestamp to be used to track unreferenced nodes by GC.
     */
    getCurrentReferenceTimestampMs() {
        // Use the timestamp of the last message seen by this client as that is server generated. If no messages have
        // been processed, use the timestamp of the message from the last summary.
        return this.deltaManager.lastMessage?.timestamp ?? this.messageAtLastSummary?.timestamp;
    }
    /**
     * Returns the type of the GC node. Currently, there are nodes that belong to the root ("/"), data stores or
     * blob manager.
     */
    getNodeType(nodePath) {
        if ((0, index_js_1.isBlobPath)(nodePath)) {
            return index_js_2.GCNodeType.Blob;
        }
        return this.channelCollection.getGCNodeType(nodePath) ?? index_js_2.GCNodeType.Other;
    }
    /**
     * Called by GC to retrieve the package path of the node with the given path. The node should belong to a
     * data store or an attachment blob.
     */
    async getGCNodePackagePath(nodePath) {
        // GC uses "/" when adding "root" references, e.g. for Aliasing or as part of Tombstone Auto-Recovery.
        // These have no package path so return a special value.
        if (nodePath === "/") {
            return ["_gcRoot"];
        }
        switch (this.getNodeType(nodePath)) {
            case index_js_2.GCNodeType.Blob: {
                return [index_js_1.blobManagerBasePath];
            }
            case index_js_2.GCNodeType.DataStore:
            case index_js_2.GCNodeType.SubDataStore: {
                return this.channelCollection.getDataStorePackagePath(nodePath);
            }
            default: {
                (0, internal_2.assert)(false, 0x2de /* "Package path requested for unsupported node type." */);
            }
        }
    }
    /**
     * From a given list of routes, separate and return routes that belong to blob manager and data stores.
     * @param routes - A list of routes that can belong to data stores or blob manager.
     * @returns Two route lists - One that contains routes for blob manager and another one that contains routes
     * for data stores.
     */
    getDataStoreAndBlobManagerRoutes(routes) {
        const blobManagerRoutes = [];
        const dataStoreRoutes = [];
        for (const route of routes) {
            if ((0, index_js_1.isBlobPath)(route)) {
                blobManagerRoutes.push(route);
            }
            else {
                dataStoreRoutes.push(route);
            }
        }
        return { blobManagerRoutes, dataStoreRoutes };
    }
    /**
     * Runs garbage collection and updates the reference / used state of the nodes in the container.
     * @returns the statistics of the garbage collection run; undefined if GC did not run.
     */
    async collectGarbage(options, telemetryContext) {
        return this.garbageCollector.collectGarbage(options, telemetryContext);
    }
    /**
     * Called when a new outbound route is added to another node. This is used by garbage collection to identify
     * all references added in the system.
     * @param fromPath - The absolute path of the node that added the reference.
     * @param toPath - The absolute path of the outbound node that is referenced.
     * @param messageTimestampMs - The timestamp of the message that added the reference.
     */
    addedGCOutboundRoute(fromPath, toPath, messageTimestampMs) {
        // This is always called when processing an op so messageTimestampMs should exist. Due to back-compat
        // across the data store runtime / container runtime boundary, this may be undefined and if so, get
        // the timestamp from the last processed message which should exist.
        // If a timestamp doesn't exist, log so we can learn about these cases and return.
        const timestampMs = messageTimestampMs ?? this.getCurrentReferenceTimestampMs();
        if (timestampMs === undefined) {
            this.mc.logger.sendTelemetryEvent({
                eventName: "NoTimestampInGCOutboundRoute",
                ...(0, internal_8.tagCodeArtifacts)({
                    id: toPath,
                    fromId: fromPath,
                }),
            });
            return;
        }
        this.garbageCollector.addedOutboundReference(fromPath, toPath, timestampMs);
    }
    /**
     * Generates the summary tree, uploads it to storage, and then submits the summarize op.
     * This is intended to be called by the summarizer, since it is the implementation of
     * ISummarizerInternalsProvider.submitSummary.
     * It takes care of state management at the container level, including pausing inbound
     * op processing, updating SummarizerNode state tracking, and garbage collection.
     * @param options - options controlling how the summary is generated or submitted
     */
    async submitSummary(options) {
        const { cancellationToken, fullTree = false, finalAttempt = false, summaryLogger, latestSummaryRefSeqNum, } = options;
        // The summary number for this summary. This will be updated during the summary process, so get it now and
        // use it for all events logged during this summary.
        const summaryNumber = this.nextSummaryNumber;
        let summaryRefSeqNum;
        const summaryNumberLogger = (0, internal_8.createChildLogger)({
            logger: summaryLogger,
            properties: {
                all: {
                    summaryNumber,
                    referenceSequenceNumber: () => summaryRefSeqNum,
                },
            },
        });
        // legacy: assert 0x3d1
        if (!this.outbox.isEmpty) {
            throw internal_8.DataProcessingError.create("Can't trigger summary in the middle of a batch", "submitSummary", undefined, {
                summaryNumber,
                pendingMessages: this.pendingMessagesCount,
                outboxLength: this.outbox.messageCount,
                mainBatchLength: this.outbox.mainBatchMessageCount,
                blobAttachBatchLength: this.outbox.blobAttachBatchMessageCount,
                idAllocationBatchLength: this.outbox.idAllocationBatchMessageCount,
            });
        }
        // If the container is dirty, i.e., there are pending unacked ops, the summary will not be eventual consistent
        // and it may even be incorrect. So, wait for the container to be saved with a timeout. If the container is not
        // saved within the timeout, check if it should be failed or can continue.
        if (this.isDirty) {
            const countBefore = this.pendingMessagesCount;
            // The timeout for waiting for pending ops can be overridden via configurations.
            const pendingOpsTimeout = this.mc.config.getNumber("Fluid.Summarizer.waitForPendingOpsTimeoutMs") ??
                exports.defaultPendingOpsWaitTimeoutMs;
            await new Promise((resolve, reject) => {
                const timeoutId = setTimeout(() => resolve(), pendingOpsTimeout);
                this.once("saved", () => {
                    clearTimeout(timeoutId);
                    resolve();
                });
                this.once("dispose", () => {
                    clearTimeout(timeoutId);
                    reject(new Error("Runtime is disposed while summarizing"));
                });
            });
            // Log that there are pending ops while summarizing. This will help us gather data on how often this
            // happens, whether we attempted to wait for these ops to be acked and what was the result.
            summaryNumberLogger.sendTelemetryEvent({
                eventName: "PendingOpsWhileSummarizing",
                saved: !this.isDirty,
                timeout: pendingOpsTimeout,
                countBefore,
                countAfter: this.pendingMessagesCount,
            });
            // There could still be pending ops. Check if summary should fail or continue.
            const pendingMessagesFailResult = await this.shouldFailSummaryOnPendingOps(summaryNumberLogger, this.deltaManager.lastSequenceNumber, this.deltaManager.minimumSequenceNumber, finalAttempt, true /* beforeSummaryGeneration */);
            if (pendingMessagesFailResult !== undefined) {
                return pendingMessagesFailResult;
            }
        }
        const shouldPauseInboundSignal = this.mc.config.getBoolean("Fluid.ContainerRuntime.SubmitSummary.disableInboundSignalPause") !== true;
        const shouldValidatePreSummaryState = this.mc.config.getBoolean("Fluid.ContainerRuntime.SubmitSummary.shouldValidatePreSummaryState") === true;
        try {
            await this._deltaManager.inbound.pause();
            if (shouldPauseInboundSignal) {
                await this.deltaManager.inboundSignal.pause();
            }
            summaryRefSeqNum = this.deltaManager.lastSequenceNumber;
            const minimumSequenceNumber = this.deltaManager.minimumSequenceNumber;
            const message = `Summary @${summaryRefSeqNum}:${this.deltaManager.minimumSequenceNumber}`;
            const lastAckedContext = this.lastAckedSummaryContext;
            const startSummaryResult = this.summarizerNode.startSummary(summaryRefSeqNum, summaryNumberLogger, latestSummaryRefSeqNum);
            /**
             * This was added to validate that the summarizer node tree has the same reference sequence number from the
             * top running summarizer down to the lowest summarizer node.
             *
             * The order of mismatch numbers goes (validate sequence number)-(node sequence number).
             * Generally the validate sequence number comes from the running summarizer and the node sequence number comes from the
             * summarizer nodes.
             */
            if (startSummaryResult.invalidNodes > 0 || startSummaryResult.mismatchNumbers.size > 0) {
                summaryLogger.sendTelemetryEvent({
                    eventName: "LatestSummaryRefSeqNumMismatch",
                    details: {
                        ...startSummaryResult,
                        mismatchNumbers: [...startSummaryResult.mismatchNumbers],
                    },
                });
                if (shouldValidatePreSummaryState && !finalAttempt) {
                    return {
                        stage: "base",
                        referenceSequenceNumber: summaryRefSeqNum,
                        minimumSequenceNumber,
                        error: new index_js_4.RetriableSummaryError(`Summarizer node state inconsistent with summarizer state.`),
                    };
                }
            }
            // Helper function to check whether we should still continue between each async step.
            const checkContinue = () => {
                // Do not check for loss of connectivity directly! Instead leave it up to
                // RunWhileConnectedCoordinator to control policy in a single place.
                // This will allow easier change of design if we chose to. For example, we may chose to allow
                // summarizer to reconnect in the future.
                // Also checking for cancellation is a must as summary process may be abandoned for other reasons,
                // like loss of connectivity for main (interactive) client.
                if (cancellationToken.cancelled) {
                    return { continue: false, error: "disconnected" };
                }
                // That said, we rely on submitSystemMessage() that today only works in connected state.
                // So if we fail here, it either means that RunWhileConnectedCoordinator does not work correctly,
                // OR that design changed and we need to remove this check and fix submitSystemMessage.
                (0, internal_2.assert)(this.connected, 0x258 /* "connected" */);
                // Ensure that lastSequenceNumber has not changed after pausing.
                // We need the summary op's reference sequence number to match our summary sequence number,
                // otherwise we'll get the wrong sequence number stamped on the summary's .protocol attributes.
                if (this.deltaManager.lastSequenceNumber !== summaryRefSeqNum) {
                    return {
                        continue: false,
                        error: `lastSequenceNumber changed before uploading to storage. ${this.deltaManager.lastSequenceNumber} !== ${summaryRefSeqNum}`,
                    };
                }
                (0, internal_2.assert)(summaryRefSeqNum === this.deltaManager.lastMessage?.sequenceNumber, 0x395 /* it's one and the same thing */);
                if (lastAckedContext !== this.lastAckedSummaryContext) {
                    return {
                        continue: false,
                        error: `Last summary changed while summarizing. ${this.lastAckedSummaryContext} !== ${lastAckedContext}`,
                    };
                }
                return { continue: true };
            };
            let continueResult = checkContinue();
            if (!continueResult.continue) {
                return {
                    stage: "base",
                    referenceSequenceNumber: summaryRefSeqNum,
                    minimumSequenceNumber,
                    error: new index_js_4.RetriableSummaryError(continueResult.error),
                };
            }
            const trace = client_utils_1.Trace.start();
            let summarizeResult;
            try {
                summarizeResult = await this.summarize({
                    fullTree,
                    trackState: true,
                    summaryLogger: summaryNumberLogger,
                    runGC: this.garbageCollector.shouldRunGC,
                });
            }
            catch (error) {
                return {
                    stage: "base",
                    referenceSequenceNumber: summaryRefSeqNum,
                    minimumSequenceNumber,
                    error: (0, internal_8.wrapError)(error, (msg) => new index_js_4.RetriableSummaryError(msg)),
                };
            }
            // Validate that the summary generated by summarizer nodes is correct before uploading.
            const validateResult = this.summarizerNode.validateSummary();
            if (!validateResult.success) {
                const { success, ...loggingProps } = validateResult;
                const error = new index_js_4.RetriableSummaryError(validateResult.reason, validateResult.retryAfterSeconds, { ...loggingProps });
                return {
                    stage: "base",
                    referenceSequenceNumber: summaryRefSeqNum,
                    minimumSequenceNumber,
                    error,
                };
            }
            // If there are pending unacked ops, this summary attempt may fail as the uploaded
            // summary would be eventually inconsistent.
            const pendingMessagesFailResult = await this.shouldFailSummaryOnPendingOps(summaryNumberLogger, summaryRefSeqNum, minimumSequenceNumber, finalAttempt, false /* beforeSummaryGeneration */);
            if (pendingMessagesFailResult !== undefined) {
                return pendingMessagesFailResult;
            }
            const { summary: summaryTree, stats: partialStats } = summarizeResult;
            // Now that we have generated the summary, update the message at last summary to the last message processed.
            this.messageAtLastSummary = this.deltaManager.lastMessage;
            // Counting dataStores and handles
            // Because handles are unchanged dataStores in the current logic,
            // summarized dataStore count is total dataStore count minus handle count
            const dataStoreTree = summaryTree.tree[internal_6.channelsTreeName];
            (0, internal_2.assert)(dataStoreTree.type === driver_definitions_1.SummaryType.Tree, 0x1fc /* "summary is not a tree" */);
            const handleCount = Object.values(dataStoreTree.tree).filter((value) => value.type === driver_definitions_1.SummaryType.Handle).length;
            const gcSummaryTreeStats = summaryTree.tree[internal_6.gcTreeKey]
                ? (0, internal_7.calculateStats)(summaryTree.tree[internal_6.gcTreeKey])
                : undefined;
            const summaryStats = {
                dataStoreCount: this.channelCollection.size,
                summarizedDataStoreCount: this.channelCollection.size - handleCount,
                gcStateUpdatedDataStoreCount: this.garbageCollector.updatedDSCountSinceLastSummary,
                gcBlobNodeCount: gcSummaryTreeStats?.blobNodeCount,
                gcTotalBlobsSize: gcSummaryTreeStats?.totalBlobSize,
                summaryNumber,
                ...partialStats,
            };
            const generateSummaryData = {
                referenceSequenceNumber: summaryRefSeqNum,
                minimumSequenceNumber,
                summaryTree,
                summaryStats,
                generateDuration: trace.trace().duration,
            };
            continueResult = checkContinue();
            if (!continueResult.continue) {
                return {
                    stage: "generate",
                    ...generateSummaryData,
                    error: new index_js_4.RetriableSummaryError(continueResult.error),
                };
            }
            const summaryContext = {
                proposalHandle: this.lastAckedSummaryContext?.proposalHandle ?? undefined,
                ackHandle: this.lastAckedSummaryContext?.ackHandle ?? this.loadedFromVersionId,
                referenceSequenceNumber: summaryRefSeqNum,
            };
            let handle;
            try {
                handle = await this.storage.uploadSummaryWithContext(summaryTree, summaryContext);
            }
            catch (error) {
                return {
                    stage: "generate",
                    ...generateSummaryData,
                    error: (0, internal_8.wrapError)(error, (msg) => new index_js_4.RetriableSummaryError(msg)),
                };
            }
            const parent = summaryContext.ackHandle;
            const summaryMessage = {
                handle,
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                head: parent,
                message,
                parents: parent ? [parent] : [],
            };
            const uploadData = {
                ...generateSummaryData,
                handle,
                uploadDuration: trace.trace().duration,
            };
            continueResult = checkContinue();
            if (!continueResult.continue) {
                return {
                    stage: "upload",
                    ...uploadData,
                    error: new index_js_4.RetriableSummaryError(continueResult.error),
                };
            }
            let clientSequenceNumber;
            try {
                clientSequenceNumber = this.submitSummaryMessage(summaryMessage, summaryRefSeqNum);
            }
            catch (error) {
                return {
                    stage: "upload",
                    ...uploadData,
                    error: (0, internal_8.wrapError)(error, (msg) => new index_js_4.RetriableSummaryError(msg)),
                };
            }
            const submitData = {
                stage: "submit",
                ...uploadData,
                clientSequenceNumber,
                submitOpDuration: trace.trace().duration,
            };
            try {
                this.summarizerNode.completeSummary(handle);
            }
            catch (error) {
                return {
                    stage: "upload",
                    ...uploadData,
                    error: (0, internal_8.wrapError)(error, (msg) => new index_js_4.RetriableSummaryError(msg)),
                };
            }
            return submitData;
        }
        finally {
            // Cleanup wip summary in case of failure
            this.summarizerNode.clearSummary();
            // ! This needs to happen before we resume inbound queues to ensure heuristics are tracked correctly
            this._summarizer?.recordSummaryAttempt?.(summaryRefSeqNum);
            // Restart the delta manager
            this._deltaManager.inbound.resume();
            if (shouldPauseInboundSignal) {
                this.deltaManager.inboundSignal.resume();
            }
        }
    }
    /**
     * This helper is called during summarization. If the container is dirty, it will return a failed summarize result
     * (IBaseSummarizeResult) unless this is the final summarize attempt and SkipFailingIncorrectSummary option is set.
     * @param logger - The logger to be used for sending telemetry.
     * @param referenceSequenceNumber - The reference sequence number of the summary attempt.
     * @param minimumSequenceNumber - The minimum sequence number of the summary attempt.
     * @param finalAttempt - Whether this is the final summary attempt.
     * @param beforeSummaryGeneration - Whether this is called before summary generation or after.
     * @returns failed summarize result (IBaseSummarizeResult) if summary should be failed, undefined otherwise.
     */
    async shouldFailSummaryOnPendingOps(logger, referenceSequenceNumber, minimumSequenceNumber, finalAttempt, beforeSummaryGeneration) {
        if (!this.isDirty) {
            return;
        }
        // If "SkipFailingIncorrectSummary" option is true, don't fail the summary in the last attempt.
        // This is a fallback to make progress in documents where there are consistently pending ops in
        // the summarizer.
        if (finalAttempt &&
            this.mc.config.getBoolean("Fluid.Summarizer.SkipFailingIncorrectSummary")) {
            const error = internal_8.DataProcessingError.create("Pending ops during summarization", "submitSummary", undefined, { pendingMessages: this.pendingMessagesCount });
            logger.sendErrorEvent({
                eventName: "SkipFailingIncorrectSummary",
                referenceSequenceNumber,
                minimumSequenceNumber,
                beforeGenerate: beforeSummaryGeneration,
            }, error);
        }
        else {
            // The retry delay when there are pending ops can be overridden via config so that we can adjust it
            // based on telemetry while we decide on a stable number.
            const retryDelayMs = this.mc.config.getNumber("Fluid.Summarizer.PendingOpsRetryDelayMs") ??
                exports.defaultPendingOpsRetryDelayMs;
            const error = new index_js_4.RetriableSummaryError("PendingOpsWhileSummarizing", retryDelayMs / 1000, {
                count: this.pendingMessagesCount,
                beforeGenerate: beforeSummaryGeneration,
            });
            return {
                stage: "base",
                referenceSequenceNumber,
                minimumSequenceNumber,
                error,
            };
        }
    }
    get pendingMessagesCount() {
        return this.pendingStateManager.pendingMessagesCount + this.outbox.messageCount;
    }
    hasPendingMessages() {
        return this.pendingMessagesCount !== 0;
    }
    updateDocumentDirtyState(dirty) {
        if (this.attachState === container_definitions_1.AttachState.Attached) {
            // Other way is not true = see this.isContainerMessageDirtyable()
            (0, internal_2.assert)(!dirty || this.hasPendingMessages(), 0x3d3 /* if doc is dirty, there has to be pending ops */);
        }
        else {
            (0, internal_2.assert)(dirty, 0x3d2 /* Non-attached container is dirty */);
        }
        if (this.dirtyContainer === dirty) {
            return;
        }
        this.dirtyContainer = dirty;
        if (this.emitDirtyDocumentEvent) {
            this.emit(dirty ? "dirty" : "saved");
        }
    }
    submitMessage(type, 
    // TODO: better typing
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types, @typescript-eslint/no-explicit-any
    contents, localOpMetadata = undefined) {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
        this.submit({ type, contents }, localOpMetadata);
    }
    async uploadBlob(blob, signal) {
        this.verifyNotClosed();
        return this.blobManager.createBlob(blob, signal);
    }
    submitIdAllocationOpIfNeeded(resubmitOutstandingRanges) {
        if (this._idCompressor) {
            const idRange = resubmitOutstandingRanges
                ? this._idCompressor.takeUnfinalizedCreationRange()
                : this._idCompressor.takeNextCreationRange();
            // Don't include the idRange if there weren't any Ids allocated
            if (idRange.ids !== undefined) {
                const idAllocationMessage = {
                    type: messageTypes_js_1.ContainerMessageType.IdAllocation,
                    contents: idRange,
                };
                const idAllocationBatchMessage = {
                    runtimeOp: idAllocationMessage,
                    referenceSequenceNumber: this.deltaManager.lastSequenceNumber,
                    // Note: For now, we will never stage ID Allocation messages.
                    // They won't contain personal info and no harm in extra allocations in case of discarding the staged changes
                    staged: false,
                };
                this.outbox.submitIdAllocation(idAllocationBatchMessage);
            }
        }
    }
    submit(containerRuntimeMessage, localOpMetadata = undefined, metadata) {
        this.verifyNotClosed();
        // There should be no ops in detached container state!
        (0, internal_2.assert)(this.attachState !== container_definitions_1.AttachState.Detached, 0x132 /* "sending ops in detached container" */);
        (0, internal_2.assert)(metadata === undefined ||
            containerRuntimeMessage.type === messageTypes_js_1.ContainerMessageType.BlobAttach, 0x93f /* metadata */);
        // Note that the real (non-proxy) delta manager is used here to get the readonly info. This is because
        // container runtime's ability to submit ops depend on the actual readonly state of the delta manager.
        if (this.innerDeltaManager.readOnlyInfo.readonly) {
            this.mc.logger.sendTelemetryEvent({
                eventName: "SubmitOpInReadonly",
                connected: this.connected,
            });
        }
        const type = containerRuntimeMessage.type;
        (0, internal_2.assert)(type !== messageTypes_js_1.ContainerMessageType.IdAllocation, 0x9a5 /* IdAllocation should be submitted directly to outbox. */);
        try {
            this.submitIdAllocationOpIfNeeded(false);
            // Allow document schema controller to send a message if it needs to propose change in document schema.
            // If it needs to send a message, it will call provided callback with payload of such message and rely
            // on this callback to do actual sending.
            const schemaChangeMessage = this.documentsSchemaController.maybeSendSchemaMessage();
            if (schemaChangeMessage) {
                this.mc.logger.sendTelemetryEvent({
                    eventName: "SchemaChangeProposal",
                    refSeq: schemaChangeMessage.refSeq,
                    version: schemaChangeMessage.version,
                    newRuntimeSchema: JSON.stringify(schemaChangeMessage.runtime),
                    sessionRuntimeSchema: JSON.stringify(this.sessionSchema),
                    oldRuntimeSchema: JSON.stringify(this.metadata?.documentSchema?.runtime),
                });
                const msg = {
                    type: messageTypes_js_1.ContainerMessageType.DocumentSchemaChange,
                    contents: schemaChangeMessage,
                };
                this.outbox.submit({
                    runtimeOp: msg,
                    referenceSequenceNumber: this.deltaManager.lastSequenceNumber,
                    staged: this.inStagingMode,
                });
            }
            const message = {
                // This will encode any handles present in this op before serializing to string
                // Note: handles may already have been encoded by the DDS layer, but encoding handles is idempotent so there's no problem.
                runtimeOp: containerRuntimeMessage,
                metadata,
                localOpMetadata,
                referenceSequenceNumber: this.deltaManager.lastSequenceNumber,
                staged: this.inStagingMode,
            };
            if (type === messageTypes_js_1.ContainerMessageType.BlobAttach) {
                // BlobAttach ops must have their metadata visible and cannot be grouped (see opGroupingManager.ts)
                this.outbox.submitBlobAttach(message);
            }
            else {
                this.outbox.submit(message);
            }
            // Note: Technically, the system "always" batches - if this case is true we'll just have a single-message batch.
            const flushImmediatelyOnSubmit = !this.currentlyBatching();
            if (flushImmediatelyOnSubmit) {
                this.flush();
            }
            else {
                this.scheduleFlush();
            }
        }
        catch (error) {
            const dpe = internal_8.DataProcessingError.wrapIfUnrecognized(error, "ContainerRuntime.submit", {
                referenceSequenceNumber: this.deltaManager.lastSequenceNumber,
            });
            this.closeFn(dpe);
            throw dpe;
        }
        if (this.isContainerMessageDirtyable(containerRuntimeMessage)) {
            this.updateDocumentDirtyState(true);
        }
    }
    scheduleFlush() {
        if (this.flushTaskExists) {
            return;
        }
        this.flushTaskExists = true;
        // TODO: hoist this out of the function scope to save unnecessary allocations
        // eslint-disable-next-line unicorn/consistent-function-scoping -- Separate `flush` method already exists in outer scope
        const flush = () => {
            this.flushTaskExists = false;
            this.flush();
        };
        switch (this.flushMode) {
            case internal_6.FlushMode.TurnBased: {
                // When in TurnBased flush mode the runtime will buffer operations in the current turn and send them as a single
                // batch at the end of the turn
                // eslint-disable-next-line @typescript-eslint/no-floating-promises
                Promise.resolve().then(flush);
                break;
            }
            // FlushModeExperimental is experimental and not exposed directly in the runtime APIs
            case internal_6.FlushModeExperimental.Async: {
                // When in Async flush mode, the runtime will accumulate all operations across JS turns and send them as a single
                // batch when all micro-tasks are complete.
                // Compared to TurnBased, this flush mode will capture more ops into the same batch.
                setTimeout(flush, 0);
                break;
            }
            default: {
                (0, internal_2.assert)(this.batchRunner.running, 0x587 /* Unreachable unless manually accumulating a batch */);
                break;
            }
        }
    }
    submitSummaryMessage(contents, referenceSequenceNumber) {
        this.verifyNotClosed();
        (0, internal_2.assert)(this.connected, 0x133 /* "Container disconnected when trying to submit system message" */);
        // System message should not be sent in the middle of the batch.
        (0, internal_2.assert)(this.outbox.isEmpty, 0x3d4 /* System op in the middle of a batch */);
        return this.submitSummaryFn(contents, referenceSequenceNumber);
    }
    /**
     * Throw an error if the runtime is closed.  Methods that are expected to potentially
     * be called after dispose due to asynchrony should not call this.
     */
    verifyNotClosed() {
        if (this._disposed) {
            throw new Error("Runtime is closed");
        }
    }
    /**
     * Resubmits each message in the batch, and then flushes the outbox.
     *
     * @remarks - If the "Offline Load" feature is enabled, the batchId is included in the resubmitted messages,
     * for correlation to detect container forking.
     */
    reSubmitBatch(batch, { batchId, staged, squash }) {
        this.batchRunner.run(() => {
            for (const message of batch) {
                this.reSubmit(message, squash);
            }
        });
        // Only include Batch ID if "Offline Load" feature is enabled
        // It's only needed to identify batches across container forks arising from misuse of offline load.
        this.flush({ batchId: this.offlineEnabled ? batchId : undefined, staged });
    }
    reSubmit(message, squash) {
        this.reSubmitCore(message.runtimeOp, message.localOpMetadata, message.opMetadata, squash);
    }
    /**
     * Finds the right store and asks it to resubmit the message. This typically happens when we
     * reconnect and there are pending messages.
     * ! Note: successfully resubmitting an op that has been successfully sequenced is not possible due to checks in the ConnectionStateHandler (Loader layer)
     * @param message - The original LocalContainerRuntimeMessage.
     * @param localOpMetadata - The local metadata associated with the original message.
     */
    reSubmitCore(message, localOpMetadata, opMetadata, squash) {
        (0, internal_2.assert)(this._summarizer === undefined, 0x8f2 /* Summarizer never reconnects so should never resubmit */);
        switch (message.type) {
            case messageTypes_js_1.ContainerMessageType.FluidDataStoreOp:
            case messageTypes_js_1.ContainerMessageType.Attach:
            case messageTypes_js_1.ContainerMessageType.Alias: {
                // For Operations, call resubmitDataStoreOp which will find the right store
                // and trigger resubmission on it.
                this.channelCollection.reSubmit(message.type, message.contents, localOpMetadata, squash);
                break;
            }
            case messageTypes_js_1.ContainerMessageType.IdAllocation: {
                // Allocation ops are never resubmitted/rebased. This is because they require special handling to
                // avoid being submitted out of order. For example, if the pending state manager contained
                // [idOp1, dataOp1, idOp2, dataOp2] and the resubmission of dataOp1 generated idOp3, that would be
                // placed into the outbox in the same batch as idOp1, but before idOp2 is resubmitted.
                // To avoid this, allocation ops are simply never resubmitted. Prior to invoking the pending state
                // manager to replay pending ops, the runtime will always submit a new allocation range that includes
                // all pending IDs. The resubmitted allocation ops are then ignored here.
                break;
            }
            case messageTypes_js_1.ContainerMessageType.BlobAttach: {
                this.blobManager.reSubmit(opMetadata);
                break;
            }
            case messageTypes_js_1.ContainerMessageType.Rejoin: {
                this.submit(message);
                break;
            }
            case messageTypes_js_1.ContainerMessageType.GC: {
                this.submit(message);
                break;
            }
            case messageTypes_js_1.ContainerMessageType.DocumentSchemaChange: {
                // There is no need to resend this message. Document schema controller will properly resend it again (if needed)
                // on a first occasion (any ops sent after reconnect). There is a good chance, though, that it will not want to
                // send any ops, as some other client already changed schema.
                break;
            }
            default: {
                const error = getUnknownMessageTypeError(message.type, "reSubmitCore" /* codePath */);
                this.closeFn(error);
                throw error;
            }
        }
    }
    rollback(runtimeOp, localOpMetadata) {
        const { type, contents } = runtimeOp;
        switch (type) {
            case messageTypes_js_1.ContainerMessageType.FluidDataStoreOp: {
                // For operations, call rollbackDataStoreOp which will find the right store
                // and trigger rollback on it.
                this.channelCollection.rollback(type, contents, localOpMetadata);
                break;
            }
            default: {
                throw new Error(`Can't rollback ${type}`);
            }
        }
    }
    /**
     * Implementation of ISummarizerInternalsProvider.refreshLatestSummaryAck
     */
    async refreshLatestSummaryAck(options) {
        const { proposalHandle, ackHandle, summaryRefSeq, summaryLogger } = options;
        // proposalHandle is always passed from RunningSummarizer.
        (0, internal_2.assert)(proposalHandle !== undefined, 0x766 /* proposalHandle should be available */);
        const result = await this.summarizerNode.refreshLatestSummary(proposalHandle, summaryRefSeq);
        /* eslint-disable jsdoc/check-indentation */
        /**
         * If the snapshot corresponding to the ack is not tracked by this client, it was submitted by another client.
         * Take action as per the following scenarios:
         * 1. If that snapshot is older than the one tracked by this client, ignore the ack because only the latest
         *    snapshot is tracked.
         * 2. If that snapshot is newer, attempt to fetch the latest snapshot and do one of the following:
         *    2.1. If the fetched snapshot is same or newer than the one for which ack was received, close this client.
         *         The next summarizer client will likely start from this snapshot and get out of this state. Fetching
         *         the snapshot updates the cache for this client so if it's re-elected as summarizer, this will prevent
         *         any thrashing.
         *    2.2. If the fetched snapshot is older than the one for which ack was received, ignore the ack. This can
         *         happen in scenarios where the snapshot for the ack was lost in storage (in scenarios like DB rollback,
         *         etc.) but the summary ack is still there because it's tracked a different service. In such cases,
         *         ignoring the ack is the correct thing to do because the latest snapshot in storage is not the one for
         *         the ack but is still the one tracked by this client. If we were to close the summarizer like in the
         *         previous scenario, it will result in this document stuck in this state in a loop.
         */
        /* eslint-enable jsdoc/check-indentation */
        if (!result.isSummaryTracked) {
            if (result.isSummaryNewer) {
                await this.fetchLatestSnapshotAndMaybeClose(summaryRefSeq, ackHandle, summaryLogger);
            }
            return;
        }
        // Notify the garbage collector so it can update its latest summary state.
        await this.garbageCollector.refreshLatestSummary(result);
        // If we here, the ack was tracked by this client. Update the summary context of the last ack.
        this.lastAckedSummaryContext = {
            proposalHandle,
            ackHandle,
            referenceSequenceNumber: summaryRefSeq,
        };
    }
    /**
     * Fetches the latest snapshot from storage. If the fetched snapshot is same or newer than the one for which ack
     * was received, close this client. Fetching the snapshot will update the cache for this client so if it's
     * re-elected as summarizer, this will prevent any thrashing.
     * If the fetched snapshot is older than the one for which ack was received, ignore the ack and return. This can
     * happen in scenarios where the snapshot for the ack was lost in storage in scenarios like DB rollback, etc.
     */
    async fetchLatestSnapshotAndMaybeClose(targetRefSeq, targetAckHandle, logger) {
        const fetchedSnapshotRefSeq = await internal_8.PerformanceEvent.timedExecAsync(logger, { eventName: "RefreshLatestSummaryAckFetch" }, async (perfEvent) => {
            const props = { targetRefSeq, targetAckHandle };
            const trace = client_utils_1.Trace.start();
            let snapshotTree;
            const scenarioName = "RefreshLatestSummaryAckFetch";
            // If loader supplied us the ISnapshot when loading, the new getSnapshotApi is supported and feature gate is ON, then use the
            // new API, otherwise it will reduce the service performance because the service will need to recalculate the full snapshot
            // in case previously getSnapshotApi was used and now we use the getVersions API.
            if (this.isSnapshotInstanceOfISnapshot &&
                this.storage.getSnapshot !== undefined &&
                this.mc.config.getBoolean("Fluid.Container.UseLoadingGroupIdForSnapshotFetch2") ===
                    true) {
                const snapshot = await this.storage.getSnapshot({
                    scenarioName,
                    fetchSource: internal_3.FetchSource.noCache,
                });
                const id = snapshot.snapshotTree.id;
                (0, internal_2.assert)(id !== undefined, 0x9d0 /* id of the fetched snapshot should be defined */);
                props.snapshotVersion = id;
                snapshotTree = snapshot.snapshotTree;
            }
            else {
                const versions = await this.storage.getVersions(
                // eslint-disable-next-line unicorn/no-null
                null, 1, scenarioName, internal_3.FetchSource.noCache);
                (0, internal_2.assert)(!!versions && !!versions[0], 0x137 /* "Failed to get version from storage" */);
                snapshotTree = await this.storage.getSnapshotTree(versions[0]);
                (0, internal_2.assert)(!!snapshotTree, 0x138 /* "Failed to get snapshot from storage" */);
                props.snapshotVersion = versions[0].id;
            }
            props.getSnapshotDuration = trace.trace().duration;
            const snapshotRefSeq = await (0, internal_7.seqFromTree)(snapshotTree, this.readAndParseBlob);
            props.snapshotRefSeq = snapshotRefSeq;
            props.newerSnapshotPresent = snapshotRefSeq >= targetRefSeq;
            perfEvent.end({ details: props });
            return snapshotRefSeq;
        });
        // If the snapshot that was fetched is older than the target snapshot, return. The summarizer will not be closed
        // because the snapshot is likely deleted from storage and it so, closing the summarizer will result in the
        // document being stuck in this state.
        if (fetchedSnapshotRefSeq < targetRefSeq) {
            return;
        }
        await (0, internal_2.delay)(this.closeSummarizerDelayMs);
        this._summarizer?.stop("latestSummaryStateStale");
        this.disposeFn();
    }
    getPendingLocalState(props) {
        this.verifyNotClosed();
        if (this.batchRunner.running) {
            throw new internal_8.UsageError("can't get state while manually accumulating a batch");
        }
        this.imminentClosure ||= props?.notifyImminentClosure ?? false;
        const getSyncState = (pendingAttachmentBlobs) => {
            const pending = this.pendingStateManager.getLocalState(props?.snapshotSequenceNumber);
            const sessionExpiryTimerStarted = props?.sessionExpiryTimerStarted ?? this.garbageCollector.sessionExpiryTimerStarted;
            const pendingIdCompressorState = this._idCompressor?.serialize(true);
            return {
                pending,
                pendingIdCompressorState,
                pendingAttachmentBlobs,
                sessionExpiryTimerStarted,
            };
        };
        const perfEvent = {
            eventName: "getPendingLocalState",
            notifyImminentClosure: props?.notifyImminentClosure,
        };
        const logAndReturnPendingState = (event, pendingState) => {
            event.end({
                attachmentBlobsSize: Object.keys(pendingState?.pendingAttachmentBlobs ?? {}).length,
                pendingOpsSize: pendingState?.pending?.pendingStates.length,
            });
            return pendingState;
        };
        // Flush pending batch.
        // getPendingLocalState() is only exposed through Container.closeAndGetPendingLocalState(), so it's safe
        // to close current batch.
        this.flush();
        return props?.notifyImminentClosure === true
            ? internal_8.PerformanceEvent.timedExecAsync(this.mc.logger, perfEvent, async (event) => logAndReturnPendingState(event, getSyncState(await this.blobManager.attachAndGetPendingBlobs(props?.stopBlobAttachingSignal))))
            : internal_8.PerformanceEvent.timedExec(this.mc.logger, perfEvent, (event) => logAndReturnPendingState(event, getSyncState()));
    }
    summarizeOnDemand(options) {
        if (this._summarizer !== undefined) {
            return this._summarizer.summarizeOnDemand(options);
        }
        else if (this.summaryManager === undefined) {
            // If we're not the summarizer, and we don't have a summaryManager, we expect that
            // disableSummaries is turned on. We are throwing instead of returning a failure here,
            // because it is a misuse of the API rather than an expected failure.
            throw new internal_8.UsageError(`Can't summarize, disableSummaries: ${this.summariesDisabled}`);
        }
        else {
            return this.summaryManager.summarizeOnDemand(options);
        }
    }
    enqueueSummarize(options) {
        if (this._summarizer !== undefined) {
            return this._summarizer.enqueueSummarize(options);
        }
        else if (this.summaryManager === undefined) {
            // If we're not the summarizer, and we don't have a summaryManager, we expect that
            // generateSummaries is turned off. We are throwing instead of returning a failure here,
            // because it is a misuse of the API rather than an expected failure.
            throw new internal_8.UsageError(`Can't summarize, disableSummaries: ${this.summariesDisabled}`);
        }
        else {
            return this.summaryManager.enqueueSummarize(options);
        }
    }
    get groupedBatchingEnabled() {
        return this.sessionSchema.opGroupingEnabled === true;
    }
}
exports.ContainerRuntime = ContainerRuntime;
function createNewSignalEnvelope(address, type, content) {
    const newEnvelope = {
        address,
        contents: { type, content },
    };
    return newEnvelope;
}
exports.createNewSignalEnvelope = createNewSignalEnvelope;
//# sourceMappingURL=containerRuntime.js.map