"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSummarizerFuzzSuite = exports.defaultSummarizerFuzzSuiteOptions = void 0;
/* eslint-disable import/no-nodejs-modules */
const node_assert_1 = require("node:assert");
const node_fs_1 = require("node:fs");
const client_utils_1 = require("@fluid-internal/client-utils");
const stochastic_test_utils_1 = require("@fluid-private/stochastic-test-utils");
const internal_1 = require("@fluidframework/test-runtime-utils/internal");
const summarizerFuzzMocks_js_1 = require("./summarizerFuzzMocks.js");
/**
 * @internal
 */
exports.defaultSummarizerFuzzSuiteOptions = {
    defaultTestCount: stochastic_test_utils_1.defaultOptions.defaultTestCount,
    emitter: new client_utils_1.TypedEventEmitter(),
    only: [],
    skip: [],
    saveFailures: false,
    parseOperations: (serialized) => JSON.parse(serialized),
};
function createSummarizerFuzzSuite(model, providedOptions) {
    const options = {
        ...exports.defaultSummarizerFuzzSuiteOptions,
        ...providedOptions,
    };
    const only = new Set(options.only);
    const skip = new Set(options.skip);
    Object.assign(options, { only, skip });
    (0, node_assert_1.strict)(isInternalOptions(options));
    const describeFuzz = (0, stochastic_test_utils_1.createFuzzDescribe)({ defaultTestCount: options.defaultTestCount });
    describeFuzz(model.workloadName, ({ testCount }) => {
        before(() => {
            if (options.saveFailures !== undefined && options.saveFailures !== false) {
                (0, node_fs_1.mkdirSync)((0, stochastic_test_utils_1.getSaveDirectory)(options.saveFailures.directory, model), {
                    recursive: true,
                });
            }
        });
        for (let seed = 0; seed < testCount; seed++) {
            runTest(model, options, seed, (0, stochastic_test_utils_1.getSaveInfo)(model, options, seed));
        }
        if (options.replay !== undefined) {
            const seed = options.replay;
            describe.only(`replay from file`, () => {
                const saveInfo = (0, stochastic_test_utils_1.getSaveInfo)(model, options, seed);
                (0, node_assert_1.strict)(saveInfo.saveOnFailure !== false, "Cannot replay a file without a directory to save files in!");
                const operations = options.parseOperations((0, node_fs_1.readFileSync)(saveInfo.saveOnFailure.path).toString());
                const replayModel = {
                    ...model,
                    // We lose some type safety here because the options interface isn't generic
                    generatorFactory: () => (0, stochastic_test_utils_1.asyncGeneratorFromArray)(operations),
                };
                runTest(replayModel, options, seed, undefined);
            });
        }
    });
}
exports.createSummarizerFuzzSuite = createSummarizerFuzzSuite;
/**
 * Runs the provided Summarizer fuzz model. All functionality is already assumed to be mixed in.
 * @privateRemarks This is currently file-exported for testing purposes, but it could be reasonable to
 * expose at the package level if we want to expose some of the harness's building blocks.
 */
async function runTestForSeed(model, options, seed, saveInfo) {
    const random = (0, stochastic_test_utils_1.makeRandom)(seed);
    const containerRuntimeFactory = new summarizerFuzzMocks_js_1.MockContainerRuntimeFactoryForSummarizer(options.containerRuntimeOptions);
    const containerRuntime = containerRuntimeFactory.createContainerRuntime(new internal_1.MockFluidDataStoreRuntime());
    const initialState = {
        containerRuntimeFactory,
        random,
        containerRuntime,
    };
    options.emitter.emit("testStart", initialState);
    const finalState = await (0, stochastic_test_utils_1.performFuzzActionsAsync)(model.generatorFactory(), model.reducer, initialState, saveInfo);
    const oldRuntime = finalState.containerRuntime;
    oldRuntime.disposeFn();
    const newRuntime = containerRuntimeFactory.createContainerRuntime(new internal_1.MockFluidDataStoreRuntime());
    await newRuntime.initializeWithStashedOps(oldRuntime);
    await newRuntime.summarize();
    options.emitter.emit("testEnd", finalState);
    return finalState;
}
function runTest(model, options, seed, saveInfo) {
    const itFn = options.only.has(seed) ? it.only : options.skip.has(seed) ? it.skip : it;
    itFn(`seed ${seed}`, async () => {
        const inCi = !!process.env.TF_BUILD;
        await runTestForSeed(model, options, seed, inCi ? undefined : saveInfo);
    });
}
function isInternalOptions(options) {
    return options.only instanceof Set && options.skip instanceof Set;
}
//# sourceMappingURL=summarizerFuzzSuite.js.map